searchNodes=[{"doc":"Oban is a robust job processing library which uses PostgreSQL for storage and coordination. Each Oban instance is a supervision tree and not an application . That means it won't be started automatically and must be included in your application's supervision tree. All of your configuration is passed into the supervisor, allowing you to configure Oban like the rest of your application: # config/config.exs config :my_app , Oban , repo : MyApp.Repo , plugins : [ Oban.Plugins.Pruner ] , queues : [ default : 10 , events : 50 , media : 20 ] # lib/my_app/application.ex defmodule MyApp.Application do @moduledoc false use Application alias MyApp.Repo alias MyAppWeb.Endpoint def start ( _type , _args ) do children = [ Repo , Endpoint , { Oban , oban_config ( ) } ] Supervisor . start_link ( children , strategy : :one_for_one , name : MyApp.Supervisor ) end # Conditionally disable queues or plugins here. defp oban_config do Application . fetch_env! ( :my_app , Oban ) end end If you are running tests (which you should be) you'll want to disable plugins, enqueueing scheduled jobs and job dispatching altogether when testing: # config/test.exs config :my_app , Oban , queues : false , plugins : false See the installation instructions in the README or on the Hexdocs guide for details on how to migrate your database. Configuring Queues Queues are specified as a keyword list where the key is the name of the queue and the value is the maximum number of concurrent jobs. The following configuration would start four queues with concurrency ranging from 5 to 50: queues : [ default : 10 , mailers : 20 , events : 50 , media : 5 ] You may also use an expanded form to configure queues with individual overrides: queues : [ default : 10 , events : [ limit : 50 , paused : true ] ] The events queue will now start in a paused state, which means it won't process anything until Oban.resume_queue/2 is called to start it. There isn't a limit to the number of queues or how many jobs may execute concurrently in each queue. Some additional guidelines: Caveats &amp; Guidelines Each queue will run as many jobs as possible concurrently, up to the configured limit. Make sure your system has enough resources (i.e. database connections) to handle the concurrent load. Queue limits are local (per-node), not global (per-cluster). For example, running a queue with a local limit of one on three separate nodes is effectively a global limit of three. If you require a global limit you must restrict the number of nodes running a particular queue. Only jobs in the configured queues will execute. Jobs in any other queue will stay in the database untouched. Be careful how many concurrent jobs make expensive system calls (i.e. FFMpeg, ImageMagick). The BEAM ensures that the system stays responsive under load, but those guarantees don't apply when using ports or shelling out commands. Defining Workers Worker modules do the work of processing a job. At a minimum they must define a perform/1 function, which is called with an %Oban.Job{} struct. Note that the args field of the job struct will always have string keys, regardless of the key type when the job was enqueued. The args are stored as jsonb in PostgreSQL and the serialization process automatically stringifies all keys. Define a worker to process jobs in the events queue: defmodule MyApp.Business do use Oban.Worker , queue : :events @impl Oban.Worker def perform ( % Oban.Job { args : %{ &quot;id&quot; =&gt; id } = args } ) do model = MyApp.Repo . get ( MyApp.Business.Man , id ) case args do %{ &quot;in_the&quot; =&gt; &quot;business&quot; } -&gt; IO . inspect ( model ) %{ &quot;vote_for&quot; =&gt; vote } -&gt; IO . inspect ( [ vote , model ] ) _ -&gt; IO . inspect ( model ) end :ok end end The use macro also accepts options to customize max attempts, priority, tags, and uniqueness: defmodule MyApp.LazyBusiness do use Oban.Worker , queue : :events , priority : 3 , max_attempts : 3 , tags : [ &quot;business&quot; ] , unique : [ period : 30 ] @impl Oban.Worker def perform ( _job ) do # do business slowly :ok end end Successful jobs should return :ok or an {:ok, value} tuple. The value returned from perform/1 is used to control whether the job is treated as a success, a failure, discarded completely or deferred until later. See the Oban.Worker docs for more details on failure conditions and Oban.Telemetry for details on job reporting. Enqueueing Jobs Jobs are simply Ecto structs and are enqueued by inserting them into the database. For convenience and consistency all workers provide a new/2 function that converts an args map into a job changeset suitable for insertion: %{ id : 1 , in_the : &quot;business&quot; , of_doing : &quot;business&quot; } |&gt; MyApp.Business . new ( ) |&gt; Oban . insert ( ) The worker's defaults may be overridden by passing options: %{ id : 1 , vote_for : &quot;none of the above&quot; } |&gt; MyApp.Business . new ( queue : :special , max_attempts : 5 ) |&gt; Oban . insert ( ) Jobs may be scheduled at a specific datetime in the future: %{ id : 1 } |&gt; MyApp.Business . new ( scheduled_at : ~U[2020-12-25 19:00:56.0Z] ) |&gt; Oban . insert ( ) Jobs may also be scheduled down to the second any time in the future: %{ id : 1 } |&gt; MyApp.Business . new ( schedule_in : 5 ) |&gt; Oban . insert ( ) Unique jobs can be configured in the worker, or when the job is built: %{ email : &quot;brewster@example.com&quot; } |&gt; MyApp.Mailer . new ( unique : [ period : 300 , fields : [ :queue , :worker ] ) |&gt; Oban . insert ( ) Job priority can be specified using an integer from 0 to 3, with 0 being the default and highest priority: %{ id : 1 } |&gt; MyApp.Backfiller . new ( priority : 2 ) |&gt; Oban . insert ( ) Any number of tags can be added to a job dynamically, at the time it is inserted: id = 1 %{ id : id } |&gt; MyApp.OnboardMailer . new ( tags : [ &quot;mailer&quot; , &quot;record- \#{ id } &quot; ] ) |&gt; Oban . insert ( ) Multiple jobs can be inserted in a single transaction: Ecto.Multi . new ( ) |&gt; Oban . insert ( :b_job , MyApp.Business . new ( %{ id : 1 } ) ) |&gt; Oban . insert ( :m_job , MyApp.Mailer . new ( %{ email : &quot;brewser@example.com&quot; } ) ) |&gt; Repo . transaction ( ) Occasionally you may need to insert a job for a worker that exists in another application. In that case you can use Oban.Job.new/2 to build the changeset manually: %{ id : 1 , user_id : 2 } |&gt; Oban.Job . new ( queue : :default , worker : OtherApp.Worker ) |&gt; Oban . insert ( ) Oban.insert/2,4 is the preferred way of inserting jobs as it provides some of Oban's advanced features (i.e., unique jobs). However, you can use your application's Repo.insert/2 function if necessary. See Oban.Job.new/2 for a full list of job options. Pruning Historic Jobs Job stats and queue introspection are built on keeping job rows in the database after they have completed. This allows administrators to review completed jobs and build informative aggregates, at the expense of storage and an unbounded table size. To prevent the oban_jobs table from growing indefinitely, Oban provides active pruning of completed , cancelled and discarded jobs. By default, the Pruner plugin retains jobs for 60 seconds. You can configure a longer retention period by providing a max_age in seconds to the Pruner plugin. # Set the max_age for 5 minutes config :my_app , Oban , plugins : [ { Oban.Plugins.Pruner , max_age : 300 } ] ... Caveats &amp; Guidelines Pruning is best-effort and performed out-of-band. This means that all limits are soft; jobs beyond a specified age may not be pruned immediately after jobs complete. Pruning is only applied to jobs that are completed , cancelled or discarded . It'll never delete a new job, a scheduled job or a job that failed and will be retried. Unique Jobs The unique jobs feature lets you specify constraints to prevent enqueueing duplicate jobs. Uniqueness is based on a combination of args , queue , worker , state and insertion time. It is configured at the worker or job level using the following options: :period — The number of seconds until a job is no longer considered duplicate. You should always specify a period. :infinity can be used to indicate the job be considered a duplicate as long as jobs are retained. :fields — The fields to compare when evaluating uniqueness. The available fields are :args , :queue and :worker , by default all three are used. :keys — A specific subset of the :args to consider when comparing against historic jobs. This allows a job with multiple key/value pairs in the args to be compared using only a subset of them. :states — The job states that are checked for duplicates. The available states are :available , :scheduled , :executing , :retryable , :completed , :cancelled and :discarded . By default all states except for :discarded and :cancelled are checked, which prevents duplicates even if the previous job has been completed. For example, configure a worker to be unique across all fields and states for 60 seconds: use Oban.Worker , unique : [ period : 60 ] Configure the worker to be unique only by :worker and :queue : use Oban.Worker , unique : [ fields : [ :queue , :worker ] , period : 60 ] Or, configure a worker to be unique until it has executed: use Oban.Worker , unique : [ period : 300 , states : [ :available , :scheduled , :executing ] ] Only consider the :url key rather than the entire args : use Oban.Worker , unique : [ fields : [ :args , :worker ] , keys : [ :url ] ] You can use Oban.Job.states/0 to specify uniqueness across all states, including :discarded : use Oban.Worker , unique : [ period : 300 , states : Oban.Job . states ( ) ] Detecting Unique Conflicts When unique settings match an existing job, the return value of Oban.insert/2 is still {:ok, job} . However, you can detect a unique conflict by checking the jobs' :conflict? field. If there was an existing job, the field is true ; otherwise it is false . You can use the :conflict? field to customize responses after insert: with { :ok , % Job { conflict? : true } } &lt;- Oban . insert ( changeset ) do { :error , :job_already_exists } end Replacing Values In addition to detecting unique conflicts, passing options to replace can update any job field when there is a conflict. Any of the following fields can be replaced: args , max_attempts , meta , priority , queue , scheduled_at , tags , worker . For example, to change the priority and increase max_attempts when there is a conflict: BusinessWorker . new ( args , max_attempts : 5 , priority : 0 , replace : [ :max_attempts , :priority ] ) Another example is bumping the scheduled time on conflict. Either scheduled_at or schedule_in values will work, but the replace option is always scheduled_at . UrgentWorker . new ( args , schedule_in : 1 , replace : [ :scheduled_at ] ) Strong Guarantees Unique jobs are guaranteed through transactional locks and database queries: they do not rely on unique constraints in the database. This makes uniqueness entirely configurable by application code, without the need for database migrations. Periodic Jobs Oban's Cron plugin registers workers a cron-like schedule and enqueues jobs automatically. Periodic jobs are declared as a list of {cron, worker} or {cron, worker, options} tuples: config :my_app , Oban , repo : MyApp.Repo , plugins : [ { Oban.Plugins.Cron , crontab : [ { &quot;* * * * *&quot; , MyApp.MinuteWorker } , { &quot;0 * * * *&quot; , MyApp.HourlyWorker , args : %{ custom : &quot;arg&quot; } } , { &quot;0 0 * * *&quot; , MyApp.DailyWorker , max_attempts : 1 } , { &quot;0 12 * * MON&quot; , MyApp.MondayWorker , queue : :scheduled , tags : [ &quot;mondays&quot; ] } , { &quot;@daily&quot; , MyApp.AnotherDailyWorker } ] } ] The crontab would insert jobs as follows: MyApp.MinuteWorker — Inserted once every minute MyApp.HourlyWorker — Inserted at the first minute of every hour with custom args MyApp.DailyWorker — Inserted at midnight every day with no retries MyApp.MondayWorker — Inserted at noon every Monday in the &quot;scheduled&quot; queue MyApp.AnotherDailyWorker — Inserted at midnight every day with no retries The crontab format respects all standard rules and has one minute resolution. Jobs are considered unique for most of each minute, which prevents duplicate jobs with multiple nodes and across node restarts. Like other jobs, recurring jobs will use the :queue specified by the worker module (or :default if one is not specified). Cron Expressions Standard Cron expressions are composed of rules specifying the minutes, hours, days, months and weekdays. Rules for each field are comprised of literal values, wildcards, step values or ranges: * — Wildcard, matches any value (0, 1, 2, ...) 0 — Literal, matches only itself (only 0) */15 — Step, matches any value that is a multiple (0, 15, 30, 45) 0-5 — Range, matches any value within the range (0, 1, 2, 3, 4, 5) 0-9/2 - Step values can be used in conjunction with ranges (0, 2, 4, 6, 8) Each part may have multiple rules, where rules are separated by a comma. The allowed values for each field are as follows: minute — 0-59 hour — 0-23 days — 1-31 month — 1-12 (or aliases, JAN , FEB , MAR , etc.) weekdays — 0-6 (or aliases, SUN , MON , TUE , etc.) The following Cron extensions are supported: @hourly — 0 * * * * @daily (as well as @midnight ) — 0 0 * * * @weekly — 0 0 * * 0 @monthly — 0 0 1 * * @yearly (as well as @annually ) — 0 0 1 1 * @reboot — Run once at boot across the entire cluster Some specific examples that demonstrate the full range of expressions: 0 * * * * — The first minute of every hour */15 9-17 * * * — Every fifteen minutes during standard business hours 0 0 * DEC * — Once a day at midnight during December 0 7-9,4-6 13 * FRI — Once an hour during both rush hours on Friday the 13th For more in depth information see the man documentation for cron and crontab in your system. Alternatively you can experiment with various expressions online at Crontab Guru . Caveats &amp; Guidelines All schedules are evaluated as UTC unless a different timezone is provided. See Oban.Plugins.Cron for information about configuring a timezone. Workers can be used for regular and scheduled jobs so long as they accept different arguments. Duplicate jobs are prevented through transactional locks and unique constraints. Workers that are used for regular and scheduled jobs must not specify unique options less than 60s . Long running jobs may execute simultaneously if the scheduling interval is shorter than it takes to execute the job. You can prevent overlap by passing custom unique opts in the crontab config: custom_args = %{ scheduled : true } unique_opts = [ period : 60 * 60 * 24 , states : [ :available , :scheduled , :executing ] ] config :my_app , Oban , repo : MyApp.Repo , plugins : [ { Oban.Plugins.Cron , crontab : [ { &quot;* * * * *&quot; , MyApp.SlowWorker , args : custom_args , unique : unique_opts } ] } ] Prioritizing Jobs Normally, all available jobs within a queue are executed in the order they were scheduled. You can override the normal behavior and prioritize or de-prioritize a job by assigning a numerical priority . Priorities from 0-3 are allowed, where 0 is the highest priority and 3 is the lowest. The default priority is 0, unless specified all jobs have an equally high priority. All jobs with a higher priority will execute before any jobs with a lower priority. Within a particular priority jobs are executed in their scheduled order. Testing Oban provides some helpers to facilitate testing. The helpers handle the boilerplate of making assertions on which jobs are enqueued. To use the perform_job/2,3 , assert_enqueued/1 and refute_enqueued/1 helpers in your tests you must include them in your testing module and specify your app's Ecto repo: use Oban.Testing , repo : MyApp.Repo Now you can assert, refute or list jobs that have been enqueued within your integration tests: assert_enqueued worker : MyWorker , args : %{ id : 1 } # or refute_enqueued queue : :special , args : %{ id : 2 } # or assert [ %{ args : %{ &quot;id&quot; =&gt; 1 } } ] = all_enqueued worker : MyWorker You can also easily unit test workers with the perform_job/2,3 function, which automates validating job args, options, and worker results from a single function call: assert :ok = perform_job ( MyWorker , %{ id : 1 } ) # or assert :ok = perform_job ( MyWorker , %{ id : 1 } , attempt : 3 , max_attempts : 3 ) # or assert { :error , _ } = perform_job ( MyWorker , %{ bad : :arg } ) See the Oban.Testing module for more details. Caveats &amp; Guidelines As noted in Usage , there are some guidelines for running tests: Disable all job dispatching by setting queues: false or queues: nil in your test.exs config. Keyword configuration is deep merged, so setting queues: [] won't have any effect. Disable plugins via plugins: false . Default plugins, such as the fixed pruner, aren't necessary in testing mode because jobs created within the sandbox are rolled back at the end of the test. Additionally, the periodic pruning queries will raise DBConnection.OwnershipError when the application boots. Be sure to use the Ecto Sandbox for testing. Oban makes use of database Pub/Sub events to dispatch jobs, but Pub/Sub events never fire within a transaction. Since sandbox tests run within a transaction no events will fire and jobs won't be dispatched. config :my_app , MyApp.Repo , pool : Ecto.Adapters.SQL.Sandbox Integration Testing During integration testing it may be necessary to run jobs because they do work essential for the test to complete, i.e. sending an email, processing media, etc. You can execute all available jobs in a particular queue by calling Oban.drain_queue/1,2 directly from your tests. For example, to process all pending jobs in the &quot;mailer&quot; queue while testing some business logic: defmodule MyApp.BusinessTest do use MyApp.DataCase , async : true alias MyApp . { Business , Worker } test &quot;we stay in the business of doing business&quot; do :ok = Business . schedule_a_meeting ( %{ email : &quot;monty@brewster.com&quot; } ) assert %{ success : 1 , failure : 0 } = Oban . drain_queue ( queue : :mailer ) # Now, make an assertion about the email delivery end end See Oban.drain_queue/1,2 for additional details. Error Handling When a job returns an error value, raises an error or exits during execution the details are recorded within the errors array on the job. When the number of execution attempts is below the configured max_attempts limit, the job will automatically be retried in the future. The retry delay has an exponential backoff, meaning the job's second attempt will be after 16s, third after 31s, fourth after 1m 36s, etc. See the Oban.Worker documentation on &quot;Customizing Backoff&quot; for alternative backoff strategies. Error Details Execution errors are stored as a formatted exception along with metadata about when the failure occurred and which attempt caused it. Each error is stored with the following keys: at The UTC timestamp when the error occurred at attempt The attempt number when the error occurred error A formatted error message and stacktrace See the Instrumentation docs for an example of integrating with external error reporting systems. Limiting Retries By default, jobs are retried up to 20 times. The number of retries is controlled by the max_attempts value, which can be set at the Worker or Job level. For example, to instruct a worker to discard jobs after three failures: use Oban.Worker , queue : :limited , max_attempts : 3 Limiting Execution Time By default, individual jobs may execute indefinitely. If this is undesirable you may define a timeout in milliseconds with the timeout/1 callback on your worker module. For example, to limit a worker's execution time to 30 seconds: def MyApp.Worker do use Oban.Worker @impl Oban.Worker def perform ( _job ) do something_that_may_take_a_long_time ( ) :ok end @impl Oban.Worker def timeout ( _job ) , do : :timer . seconds ( 30 ) end The timeout/1 function accepts an Oban.Job struct, so you can customize the timeout using any job attributes. Define the timeout value through job args: def timeout ( % _ { args : %{ &quot;timeout&quot; =&gt; timeout } } ) , do : timeout Define the timeout based on the number of attempts: def timeout ( % _ { attempt : attempt } ) , do : attempt * :timer . seconds ( 5 ) Instrumentation and Logging Oban provides integration with Telemetry , a dispatching library for metrics. It is easy to report Oban metrics to any backend by attaching to :oban events. Here is an example of a sample unstructured log handler: defmodule MyApp.ObanLogger do require Logger def handle_event ( [ :oban , :job , :start ] , measure , meta , _ ) do Logger . warn ( &quot;[Oban] :started \#{ meta . worker } at \#{ measure . system_time } &quot; ) end def handle_event ( [ :oban , :job , event ] , measure , meta , _ ) do Logger . warn ( &quot;[Oban] \#{ event } \#{ meta . worker } ran in \#{ measure . duration } &quot; ) end end Attach the handler to success and failure events in application.ex : events = [ [ :oban , :job , :start ] , [ :oban , :job , :stop ] , [ :oban , :job , :exception ] ] :telemetry . attach_many ( &quot;oban-logger&quot; , events , &amp; MyApp.ObanLogger . handle_event / 4 , [ ] ) The Oban.Telemetry module provides a robust structured logger that handles all of Oban's telemetry events. As in the example above, attach it within your application.ex module: :ok = Oban.Telemetry . attach_default_logger ( ) For more details on the default structured logger and information on event metadata see docs for the Oban.Telemetry module. Reporting Errors Another great use of execution data is error reporting. Here is an example of integrating with Sentry to report job failures: defmodule ErrorReporter do def handle_event ( [ :oban , :job , :exception ] , measure , meta , _ ) do extra = meta . job |&gt; Map . take ( [ :id , :args , :meta , :queue , :worker ] ) |&gt; Map . merge ( measure ) Sentry . capture_exception ( meta . reason , stacktrace : meta . stacktrace , extra : extra ) end def handle_event ( [ :oban , :circuit , :trip ] , _measure , meta , _ ) do Sentry . capture_exception ( meta . reason , stacktrace : meta . stacktrace , extra : meta ) end end :telemetry . attach ( &quot;oban-errors&quot; , [ :oban , :job , :exception ] , &amp; ErrorReporter . handle_event / 4 , [ ] ) You can use exception events to send error reports to Honeybadger, Rollbar, AppSignal or any other application monitoring platform. Isolation Oban supports namespacing through PostgreSQL schemas, also called &quot;prefixes&quot; in Ecto. With prefixes your jobs table can reside outside of your primary schema (usually public) and you can have multiple separate job tables. To use a prefix you first have to specify it within your migration: defmodule MyApp.Repo.Migrations.AddPrefixedObanJobsTable do use Ecto.Migration def up do Oban.Migrations . up ( prefix : &quot;private&quot; ) end def down do Oban.Migrations . down ( prefix : &quot;private&quot; ) end end The migration will create the &quot;private&quot; schema and all tables, functions and triggers within that schema. With the database migrated you'll then specify the prefix in your configuration: config :my_app , Oban , prefix : &quot;private&quot; , repo : MyApp.Repo , queues : [ default : 10 ] Now all jobs are inserted and executed using the private.oban_jobs table. Note that Oban.insert/2,4 will write jobs in the private.oban_jobs table, you'll need to specify a prefix manually if you insert jobs directly through a repo. Supervisor Isolation Not only is the oban_jobs table isolated within the schema, but all notification events are also isolated. That means that insert/update events will only dispatch new jobs for their prefix. You can run multiple Oban instances with different prefixes on the same system and have them entirely isolated, provided you give each supervisor a distinct id. Here we configure our application to start three Oban supervisors using the &quot;public&quot;, &quot;special&quot; and &quot;private&quot; prefixes, respectively: def start ( _type , _args ) do children = [ Repo , Endpoint , { Oban , name : ObanA , repo : Repo } , { Oban , name : ObanB , repo : Repo , prefix : &quot;special&quot; } , { Oban , name : ObanC , repo : Repo , prefix : &quot;private&quot; } ] Supervisor . start_link ( children , strategy : :one_for_one , name : MyApp.Supervisor ) end Dynamic Repositories Oban supports Ecto dynamic repositories through the :get_dynamic_repo option. To make this work, you need to run a separate Oban instance per each dynamic repo instance. Most often it's worth bundling each Oban and repo instance under the same supervisor: def start_repo_and_oban ( instance_id ) do children = [ { MyDynamicRepo , name : nil , url : repo_url ( instance_id ) } , { Oban , name : instance_id , get_dynamic_repo : fn -&gt; repo_pid ( instance_id ) end } ] Supervisor . start_link ( children , strategy : :one_for_one ) end The function repo_pid/1 must return the pid of the repo for the given instance. You can use Registry to register the repo (for example in the repo's init/2 callback) and discover it. If your application exclusively uses dynamic repositories and doesn't specify all credentials upfront, you must implement an init/1 callback in your Ecto Repo. Doing so provides the Postgres notifier with the correct credentials on init, allowing jobs to process as expected.","ref":"Oban.html","title":"Oban","type":"module"},{"doc":"Cancel many jobs based on a queryable and mark them as cancelled to prevent them from running. Any currently executing jobs are killed while the others are ignored. If executing jobs happen to fail before cancellation then the state is set to cancelled . However, any that complete successfully will remain completed . Only jobs with the statuses executing , available , scheduled , or retryable can be cancelled. Example Cancel all jobs: Oban . cancel_all_jobs ( Oban.Job ) { :ok , 9 } Cancel all jobs for a specific worker: Oban.Job |&gt; Ecto.Query . where ( worker : &quot;MyApp.MyWorker&quot; ) |&gt; Oban . cancel_all_jobs ( ) { :ok , 2 }","ref":"Oban.html#cancel_all_jobs/2","title":"Oban.cancel_all_jobs/2","type":"function"},{"doc":"Cancel an executing , available , scheduled or retryable job and mark it as cancelled to prevent it from running. If the job is currently executing it will be killed and otherwise it is ignored. If an executing job happens to fail before it can be cancelled the state is set to cancelled . However, if it manages to complete successfully then the state will still be completed . Example Cancel a scheduled job with the id 1 : Oban . cancel_job ( 1 ) :ok","ref":"Oban.html#cancel_job/2","title":"Oban.cancel_job/2","type":"function"},{"doc":"Check the current state of a queue producer. This allows you to introspect on a queue's health by retrieving key attributes of the producer's state; values such as the current limit , the running job ids, and when the producer was started. Options :queue - a string or atom specifying the queue to check, required Example Oban . check_queue ( queue : :default ) %{ limit : 10 , node : &quot;me@local&quot; , paused : false , queue : &quot;default&quot; , running : [ 100 , 102 ] , started_at : ~D[2020-10-07 15:31:00] , updated_at : ~D[2020-10-07 15:31:00] }","ref":"Oban.html#check_queue/2","title":"Oban.check_queue/2","type":"function"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Oban.html#child_spec/1","title":"Oban.child_spec/1","type":"function"},{"doc":"Retrieve the config struct for a named Oban supervision tree.","ref":"Oban.html#config/1","title":"Oban.config/1","type":"function"},{"doc":"Synchronously execute all available jobs in a queue. All execution happens within the current process and it is guaranteed not to raise an error or exit. Draining a queue from within the current process is especially useful for testing. Jobs that are enqueued by a process when Ecto is in sandbox mode are only visible to that process. Calling drain_queue/2 allows you to control when the jobs are executed and to wait synchronously for all jobs to complete. Failures &amp; Retries Draining a queue uses the same execution mechanism as regular job dispatch. That means that any job failures or crashes are captured and result in a retry. Retries are scheduled in the future with backoff and won't be retried immediately. By default jobs are executed in safe mode, just as they are in production. Safe mode catches any errors or exits and records the formatted error in the job's errors array. That means exceptions and crashes are not bubbled up to the calling process. If you expect jobs to fail, would like to track failures, or need to check for specific errors you can pass the with_safety: false flag. Scheduled Jobs By default, drain_queue/2 will execute all currently available jobs. In order to execute scheduled jobs, you may pass the :with_scheduled flag which will cause scheduled jobs to be marked as available beforehand. Options :queue - a string or atom specifying the queue to drain, required :with_limit — the maximum number of jobs to drain at once. When recursion is enabled this is how many jobs are processed per-iteration. :with_recursion — whether to keep draining a queue repeatedly when jobs insert more jobs :with_safety — whether to silently catch errors when draining, default true :with_scheduled — whether to include any scheduled jobs when draining, default false Example Drain a queue with three available jobs, two of which succeed and one of which fails: Oban . drain_queue ( queue : :default ) %{ failure : 1 , snoozed : 0 , success : 2 } Drain a queue including any scheduled jobs: Oban . drain_queue ( queue : :default , with_scheduled : true ) %{ failure : 0 , snoozed : 0 , success : 1 } Drain a queue and assert an error is raised: assert_raise RuntimeError , fn -&gt; Oban . drain_queue ( queue : :risky , with_safety : false ) end Drain a queue repeatedly until there aren't any more jobs to run. This is particularly useful for testing jobs that enqueue other jobs: Oban . drain_queue ( queue : :default , with_recursion : true ) %{ failure : 1 , snoozed : 0 , success : 2 } Drain only the top (by scheduled time and priority) five jobs off a queue: Oban . drain_queue ( queue : :default , with_limit : 5 ) %{ failure : 0 , snoozed : 0 , success : 1 } Drain a queue recursively, only one job at a time: Oban . drain_queue ( queue : :default , with_limit : 1 , with_recursion : true ) %{ failure : 0 , snoozed : 0 , success : 3 }","ref":"Oban.html#drain_queue/2","title":"Oban.drain_queue/2","type":"function"},{"doc":"Insert a new job into the database for execution. This and the other insert variants are the recommended way to enqueue jobs because they support features like unique jobs. See the section on &quot;Unique Jobs&quot; for more details. Example Insert a single job: { :ok , job } = Oban . insert ( MyApp.Worker . new ( %{ id : 1 } ) ) Insert a job while ensuring that it is unique within the past 30 seconds: { :ok , job } = Oban . insert ( MyApp.Worker . new ( %{ id : 1 } , unique : [ period : 30 ] ) )","ref":"Oban.html#insert/2","title":"Oban.insert/2","type":"function"},{"doc":"Put a job insert operation into an Ecto.Multi . Like insert/2 , this variant is recommended over Ecto.Multi.insert because it supports all of Oban's features, i.e. unique jobs. See the section on &quot;Unique Jobs&quot; for more details. Example Ecto.Multi . new ( ) |&gt; Oban . insert ( &quot;job-1&quot; , MyApp.Worker . new ( %{ id : 1 } ) ) |&gt; Oban . insert ( &quot;job-2&quot; , fn _ -&gt; MyApp.Worker . new ( %{ id : 2 } ) end ) |&gt; MyApp.Repo . transaction ( )","ref":"Oban.html#insert/4","title":"Oban.insert/4","type":"function"},{"doc":"Similar to insert/2 , but raises an Ecto.InvalidChangesetError if the job can't be inserted. Example job = Oban . insert! ( MyApp.Worker . new ( %{ id : 1 } ) )","ref":"Oban.html#insert!/2","title":"Oban.insert!/2","type":"function"},{"doc":"Insert multiple jobs into the database for execution. Insertion respects prefix and log settings, but it does not use per-job unique configuration. You must use insert/2,4 or insert!/2 for per-job unique support. There are a few important differences between this function and Ecto.Repo.insert_all/3 : This function always returns a list rather than a tuple of {count, records} This function requires a list of changesets rather than a list of maps or keyword lists Example 1 .. 100 |&gt; Enum . map ( &amp; MyApp.Worker . new ( %{ id : &amp;1 } ) ) |&gt; Oban . insert_all ( )","ref":"Oban.html#insert_all/2","title":"Oban.insert_all/2","type":"function"},{"doc":"Put an insert_all operation into an Ecto.Multi . This function supports the same features and has the same caveats as insert_all/2 . Example changesets = Enum . map ( 0 .. 100 , &amp; MyApp.Worker . new ( %{ id : &amp;1 } ) ) Ecto.Multi . new ( ) |&gt; Oban . insert_all ( :jobs , changesets ) |&gt; MyApp.Repo . transaction ( )","ref":"Oban.html#insert_all/4","title":"Oban.insert_all/4","type":"function"},{"doc":"Pause a running queue, preventing it from executing any new jobs. All running jobs will remain running until they are finished. When shutdown begins all queues are paused. Options :queue - a string or atom specifying the queue to pause, required :local_only - whether the queue will be paused only on the local node, default: false Note: by default, Oban does not verify that the given queue exists unless :local_only is set to true as even if the queue does not exist locally, it might be running on another node. Example Pause the default queue: Oban . pause_queue ( queue : :default ) :ok Pause the default queue, but only on the local node: Oban . pause_queue ( queue : :default , local_only : true ) :ok","ref":"Oban.html#pause_queue/2","title":"Oban.pause_queue/2","type":"function"},{"doc":"Resume executing jobs in a paused queue. Options :queue - a string or atom specifying the queue to resume, required :local_only - whether the queue will be resumed only on the local node, default: false Note: by default, Oban does not verify that the given queue exists unless :local_only is set to true as even if the queue does not exist locally, it might be running on another node. Example Resume a paused default queue: Oban . resume_queue ( queue : :default ) :ok Resume the default queue, but only on the local node: Oban . resume_queue ( queue : :default , local_only : true ) :ok","ref":"Oban.html#resume_queue/2","title":"Oban.resume_queue/2","type":"function"},{"doc":"Retries all jobs that match on the given queryable. Please note that no matter the queryable constraints, it will never retry available , executing or scheduled jobs. If no queryable is given, Oban will retry all jobs in retryable states. Example Retries all retryable jobs Oban . retry_all_jobs ( Oban.Job ) { :ok , 9 } Retries all retryable jobs with priority 0 Oban.Job |&gt; Ecto.Query . where ( priority : 0 ) |&gt; Oban . retry_all_jobs ( ) { :ok , 5 }","ref":"Oban.html#retry_all_jobs/2","title":"Oban.retry_all_jobs/2","type":"function"},{"doc":"Sets a job as available , adding attempts if already maxed out. If the job is currently available , executing or scheduled it will be ignored. The job is scheduled for immediate execution. Example Retry a discarded job with the id 1 : Oban . retry_job ( 1 ) :ok","ref":"Oban.html#retry_job/2","title":"Oban.retry_job/2","type":"function"},{"doc":"Scale the concurrency for a queue. Options :queue - a string or atom specifying the queue to scale, required :limit — the new concurrency limit, required :local_only — whether the queue will be scaled only on the local node, default: false In addition, all engine-specific queue options are passed along after validation. Note: by default, Oban does not verify that the given queue exists unless :local_only is set to true as even if the queue does not exist locally, it might be running on another node. Example Scale a queue up, triggering immediate execution of queued jobs: Oban . scale_queue ( queue : :default , limit : 50 ) :ok Scale the queue back down, allowing executing jobs to finish: Oban . scale_queue ( queue : :default , limit : 5 ) :ok Scale the queue only on the local node: Oban . scale_queue ( queue : :default , limit : 10 , local_only : true ) :ok","ref":"Oban.html#scale_queue/2","title":"Oban.scale_queue/2","type":"function"},{"doc":"Starts an Oban supervision tree linked to the current process. Options These options are required; without them the supervisor won't start :name — used for supervisor registration, defaults to Oban :repo — specifies the Ecto repo used to insert and retrieve jobs Primary Options These options determine what the system does at a high level, i.e. which queues to run. :node — used to identify the node that the supervision tree is running in. If no value is provided it will use the node name in a distributed system, or the hostname in an isolated node. See &quot;Node Name&quot; below. :plugins — a list or modules or module/option tuples that are started as children of an Oban supervisor. Any supervisable module is a valid plugin, i.e. a GenServer or an Agent . :prefix — the query prefix, or schema, to use for inserting and executing jobs. An oban_jobs table must exist within the prefix. See the &quot;Prefix Support&quot; section in the module documentation for more details. :queues — a keyword list where the keys are queue names and the values are the concurrency setting or a keyword list of queue options. For example, setting queues to [default: 10, exports: 5] would start the queues default and exports with a combined concurrency level of 15. The concurrency setting specifies how many jobs each queue will run concurrently. Queues accept additional override options to customize their behavior, e.g. by setting paused or dispatch_cooldown for a specific queue. For testing purposes :queues may be set to false or nil , which effectively disables all job dispatching. :telemetry_prefix - a list of atoms that prefixes telemetry events. For example, a telemetry prefix of [:my_app, :oban] would span job start telemetry events as [:my_app, :oban, :job, :start] . Defaults to [:oban] . :log — either false to disable logging or a standard log level ( :error , :warn , :info , :debug ). This determines whether queries are logged or not; overriding the repo's configured log level. Defaults to false , where no queries are logged. Twiddly Options Additional options used to tune system behaviour. These are primarily useful for testing or troubleshooting and don't usually need modification. :circuit_backoff — the number of milliseconds until queries are attempted after a database error. All processes communicating with the database are equipped with circuit breakers and will use this for the backoff. Defaults to 30_000ms . :dispatch_cooldown — the minimum number of milliseconds a producer will wait before fetching and running more jobs. A slight cooldown period prevents a producer from flooding with messages and thrashing the database. The cooldown period directly impacts a producer's throughput: jobs per second for a single queue is calculated by (1000 / cooldown) * limit . For example, with a 5ms cooldown and a queue limit of 25 a single queue can run 5,000 jobs/sec. The default is 5ms and the minimum is 1ms , which is likely faster than the database can return new jobs to run. :shutdown_grace_period - the amount of time a queue will wait for executing jobs to complete before hard shutdown, specified in milliseconds. The default is 15_000 , or 15 seconds. Example To start an Oban supervisor within an application's supervision tree: def start ( _type , _args ) do children = [ MyApp.Repo , { Oban , queues : [ default : 50 ] } ] Supervisor . start_link ( children , strategy : :one_for_one , name : MyApp.Supervisor ) end Node Name When the node value hasn't been configured it is generated based on the environment: In a distributed system the node name is used In a Heroku environment the system environment's DYNO value is used Otherwise, the system hostname is used","ref":"Oban.html#start_link/1","title":"Oban.start_link/1","type":"function"},{"doc":"Start a new supervised queue. By default this starts a new supervised queue across all nodes running Oban on the same database and prefix. You can pass the option local_only: true if you prefer to start the queue only on the local node. Options :queue - a string or atom specifying the queue to start, required :local_only - whether the queue will be started only on the local node, default: false :limit - set the concurrency limit, required :paused — set whether the queue starts in the &quot;paused&quot; state, optional In addition, all engine-specific queue options are passed along after validation. Example Start the :priority queue with a concurrency limit of 10 across the connected nodes. Oban . start_queue ( queue : :priority , limit : 10 ) :ok Start the :media queue with a concurrency limit of 5 only on the local node. Oban . start_queue ( queue : :media , limit : 5 , local_only : true ) :ok Start the :media queue in a paused state. Oban . start_queue ( queue : :media , limit : 5 , paused : true ) :ok","ref":"Oban.html#start_queue/2","title":"Oban.start_queue/2","type":"function"},{"doc":"Shutdown a queue's supervision tree and stop running jobs for that queue. By default this action will occur across all the running nodes. Still, if you prefer to stop the queue's supervision tree and stop running jobs for that queue only on the local node, you can pass the option: local_only: true The shutdown process pauses the queue first and allows current jobs to exit gracefully, provided they finish within the shutdown limit. Note: by default, Oban does not verify that the given queue exists unless :local_only is set to true as even if the queue does not exist locally, it might be running on another node. Options :queue - a string or atom specifying the queue to stop, required :local_only - whether the queue will be stopped only on the local node, default: false Example Oban . stop_queue ( queue : :default ) :ok Oban . stop_queue ( queue : :media , local_only : true ) :ok","ref":"Oban.html#stop_queue/2","title":"Oban.stop_queue/2","type":"function"},{"doc":"Returns the pid of the root Oban process for the given name. Example Find the default instance: Oban . whereis ( Oban ) Find a dynamically named instance: Oban . whereis ( { :oban , 1 } )","ref":"Oban.html#whereis/1","title":"Oban.whereis/1","type":"function"},{"doc":"","ref":"Oban.html#t:changesets_or_wrapper/0","title":"Oban.changesets_or_wrapper/0","type":"type"},{"doc":"","ref":"Oban.html#t:drain_option/0","title":"Oban.drain_option/0","type":"type"},{"doc":"","ref":"Oban.html#t:drain_result/0","title":"Oban.drain_result/0","type":"type"},{"doc":"","ref":"Oban.html#t:name/0","title":"Oban.name/0","type":"type"},{"doc":"","ref":"Oban.html#t:option/0","title":"Oban.option/0","type":"type"},{"doc":"","ref":"Oban.html#t:queue_name/0","title":"Oban.queue_name/0","type":"type"},{"doc":"","ref":"Oban.html#t:queue_option/0","title":"Oban.queue_option/0","type":"type"},{"doc":"","ref":"Oban.html#t:queue_state/0","title":"Oban.queue_state/0","type":"type"},{"doc":"","ref":"Oban.html#t:wrapper/0","title":"Oban.wrapper/0","type":"type"},{"doc":"A Job is an Ecto schema used for asynchronous execution. Job changesets are created by your application code and inserted into the database for asynchronous execution. Jobs can be inserted along with other application data as part of a transaction, which guarantees that jobs will only be triggered from a successful transaction.","ref":"Oban.Job.html","title":"Oban.Job","type":"module"},{"doc":"Construct a new job changeset ready for insertion into the database. Options :max_attempts — the maximum number of times a job can be retried if there are errors during execution :meta — a map containing additional information about the job :priority — a numerical indicator from 0 to 3 of how important this job is relative to other jobs in the same queue. The lower the number, the higher priority the job. :queue — a named queue to push the job into. Jobs may be pushed into any queue, regardless of whether jobs are currently being processed for the queue. :schedule_in - the number of seconds until the job should be executed or a tuple containing a number and unit :replace - a list of keys to replace on a unique conflict :scheduled_at - a time in the future after which the job should be executed :tags — a list of tags to group and organize related jobs, i.e. to identify scheduled jobs :unique — a keyword list of options specifying how uniqueness will be calculated. The options define which fields will be used, for how long, with which keys, and for which states. :worker — a module to execute the job in. The module must implement the Oban.Worker behaviour. Examples Insert a job with the :default queue: %{ id : 1 , user_id : 2 } |&gt; Oban.Job . new ( queue : :default , worker : MyApp.Worker ) |&gt; Oban . insert ( ) Generate a pre-configured job for MyApp.Worker and push it: %{ id : 1 , user_id : 2 } |&gt; MyApp.Worker . new ( ) |&gt; Oban . insert ( ) Schedule a job to run in 5 seconds: %{ id : 1 } |&gt; MyApp.Worker . new ( schedule_in : 5 ) |&gt; Oban . insert ( ) Schedule a job to run in 5 minutes: %{ id : 1 } |&gt; MyApp.Worker . new ( schedule_in : { 5 , :minutes } ) |&gt; Oban . insert ( ) Insert a job, ensuring that it is unique within the past minute: %{ id : 1 } |&gt; MyApp.Worker . new ( unique : [ period : 60 ] ) |&gt; Oban . insert ( ) Insert a unique job based only on the worker field, and within multiple states: fields = [ :worker ] states = [ :available , :scheduled , :executing , :retryable , :completed ] %{ id : 1 } |&gt; MyApp.Worker . new ( unique : [ fields : fields , period : 60 , states : states ] ) |&gt; Oban . insert ( ) Insert a unique job considering only the worker and specified keys in the args: keys = [ :account_id , :url ] %{ account_id : 1 , url : &quot;https://example.com&quot; } |&gt; MyApp.Worker . new ( unique : [ fields : [ :args , :worker ] , keys : keys ] ) |&gt; Oban . insert ( ) Insert a unique job considering only specified keys in the meta: unique = [ fields : [ :meta ] , keys : [ :slug ] ] %{ id : 1 } |&gt; MyApp.Worker . new ( meta : %{ slug : &quot;unique-key&quot; } , unique : unique ) |&gt; Oban . insert ( )","ref":"Oban.Job.html#new/2","title":"Oban.Job.new/2","type":"function"},{"doc":"A canonical list of all possible job states. This may be used to build up :unique options without duplicating states in application code. Examples iex&gt; Oban.Job . states ( ) -- [ :completed , :discarded ] [ :scheduled , :available , :executing , :retryable , :cancelled ]","ref":"Oban.Job.html#states/0","title":"Oban.Job.states/0","type":"function"},{"doc":"Convert a Job changeset into a map suitable for database insertion. Examples Convert a worker generated changeset into a plain map: %{ id : 123 } |&gt; MyApp.Worker . new ( ) |&gt; Oban.Job . to_map ( )","ref":"Oban.Job.html#to_map/1","title":"Oban.Job.to_map/1","type":"function"},{"doc":"","ref":"Oban.Job.html#t:args/0","title":"Oban.Job.args/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:changeset/0","title":"Oban.Job.changeset/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:changeset_fun/0","title":"Oban.Job.changeset_fun/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:changeset_list/0","title":"Oban.Job.changeset_list/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:changeset_list_fun/0","title":"Oban.Job.changeset_list_fun/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:errors/0","title":"Oban.Job.errors/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:option/0","title":"Oban.Job.option/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:replace_option/0","title":"Oban.Job.replace_option/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:schedule_in_option/0","title":"Oban.Job.schedule_in_option/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:t/0","title":"Oban.Job.t/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:tags/0","title":"Oban.Job.tags/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:unique_field/0","title":"Oban.Job.unique_field/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:unique_option/0","title":"Oban.Job.unique_option/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:unique_period/0","title":"Oban.Job.unique_period/0","type":"type"},{"doc":"","ref":"Oban.Job.html#t:unique_state/0","title":"Oban.Job.unique_state/0","type":"type"},{"doc":"Migrations create and modify the database tables Oban needs to function. Usage To use migrations in your application you'll need to generate an Ecto.Migration that wraps calls to Oban.Migrations : mix ecto.gen.migration add_oban Open the generated migration in your editor and call the up and down functions on Oban.Migrations : defmodule MyApp.Repo.Migrations.AddOban do use Ecto.Migration def up , do : Oban.Migrations . up ( ) def down , do : Oban.Migrations . down ( ) end This will run all of Oban's versioned migrations for your database. Migrations between versions are idempotent. As new versions are released you may need to run additional migrations. Now, run the migration to create the table: mix ecto.migrate Isolation with Prefixes Oban supports namespacing through PostgreSQL schemas, also called &quot;prefixes&quot; in Ecto. With prefixes your jobs table can reside outside of your primary schema (usually public) and you can have multiple separate job tables. To use a prefix you first have to specify it within your migration: defmodule MyApp.Repo.Migrations.AddPrefixedObanJobsTable do use Ecto.Migration def up , do : Oban.Migrations . up ( prefix : &quot;private&quot; ) def down , do : Oban.Migrations . down ( prefix : &quot;private&quot; ) end The migration will create the &quot;private&quot; schema and all tables, functions and triggers within that schema. With the database migrated you'll then specify the prefix in your configuration: config :my_app , Oban , prefix : &quot;private&quot; , ... In some cases, for example if your &quot;private&quot; schema already exists and your database user in production doesn't have permissions to create a new schema, trying to create the schema from the migration will result in an error. In such situations, it may be useful to inhibit the creation of the &quot;private&quot; schema: defmodule MyApp.Repo.Migrations.AddPrefixedObanJobsTable do use Ecto.Migration def up , do : Oban.Migrations . up ( prefix : &quot;private&quot; , create_schema : false ) def down , do : Oban.Migrations . down ( prefix : &quot;private&quot; ) end","ref":"Oban.Migrations.html","title":"Oban.Migrations","type":"module"},{"doc":"Run the down changes for all migrations between the current version and the initial version. Example Run all migrations from current version down to the first: Oban.Migrations . down ( ) Run migrations down to a specified version: Oban.Migrations . down ( version : 5 ) Run migrations in an alternate prefix: Oban.Migrations . down ( prefix : &quot;payments&quot; )","ref":"Oban.Migrations.html#down/1","title":"Oban.Migrations.down/1","type":"function"},{"doc":"Run the up changes for all migrations between the initial version and the current version. Example Run all migrations up to the current version: Oban.Migrations . up ( ) Run migrations up to a specified version: Oban.Migrations . up ( version : 2 ) Run migrations in an alternate prefix: Oban.Migrations . up ( prefix : &quot;payments&quot; ) Run migrations in an alternate prefix but don't try to create the schema: Oban.Migrations . up ( prefix : &quot;payments&quot; , create_schema : false )","ref":"Oban.Migrations.html#up/1","title":"Oban.Migrations.up/1","type":"function"},{"doc":"","ref":"Oban.PostgresNotifier.html","title":"Oban.PostgresNotifier","type":"module"},{"doc":"Postgres Listen/Notify based Notifier Caveats The notifications system is built on PostgreSQL's LISTEN/NOTIFY functionality. Notifications are only delivered after a transaction completes and are de-duplicated before publishing. Most applications run Ecto in sandbox mode while testing. Sandbox mode wraps each test in a separate transaction which is rolled back after the test completes. That means the transaction is never committed, which prevents delivering any notifications. To test using notifications you must run Ecto without sandbox mode enabled.","ref":"Oban.PostgresNotifier.html#child_spec/1","title":"Oban.PostgresNotifier.child_spec/1","type":"function"},{"doc":"Telemetry integration for event metrics, logging and error reporting. Initialization Events Oban emits the following telemetry event when an Oban supervisor is started: [:oban, :supervisor, :init] - when the Oban supervisor is started this will execute The initialization event contains the following measurements: :system_time - The system's time when Oban was started The initialization event contains the following metadata: :conf - The configuration used for the Oban supervisor instance :pid - The PID of the supervisor instance Job Events Oban emits the following telemetry events for each job: [:oban, :job, :start] — at the point a job is fetched from the database and will execute [:oban, :job, :stop] — after a job succeeds and the success is recorded in the database [:oban, :job, :exception] — after a job fails and the failure is recorded in the database All job events share the same details about the job that was executed. In addition, failed jobs provide the error type, the error itself, and the stacktrace. The following chart shows which metadata you can expect for each event: event measures metadata :start :system_time :job, :conf, :state :stop :duration , :queue_time :job, :conf, :state, :result :exception :duration , :queue_time :job, :conf, :state, :kind, :reason, :stacktrace Metadata :conf — the config of the Oban supervised producer :job — the executing Oban.Job :state — one of :success , :discard or :snoozed :result — the perform/1 return value, only included when the state is :success For :exception events the metadata includes details about what caused the failure. The :kind value is determined by how an error occurred. Here are the possible kinds: :error — from an {:error, error} return value. Some Erlang functions may also throw an :error tuple, which will be reported as :error . :exit — from a caught process exit :throw — from a caught value, this doesn't necessarily mean that an error occurred and the error value is unpredictable Producer Events Oban emits the following telemetry span events for each queue's producer: [:oban, :producer, :start | :stop | :exception] — when a producer dispatches new jobs event measures metadata :start :system_time :queue, :conf :stop :duration :queue, :conf, :dispatched_count :exception :duration :queue, :conf , :kind, :reason, :stacktrace Metadata :queue — the name of the queue as a string, e.g. &quot;default&quot; or &quot;mailers&quot; :conf — the config of the Oban supervised producer :dispatched_count — the number of jobs fetched and started by the producer Engine Events Oban emits telemetry span events for the following Engine operations: [:oban, :engine, :init, :start | :stop | :exception] event measures metadata :start :system_time :conf , engine :stop :duration :conf , engine :exception :duration :conf , engine [:oban, :engine, :refresh, :start | :stop | :exception] event measures metadata :start :system_time :conf , engine :stop :duration :conf , engine :exception :duration :conf , engine [:oban, :engine, :put_meta, :start | :stop | :exception] event measures metadata :start :system_time :conf , engine :stop :duration :conf , engine :exception :duration :conf , engine [:oban, :engine, :fetch_jobs, :start | :stop | :exception] event measures metadata :start :system_time :conf , engine :stop :duration :conf , engine :exception :duration :conf , engine [:oban, :engine, :complete_job, :start | :stop | :exception] event measures metadata :start :system_time :conf , engine , job :stop :duration :conf , engine , job :exception :duration :conf , engine , job , kind , reason , stacktrace [:oban, :engine, :discard_job, :start | :stop | :exception] event measures metadata :start :system_time :conf , engine , job :stop :duration :conf , engine , job :exception :duration :conf , engine , job , kind , reason , stacktrace [:oban, :engine, :error_job, :start | :stop | :exception] event measures metadata :start :system_time :conf , engine , job :stop :duration :conf , engine , job :exception :duration :conf , engine , job , kind , reason , stacktrace [:oban, :engine, :snooze_job, :start | :stop | :exception] event measures metadata :start :system_time :conf , engine , job :stop :duration :conf , engine , job :exception :duration :conf , engine , job , kind , reason , stacktrace [:oban, :engine, :cancel_job, :start | :stop | :exception] event measures metadata :start :system_time :conf , engine , job :stop :duration :conf , engine , job :exception :duration :conf , engine , job , kind , reason , stacktrace [:oban, :engine, :cancel_all_jobs, :start | :stop | :exception] event measures metadata :start :system_time :conf , engine :stop :duration :conf , engine :exception :duration :conf , engine , kind , reason , stacktrace Metadata :conf — the config of the Oban supervised producer :engine — the module of the engine used :job - the Oban.Job in question Notifier Events Oban emits telemetry a span event each time the Notifier is triggered: [:oban, :notifier, :notify, :start | :stop | :exception] event measures metadata :start :system_time :conf , channel , payload :stop :duration :conf , channel , payload :exception :duration :conf , channel , payload , kind , reason , stacktrace :conf — the config of the Oban supervised producer :channel — the channel on which the notification was sent :payload - the payload that was sent kind , reason , stacktrace , see the explanation above. Circuit Events All processes that interact with the database have circuit breakers to prevent errors from crashing the entire supervision tree. Processes emit a [:oban, :circuit, :trip] event when a circuit is tripped and [:oban, :circuit, :open] when the breaker is subsequently opened again. event measures metadata [:oban, :circuit, :trip] :kind, :reason, :message, :name, :stacktrace, :conf [:oban, :circuit, :open] :name, :conf Metadata :kind — the kind of error (see the explanation above) :reason — the error that tripped the circuit, see the error kinds breakdown above :name — the registered name of the process that tripped a circuit, i.e. Oban.Notifier :message — a formatted error message describing what went wrong :stacktrace — exception stacktrace, when available :conf — the config of the Oban supervisor that the producer is for Plugin Events All the Oban plugins emit telemetry events under the [:oban, :plugin, *] pattern (where * is either :start , :stop , or :exception ). You can filter out for plugin events by looking into the metadata of the event and checking the value of :plugin . The :plugin key will contain the module name of the plugin module that emitted the event. For example, to get Oban.Plugins.Cron specific events, you can filter for telemetry events with a metadata key/value of plugin: Oban.Plugins.Cron . Oban emits the following telemetry event whenever a plugin executes (be sure to check the documentation for each plugin as each plugin can also add additional metadata specific to the plugin): [:oban, :plugin, :start] — when the plugin beings performing its work [:oban, :plugin, :stop] — after the plugin completes its work [:oban, :plugin, :exception] — when the plugin encounters an error The following chart shows which metadata you can expect for each event: event measures metadata :start :system_time :conf, :plugin :stop :duration :conf, :plugin :exception :duration :kind, :reason, :stacktrace, :conf, :plugin Default Logger A default log handler that emits structured JSON is provided, see attach_default_logger/0 for usage. Otherwise, if you would prefer more control over logging or would like to instrument events you can write your own handler. Here is an example of the JSON output for the job:stop event: { &quot;args&quot;:{&quot;action&quot;:&quot;OK&quot;,&quot;ref&quot;:1}, &quot;duration&quot;:4327295, &quot;event&quot;:&quot;job:stop&quot;, &quot;queue&quot;:&quot;alpha&quot;, &quot;queue_time&quot;:3127905, &quot;source&quot;:&quot;oban&quot;, &quot;worker&quot;:&quot;Oban.Integration.Worker&quot; } All timing measurements are recorded as native time units but logged in microseconds. Examples A handler that only logs a few details about failed jobs: defmodule MicroLogger do require Logger def handle_event ( [ :oban , :job , :exception ] , %{ duration : duration } , meta , nil ) do Logger . warn ( &quot;[ \#{ meta . queue } ] \#{ meta . worker } failed in \#{ duration } &quot; ) end end :telemetry . attach ( &quot;oban-logger&quot; , [ :oban , :job , :exception ] , &amp; MicroLogger . handle_event / 4 , nil ) Another great use of execution data is error reporting. Here is an example of integrating with Honeybadger , but only reporting jobs that have failed 3 times or more: defmodule ErrorReporter do def handle_event ( [ :oban , :job , :exception ] , _ , %{ attempt : attempt } = meta , _ ) do if attempt &gt;= 3 do context = Map . take ( meta , [ :id , :args , :queue , :worker ] ) Honeybadger . notify ( meta . reason , context , meta . stacktrace ) end end end :telemetry . attach ( &quot;oban-errors&quot; , [ :oban , :job , :exception ] , &amp; ErrorReporter . handle_event / 4 , [ ] )","ref":"Oban.Telemetry.html","title":"Oban.Telemetry","type":"module"},{"doc":"Attaches a default structured JSON Telemetry handler for logging. This function attaches a handler that outputs logs with the following fields: args — a map of the job's raw arguments duration — the job's runtime duration, in the native time unit event — either :success or :failure depending on whether the job succeeded or errored queue — the job's queue source — always &quot;oban&quot; system_time — when the job started, in microseconds worker — the job's worker module Examples Attach a logger at the default :info level: :ok = Oban.Telemetry . attach_default_logger ( ) Attach a logger with a specific oban prefix and level: :ok = Oban.Telemetry . attach_default_logger ( telemetry_prefix : [ :my_oban ] , logger_level : :debug )","ref":"Oban.Telemetry.html#attach_default_logger/1","title":"Oban.Telemetry.attach_default_logger/1","type":"function"},{"doc":"","ref":"Oban.Telemetry.html#span/3","title":"Oban.Telemetry.span/3","type":"function"},{"doc":"","ref":"Oban.Telemetry.html#t:attach_option/0","title":"Oban.Telemetry.attach_option/0","type":"type"},{"doc":"This module simplifies testing workers and making assertions about enqueued jobs. Assertions may be made on any property of a job, but you'll typically want to check by args , queue or worker . If you're using namespacing through PostgreSQL schemas, also called &quot;prefixes&quot; in Ecto, you should use the prefix option when doing assertions about enqueued jobs during testing. By default the prefix option is public . Using in Tests The most convenient way to use Oban.Testing is to use the module: use Oban.Testing , repo : MyApp.Repo That will define the helper functions you'll use to make assertions on the jobs that should (or should not) be inserted in the database while testing. Along with the repo you can also specify an alternate prefix to use in all assertions: use Oban.Testing , repo : MyApp.Repo , prefix : &quot;business&quot; Some example assertions: # Assert that a job was already enqueued assert_enqueued worker : MyWorker , args : %{ id : 1 } # Assert that a job was enqueued or will be enqueued in the next 100ms assert_enqueued [ worker : MyWorker , args : %{ id : 1 } ] , 100 # Refute that a job was already enqueued refute_enqueued queue : &quot;special&quot; , args : %{ id : 2 } # Refute that a job was already enqueued or would be enqueued in the next 100ms refute_enqueued queue : &quot;special&quot; , args : %{ id : 2 } , 100 # Make assertions on a list of all jobs matching some options assert [ %{ args : %{ &quot;id&quot; =&gt; 1 } } ] = all_enqueued ( worker : MyWorker ) # Assert that no jobs are enqueued in any queues assert [ ] = all_enqueued ( ) Note that the final example, using all_enqueued/1 , returns a raw list of matching jobs and does not make an assertion by itself. This makes it possible to test using pattern matching at the expense of being more verbose. Example Given a simple module that enqueues a job: defmodule MyApp.Business do def work ( args ) do args |&gt; Oban.Job . new ( worker : MyApp.Worker , queue : :special ) |&gt; Oban . insert! ( ) end end The behaviour can be exercised in your test code: defmodule MyApp.BusinessTest do use ExUnit.Case , async : true use Oban.Testing , repo : MyApp.Repo alias MyApp.Business test &quot;jobs are enqueued with provided arguments&quot; do Business . work ( %{ id : 1 , message : &quot;Hello!&quot; } ) assert_enqueued worker : MyApp.Worker , args : %{ id : 1 , message : &quot;Hello!&quot; } end end Matching Scheduled Jobs and Timestamps In order to assert a job has been scheduled at a certain time, you will need to match against the scheduled_at attribute of the enqueued job. in_an_hour = DateTime . add ( DateTime . utc_now ( ) , 3600 , :second ) assert_enqueued worker : MyApp.Worker , scheduled_at : in_an_hour By default, Oban will apply a 1 second delta to all timestamp fields of jobs, so that small deviations between the actual value and the expected one are ignored. You may configure this delta by passing a tuple of value and a delta option (in seconds) to corresponding keyword: assert_enqueued worker : MyApp.Worker , scheduled_at : { in_an_hour , delta : 10 } Adding to Case Templates To include helpers in all of your tests you can add it to your case template: defmodule MyApp.DataCase do use ExUnit.CaseTemplate using do quote do use Oban.Testing , repo : MyApp.Repo import Ecto import Ecto.Changeset import Ecto.Query import MyApp.DataCase alias MyApp.Repo end end end","ref":"Oban.Testing.html","title":"Oban.Testing","type":"module"},{"doc":"Retrieve all currently enqueued jobs matching a set of options. Only jobs matching all of the provided arguments will be returned. Additionally, jobs are returned in descending order where the most recently enqueued job will be listed first. Examples Assert based on only some of a job's args: assert [ %{ args : %{ &quot;id&quot; =&gt; 1 } } ] = all_enqueued ( worker : MyWorker ) Assert that exactly one job was inserted for a queue: assert [ % Oban.Job { } ] = all_enqueued ( queue : :alpha ) Assert that there aren't any jobs enqueued for any queues or workers: assert [ ] = all_enqueued ( )","ref":"Oban.Testing.html#all_enqueued/2","title":"Oban.Testing.all_enqueued/2","type":"function"},{"doc":"Assert that a job with particular options has been enqueued. Only values for the provided arguments will be checked. For example, an assertion made on worker: &quot;MyWorker&quot; will match any jobs for that worker, regardless of the queue or args.","ref":"Oban.Testing.html#assert_enqueued/2","title":"Oban.Testing.assert_enqueued/2","type":"function"},{"doc":"Assert that a job with particular options is or will be enqueued within a timeout period. See assert_enqueued/2 for additional details. Examples Assert that a job will be enqueued in the next 100ms: assert_enqueued [ worker : MyWorker ] , 100","ref":"Oban.Testing.html#assert_enqueued/3","title":"Oban.Testing.assert_enqueued/3","type":"function"},{"doc":"Construct a job and execute it with a worker module. This reduces boiler plate when constructing jobs for unit tests and checks for common pitfalls. For example, it automatically converts args to string keys before calling perform/1 , ensuring that perform clauses aren't erroneously trying to match atom keys. The helper makes the following assertions: That the worker implements the Oban.Worker behaviour That the options provided build a valid job That the return is valid, e.g. :ok , {:ok, value} , {:error, value} etc. If all of the assertions pass then the function returns the result of perform/1 for you to make additional assertions on. Examples Successfully execute a job with some string arguments: assert :ok = perform_job ( MyWorker , %{ &quot;id&quot; =&gt; 1 } ) Successfully execute a job and assert that it returns an error tuple: assert { :error , _ } = perform_job ( MyWorker , %{ &quot;bad&quot; =&gt; &quot;arg&quot; } ) Execute a job with the args keys automatically stringified: assert :ok = perform_job ( MyWorker , %{ id : 1 } ) Exercise custom attempt handling within a worker by passing options: assert :ok = perform_job ( MyWorker , %{ } , attempt : 42 ) Cause a test failure because the provided worker isn't real: assert :ok = perform_job ( Vorker , %{ &quot;id&quot; =&gt; 1 } )","ref":"Oban.Testing.html#perform_job/3","title":"Oban.Testing.perform_job/3","type":"function"},{"doc":"Refute that a job with particular options has been enqueued. See assert_enqueued/2 for additional details.","ref":"Oban.Testing.html#refute_enqueued/2","title":"Oban.Testing.refute_enqueued/2","type":"function"},{"doc":"Refute that a job with particular options is or will be enqueued within a timeout period. The minimum refute timeout is 10ms. See assert_enqueued/2 for additional details. Examples Refute that a job will be enqueued in the next 100ms: refute_enqueued [ worker : MyWorker ] , 100","ref":"Oban.Testing.html#refute_enqueued/3","title":"Oban.Testing.refute_enqueued/3","type":"function"},{"doc":"Defines a behavior and macro to guide the creation of worker modules. Worker modules do the work of processing a job. At a minimum they must define a perform/1 function, which is called with the full Oban.Job struct. Defining Workers Worker modules are defined by using Oban.Worker . A bare use Oban.Worker invocation sets a worker with these defaults: :max_attempts — 20 :priority — 0 :queue — :default :unique — no uniqueness set The following example defines a worker module to process jobs in the events queue. It then dials down the priority from 0 to 1, limits retrying on failures to 10, adds a &quot;business&quot; tag, and ensures that duplicate jobs aren't enqueued within a 30 second period: defmodule MyApp.Workers.Business do use Oban.Worker , queue : :events , priority : 1 , max_attempts : 10 , tags : [ &quot;business&quot; ] , unique : [ period : 30 ] @impl Oban.Worker def perform ( % Oban.Job { attempt : attempt } ) when attempt &gt; 3 do IO . inspect ( attempt ) end def perform ( job ) do IO . inspect ( job . args ) end end The perform/1 function receives an Oban.Job struct as argument. This allows workers to change the behavior of perform/1 based on attributes of the job, e.g. the number of execution attempts or when it was inserted. The value returned from perform/1 can control whether the job is a success or a failure: :ok or {:ok, value} — the job is successful; for success tuples the value is ignored :discard or {:discard, reason} — discard the job and prevent it from being retried again. An error is recorded using the optional reason, though the job is still successful {:error, error} — the job failed, record the error and schedule a retry if possible {:snooze, seconds} — mark the job as snoozed and schedule it to run again seconds in the future. See Snoozing for more details. In addition to explicit return values, any unhandled exception , exit or throw will fail the job and schedule a retry if possible. As an example of error tuple handling, this worker will return an error tuple when the value is less than one: defmodule MyApp.Workers.ErrorExample do use Oban.Worker @impl Worker def perform ( %{ args : %{ &quot;value&quot; =&gt; value } } ) do if value &gt; 1 do :ok else { :error , &quot;invalid value given: &quot; &lt;&gt; inspect ( value ) } end end end The error tuple is wrapped in an Oban.PerformError with a formatted message. The error tuple itself is available through the exception's :reason field. Enqueuing Jobs All workers implement a new/2 function that converts an args map into a job changeset suitable for inserting into the database for later execution: %{ in_the : &quot;business&quot; , of_doing : &quot;business&quot; } |&gt; MyApp.Workers.Business . new ( ) |&gt; Oban . insert ( ) The worker's defaults may be overridden by passing options: %{ vote_for : &quot;none of the above&quot; } |&gt; MyApp.Workers.Business . new ( queue : &quot;special&quot; , max_attempts : 5 ) |&gt; Oban . insert ( ) Uniqueness options may also be overridden by passing options: %{ expensive : &quot;business&quot; } |&gt; MyApp.Workers.Business . new ( unique : [ period : 120 , fields : [ :worker ] ] ) |&gt; Oban . insert ( ) Note that unique options aren't merged, they are overridden entirely. See Oban.Job for all available options. Customizing Backoff When jobs fail they may be retried again in the future using a backoff algorithm. By default the backoff is exponential with a fixed padding of 15 seconds and a small amount of jitter. The jitter helps to prevent jobs that fail simultaneously from consistently retrying at the same time. The default backoff is clamped to a maximum of 24 days, the equivalent of the 20th attempt. If the default strategy is too aggressive or otherwise unsuited to your app's workload you can define a custom backoff function using the backoff/1 callback. The following worker defines a backoff/1 function that delays retries using a variant of the historic Resque/Sidekiq algorithm: defmodule MyApp.SidekiqBackoffWorker do use Oban.Worker @impl Worker def backoff ( % Job { attempt : attempt } ) do trunc ( :math . pow ( attempt , 4 ) + 15 + :rand . uniform ( 30 ) * attempt ) end @impl Worker def perform ( _job ) do :do_business end end Here are some alternative backoff strategies to consider: constant — delay by a fixed number of seconds, e.g. 1→15, 2→15, 3→15 linear — delay for the same number of seconds as the current attempt, e.g. 1→1, 2→2, 3→3 squared — delay by attempt number squared, e.g. 1→1, 2→4, 3→9 sidekiq — delay by a base amount plus some jitter, e.g. 1→32, 2→61, 3→135 Contextual Backoff Any error, catch or throw is temporarily recorded in the job's unsaved_error map. The unsaved error map can be used by backoff/1 to calculate a custom backoff based on the exact error that failed the job. In this example the backoff/1 callback checks to see if the error was due to rate limiting and adjusts the backoff accordingly: defmodule MyApp.ApiWorker do use Oban.Worker @five_minutes 5 * 60 @impl Worker def perform ( %{ args : args } ) do MyApp . make_external_api_call ( args ) end @impl Worker def backoff ( % Job { attempt : attempt , unsaved_error : unsaved_error } ) do %{ kind : _ , reason : reason , stacktrace : _ } = unsaved_error case reason do % MyApp.ApiError { status : 429 } -&gt; @five_minutes _ -&gt; trunc ( :math . pow ( attempt , 4 ) ) end end end Snoozing jobs When returning {:snooze, snooze_time} in perform/1 , the job is postponed for at least snooze_time seconds. Snoozing is done by incrementing the job's max_attempts field and scheduling execution for snooze_time seconds in the future. Snoozing does not change the number of retries remaining on the job, but it does increment the attempt number each time the job snoozes, which will affect the default backoff exponential retry algorithm. In the example below the backoff/1 callback compensates for snoozing: defmodule MyApp.SnoozingWorker do @max_attempts 20 use Oban.Worker , max_attempts : @max_attempts @impl Worker def backoff ( % Job { } = job ) do corrected_attempt = @max_attempts - ( job . max_attempts - job . attempt ) Worker . backoff ( %{ job | attempt : corrected_attempt } ) end @impl Worker def perform ( job ) do if MyApp . something? ( job ) , do : :ok , else : { :snooze , 60 } end end Limiting Execution Time By default, individual jobs may execute indefinitely. If this is undesirable you may define a timeout in milliseconds with the timeout/1 callback on your worker module. For example, to limit a worker's execution time to 30 seconds: def MyApp.Worker do use Oban.Worker @impl Oban.Worker def perform ( _job ) do something_that_may_take_a_long_time ( ) :ok end @impl Oban.Worker def timeout ( _job ) , do : :timer . seconds ( 30 ) end The timeout/1 function accepts an Oban.Job struct, so you can customize the timeout using any job attributes. Define the timeout value through job args: def timeout ( % _ { args : %{ &quot;timeout&quot; =&gt; timeout } } ) , do : timeout Define the timeout based on the number of attempts: def timeout ( % _ { attempt : attempt } ) , do : attempt * :timer . seconds ( 5 )","ref":"Oban.Worker.html","title":"Oban.Worker","type":"behaviour"},{"doc":"Calculate the execution backoff. In this context backoff specifies the number of seconds to wait before retrying a failed job. Defaults to an exponential algorithm with a minimum delay of 15 seconds and a small amount of jitter.","ref":"Oban.Worker.html#c:backoff/1","title":"Oban.Worker.backoff/1","type":"callback"},{"doc":"Resolve a module from a worker string. Examples iex&gt; Oban.Worker . from_string ( &quot;Oban.Integration.Worker&quot; ) { :ok , Oban.Integration.Worker } iex&gt; defmodule NotAWorker , do : [ ] ...&gt; Oban.Worker . from_string ( &quot;NotAWorker&quot; ) { :error , % RuntimeError { message : &quot;module is not a worker: NotAWorker&quot; } } iex&gt; Oban.Worker . from_string ( &quot;RandomWorker&quot; ) { :error , % RuntimeError { message : &quot;unknown worker: RandomWorker&quot; } }","ref":"Oban.Worker.html#from_string/1","title":"Oban.Worker.from_string/1","type":"function"},{"doc":"Build a job changeset for this worker with optional overrides. See Oban.Job.new/2 for the available options.","ref":"Oban.Worker.html#c:new/2","title":"Oban.Worker.new/2","type":"callback"},{"doc":"The perform/1 function is called to execute a job. Each perform/1 function should return :ok or a success tuple. When the return is an error tuple, an uncaught exception or a throw then the error is recorded and the job may be retried if there are any attempts remaining. Note that the args map provided to perform/1 will always have string keys, regardless of the key type when the job was enqueued. The args are stored as jsonb in PostgreSQL and the serialization process automatically stringifies all keys.","ref":"Oban.Worker.html#c:perform/1","title":"Oban.Worker.perform/1","type":"callback"},{"doc":"Set a job's maximum execution time in milliseconds. Jobs that exceed the time limit are considered a failure and may be retried. Defaults to :infinity .","ref":"Oban.Worker.html#c:timeout/1","title":"Oban.Worker.timeout/1","type":"callback"},{"doc":"Return a string representation of a worker module. This is particularly useful for normalizing worker names when building custom Ecto queries. Examples iex&gt; Oban.Worker . to_string ( MyApp.SomeWorker ) &quot;MyApp.SomeWorker&quot; iex&gt; Oban.Worker . to_string ( Elixir.MyApp.SomeWorker ) &quot;MyApp.SomeWorker&quot; iex&gt; Oban.Worker . to_string ( &quot;Elixir.MyApp.SomeWorker&quot; ) &quot;MyApp.SomeWorker&quot;","ref":"Oban.Worker.html#to_string/1","title":"Oban.Worker.to_string/1","type":"function"},{"doc":"","ref":"Oban.Worker.html#t:result/0","title":"Oban.Worker.result/0","type":"type"},{"doc":"","ref":"Oban.Worker.html#t:t/0","title":"Oban.Worker.t/0","type":"type"},{"doc":"Periodically enqueue jobs through CRON based scheduling. Using the Plugin Schedule various jobs using {expr, worker} and {expr, worker, opts} syntaxes: config :my_app , Oban , plugins : [ { Oban.Plugins.Cron , crontab : [ { &quot;* * * * *&quot; , MyApp.MinuteWorker } , { &quot;0 * * * *&quot; , MyApp.HourlyWorker , args : %{ custom : &quot;arg&quot; } } , { &quot;0 0 * * *&quot; , MyApp.DailyWorker , max_attempts : 1 } , { &quot;0 12 * * MON&quot; , MyApp.MondayWorker , queue : :scheduled , tags : [ &quot;mondays&quot; ] } , { &quot;@daily&quot; , MyApp.AnotherDailyWorker } ] } ] Options :crontab — a list of cron expressions that enqueue jobs on a periodic basis. See Periodic Jobs in the Oban module docs for syntax and details. :timezone — which timezone to use when scheduling cron jobs. To use a timezone other than the default of &quot;Etc/UTC&quot; you must have a timezone database like tzdata installed and configured. Instrumenting with Telemetry The Oban.Plugins.Cron plugin adds the following metadata to the [:oban, :plugin, :stop] event: :jobs - a list of jobs that were inserted into the database","ref":"Oban.Plugins.Cron.html","title":"Oban.Plugins.Cron","type":"module"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Oban.Plugins.Cron.html#child_spec/1","title":"Oban.Plugins.Cron.child_spec/1","type":"function"},{"doc":"","ref":"Oban.Plugins.Cron.html#start_link/1","title":"Oban.Plugins.Cron.start_link/1","type":"function"},{"doc":"","ref":"Oban.Plugins.Cron.html#t:cron_input/0","title":"Oban.Plugins.Cron.cron_input/0","type":"type"},{"doc":"","ref":"Oban.Plugins.Cron.html#t:option/0","title":"Oban.Plugins.Cron.option/0","type":"type"},{"doc":"The Gossip plugin uses PubSub to periodically exchange queue state information between all interested nodes. This allows Oban instances to broadcast state information regardless of which engine they are using, and without storing anything in the database. Gossip enables real-time updates across an entire cluster, and is essential to the operation of UIs like Oban Web. The Gossip plugin entirely replaced heartbeats and the legacy oban_beats table. Using the Plugin The following example demonstrates using the plugin without any configuration, which will broadcast the state of each local queue every 1 second: config :my_app , Oban , plugins : [ Oban.Plugins.Gossip ] , ... Override the default options to broadcast every 5 seconds: config :my_app , Oban , plugins : [ { Oban.Plugins.Gossip , interval : :timer . seconds ( 5 ) } ] , ... Options :interval — the number of milliseconds between gossip broadcasts Instrumenting with Telemetry The Oban.Plugins.Gossip plugin adds the following metadata to the [:oban, :plugin, :stop] event: :gossip_count - the number of queues that had activity broadcasted","ref":"Oban.Plugins.Gossip.html","title":"Oban.Plugins.Gossip","type":"module"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Oban.Plugins.Gossip.html#child_spec/1","title":"Oban.Plugins.Gossip.child_spec/1","type":"function"},{"doc":"","ref":"Oban.Plugins.Gossip.html#t:option/0","title":"Oban.Plugins.Gossip.option/0","type":"type"},{"doc":"Periodically delete completed, cancelled and discarded jobs based on age. Using the Plugin The following example demonstrates using the plugin without any configuration, which will prune jobs older than the default of 60 seconds: config :my_app , Oban , plugins : [ Oban.Plugins.Pruner ] , ... Override the default options to prune jobs after 5 minutes: config :my_app , Oban , plugins : [ { Oban.Plugins.Pruner , max_age : 300 } ] , ... Options :max_age — the number of seconds after which a job may be pruned :limit — the maximum number of jobs to prune at one time. The default is 10,000 to prevent request timeouts. Applications that steadily generate more than 10k jobs a minute should increase this value. Instrumenting with Telemetry The Oban.Plugins.Pruner plugin adds the following metadata to the [:oban, :plugin, :stop] event: :pruned_count - the number of jobs that were pruned from the database","ref":"Oban.Plugins.Pruner.html","title":"Oban.Plugins.Pruner","type":"module"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Oban.Plugins.Pruner.html#child_spec/1","title":"Oban.Plugins.Pruner.child_spec/1","type":"function"},{"doc":"","ref":"Oban.Plugins.Pruner.html#t:option/0","title":"Oban.Plugins.Pruner.option/0","type":"type"},{"doc":"Repeatedly send inserted messages to all registered producers to simulate polling. This plugin is only necessary if you're running Oban in an environment where Postgres notifications don't work, notably one of: Using a database connection pooler in transaction mode, i.e. pg_bouncer. Integration testing within the Ecto sandbox, i.e. developing Oban plugins Options :interval — the number of milliseconds between notifications. The default is 1_000ms .","ref":"Oban.Plugins.Repeater.html","title":"Oban.Plugins.Repeater","type":"module"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Oban.Plugins.Repeater.html#child_spec/1","title":"Oban.Plugins.Repeater.child_spec/1","type":"function"},{"doc":"","ref":"Oban.Plugins.Repeater.html#t:option/0","title":"Oban.Plugins.Repeater.option/0","type":"type"},{"doc":"Transition jobs to the available state when they reach their scheduled time. This module is necessary for the execution of scheduled and retryable jobs. Options :interval - the number of milliseconds between database updates. This is directly tied to the resolution of scheduled jobs. For example, with an interval of 5_000ms , scheduled jobs are checked every 5 seconds. The default is 1_000ms . :limit — the number of jobs that will be staged each time the plugin runs. Defaults to 5,000 , which you can increase if staging can't keep up with your insertion rate or decrease if you're experiencing staging timeouts. Instrumenting with Telemetry The Oban.Plugins.Stager plugin adds the following metadata to the [:oban, :plugin, :stop] event: :staged_count - the number of jobs that were staged in the database","ref":"Oban.Plugins.Stager.html","title":"Oban.Plugins.Stager","type":"module"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"Oban.Plugins.Stager.html#child_spec/1","title":"Oban.Plugins.Stager.child_spec/1","type":"function"},{"doc":"","ref":"Oban.Plugins.Stager.html#t:option/0","title":"Oban.Plugins.Stager.option/0","type":"type"},{"doc":"The Config struct validates and encapsulates Oban instance state. Options passed to Oban.start_link/1 are validated and stored in a config struct. Internal modules and plugins are always passed the config with a :conf key.","ref":"Oban.Config.html","title":"Oban.Config","type":"module"},{"doc":"","ref":"Oban.Config.html#t:option/0","title":"Oban.Config.option/0","type":"type"},{"doc":"","ref":"Oban.Config.html#t:t/0","title":"Oban.Config.t/0","type":"type"},{"doc":"The Notifier coordinates listening for and publishing notifications for events in predefined channels. Every Oban supervision tree contains a notifier process, registered as Oban.Notifier , which can be any implementation of the Oban.Notifier behaviour. The default one is Oban.PostgresNotifier , which relies on Postgres LISTEN/NOTIFY . All incoming notifications are relayed through the notifier to other processes. Channels The notifier recognizes three predefined channels, each with a distinct responsibility: gossip — arbitrary communication between nodes or jobs are sent on the gossip channel insert — as jobs are inserted into the database an event is published on the insert channel. Processes such as queue producers use this as a signal to dispatch new jobs. signal — instructions to take action, such as scale a queue or kill a running job, are sent through the signal channel. The insert and signal channels are primarily for internal use. Use the gossip channel to send notifications between jobs or processes in your application. Examples Broadcasting after a job is completed: defmodule MyApp.Worker do use Oban.Worker @impl Oban.Worker def perform ( job ) do :ok = MyApp . do_work ( job . args ) Oban.Notifier . notify ( Oban , :gossip , %{ complete : job . id } ) :ok end end Listening for job complete events from another process: def insert_and_listen ( args ) do { :ok , job } = args |&gt; MyApp.Worker . new ( ) |&gt; Oban . insert ( ) receive do { :notification , :gossip , %{ &quot;complete&quot; =&gt; ^ job . id } } -&gt; IO . puts ( &quot;Other job complete!&quot; ) after 30_000 -&gt; IO . puts ( &quot;Other job didn&#39;t finish in 30 seconds!&quot; ) end end","ref":"Oban.Notifier.html","title":"Oban.Notifier","type":"behaviour"},{"doc":"Register the current process to receive relayed messages for the provided channels. All messages are received as JSON and decoded before they are relayed to registered processes. Each registered process receives a three element notification tuple in the following format: { :notification , channel :: channel ( ) , decoded :: map ( ) } Example Register to listen for all :gossip channel messages: Oban.Notifier . listen ( [ :gossip ] ) Listen for messages on all channels: Oban.Notifier . listen ( [ :gossip , :insert , :signal ] ) Listen for messages when using a custom Oban name: Oban.Notifier . listen ( MyApp.MyOban , [ :gossip , :insert , :signal ] )","ref":"Oban.Notifier.html#listen/2","title":"Oban.Notifier.listen/2","type":"function"},{"doc":"Register current process to receive messages from some channels","ref":"Oban.Notifier.html#c:listen/2","title":"Oban.Notifier.listen/2","type":"callback"},{"doc":"Broadcast a notification to listeners on all nodes. Notifications are scoped to the configured prefix . For example, if there are instances running with the public and private prefixes, a notification published in the public prefix won't be picked up by processes listening with the private prefix. Using notify/3 with a config is soft deprecated. Use a server as the first argument instead Example Broadcast a gossip message: Oban.Notifier . notify ( :gossip , %{ message : &quot;hi!&quot; } )","ref":"Oban.Notifier.html#notify/3","title":"Oban.Notifier.notify/3","type":"function"},{"doc":"Broadcast a notification in a channel","ref":"Oban.Notifier.html#c:notify/3","title":"Oban.Notifier.notify/3","type":"callback"},{"doc":"Starts a notifier","ref":"Oban.Notifier.html#c:start_link/1","title":"Oban.Notifier.start_link/1","type":"callback"},{"doc":"Unregister the current process from receiving relayed messages on provided channels. Example Stop listening for messages on the :gossip channel: Oban.Notifier . unlisten ( [ :gossip ] ) Stop listening for messages when using a custom Oban name: Oban.Notifier . unlisten ( MyApp.MyOban , [ :gossip ] )","ref":"Oban.Notifier.html#unlisten/2","title":"Oban.Notifier.unlisten/2","type":"function"},{"doc":"Unregister current process from channels","ref":"Oban.Notifier.html#c:unlisten/2","title":"Oban.Notifier.unlisten/2","type":"callback"},{"doc":"","ref":"Oban.Notifier.html#t:channel/0","title":"Oban.Notifier.channel/0","type":"type"},{"doc":"","ref":"Oban.Notifier.html#t:option/0","title":"Oban.Notifier.option/0","type":"type"},{"doc":"","ref":"Oban.Notifier.html#t:server/0","title":"Oban.Notifier.server/0","type":"type"},{"doc":"Local process storage for Oban instances.","ref":"Oban.Registry.html","title":"Oban.Registry","type":"module"},{"doc":"Fetch the config for an Oban supervisor instance. Example Get the default instance config: Oban.Registry . config ( Oban ) Get config for a custom named instance: Oban.Registry . config ( MyApp.Oban )","ref":"Oban.Registry.html#config/1","title":"Oban.Registry.config/1","type":"function"},{"doc":"Build a via tuple suitable for calls to a supervised Oban process. Example For an Oban supervisor: Oban.Registry . via ( Oban ) For a supervised module: Oban.Registry . via ( Oban , Oban.Notifier ) For a plugin: Oban.Registry . via ( Oban , { :plugin , Oban.Plugins.Cron } )","ref":"Oban.Registry.html#via/3","title":"Oban.Registry.via/3","type":"function"},{"doc":"Returns the pid of a supervised Oban process, or nil if the process can't be found. Example Get the Oban supervisor's pid: Oban.Registry . whereis ( Oban ) Get a supervised module's pid: Oban.Registry . whereis ( Oban , Oban.Notifier ) Get the pid for a plugin: Oban.Registry . whereis ( Oban , { :plugin , MyApp.Oban.Plugin } ) Get the pid for a queue's producer: Oban.Registry . whereis ( Oban , { :producer , :default } )","ref":"Oban.Registry.html#whereis/2","title":"Oban.Registry.whereis/2","type":"function"},{"doc":"","ref":"Oban.Registry.html#t:key/0","title":"Oban.Registry.key/0","type":"type"},{"doc":"","ref":"Oban.Registry.html#t:role/0","title":"Oban.Registry.role/0","type":"type"},{"doc":"","ref":"Oban.Registry.html#t:value/0","title":"Oban.Registry.value/0","type":"type"},{"doc":"Wrappers around Ecto.Repo callbacks. These functions should be used when working with an Ecto repo inside a plugin. These functions will resolve the correct repo instance, and set the schema prefix and the log level, according to the Oban configuration.","ref":"Oban.Repo.html","title":"Oban.Repo","type":"module"},{"doc":"Wraps Ecto.Repo.all/2 .","ref":"Oban.Repo.html#all/3","title":"Oban.Repo.all/3","type":"function"},{"doc":"Wraps Ecto.Repo.checkout/2 .","ref":"Oban.Repo.html#checkout/3","title":"Oban.Repo.checkout/3","type":"function"},{"doc":"Wraps Ecto.Repo.config/0 .","ref":"Oban.Repo.html#config/1","title":"Oban.Repo.config/1","type":"function"},{"doc":"Wraps Ecto.Repo.delete/2 .","ref":"Oban.Repo.html#delete/3","title":"Oban.Repo.delete/3","type":"function"},{"doc":"Wraps Ecto.Repo.delete_all/2 .","ref":"Oban.Repo.html#delete_all/3","title":"Oban.Repo.delete_all/3","type":"function"},{"doc":"Wraps Ecto.Repo.insert/2 .","ref":"Oban.Repo.html#insert/3","title":"Oban.Repo.insert/3","type":"function"},{"doc":"Wraps Ecto.Repo.insert_all/3 .","ref":"Oban.Repo.html#insert_all/4","title":"Oban.Repo.insert_all/4","type":"function"},{"doc":"Wraps Ecto.Repo.one/2 .","ref":"Oban.Repo.html#one/3","title":"Oban.Repo.one/3","type":"function"},{"doc":"Wraps Ecto.Adapters.SQL.Repo.query/4 .","ref":"Oban.Repo.html#query/4","title":"Oban.Repo.query/4","type":"function"},{"doc":"Wraps Ecto.Repo.stream/2","ref":"Oban.Repo.html#stream/3","title":"Oban.Repo.stream/3","type":"function"},{"doc":"Wraps Ecto.Adapters.SQL.Repo.to_sql/2 .","ref":"Oban.Repo.html#to_sql/3","title":"Oban.Repo.to_sql/3","type":"function"},{"doc":"Wraps Ecto.Repo.transaction/2 .","ref":"Oban.Repo.html#transaction/3","title":"Oban.Repo.transaction/3","type":"function"},{"doc":"Wraps Ecto.Repo.update/2 .","ref":"Oban.Repo.html#update/3","title":"Oban.Repo.update/3","type":"function"},{"doc":"Wraps Ecto.Repo.update_all/3 .","ref":"Oban.Repo.html#update_all/4","title":"Oban.Repo.update_all/4","type":"function"},{"doc":"Wraps unhandled exits and throws that occur during job execution.","ref":"Oban.CrashError.html","title":"Oban.CrashError","type":"exception"},{"doc":"Wraps the reason returned by {:error, reason} , {:discard, reason} in a proper exception. The original return tuple is available in the :reason key.","ref":"Oban.PerformError.html","title":"Oban.PerformError","type":"exception"},{"doc":"Returned when a job is terminated early due to a custom timeout.","ref":"Oban.TimeoutError.html","title":"Oban.TimeoutError","type":"exception"},{"doc":"All notable changes to Oban are documented here. 🌟 Looking for changes to Web or Pro? Check the Oban.Pro Changelog or the Oban.Web Changelog . 🌟","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"Added [Oban.Telemetry] Add customizable prefix for all telemetry events. For example, a telemetry prefix of [:my_app, :oban] would span job start telemetry events as [:my_app, :oban, :job, :start] . The default is [:oban] , which matches the existing functionality. Fixed [Oban.Plugins.Stager] Use the notifier to broadcast inserted and available jobs rather than inlining them into a Postgres query. With this change the notifier is entirely swappable and there isn't any reason to use the Repeater plugin in production . [Oban.Plugins.Cron] Validate job options on init. Providing invalid job args in the cron tab, e.g. priority: 5 or unique: [] , wasn't caught until runtime. At that point each insert attempt would fail, crashing the plugin. [Oban.Queue.Producer] Prevent crashing on exception formatting when a job exits without a stacktrace, most notably with {:EXIT, pid} . [Oban.Testing] Return invalid results from perform_job , rather than always returning nil . [Oban] Validate that a queue exists when controlling or checking locally, e.g. calls to Oban.check_queue or Oban.scale_queue . [Oban.Telemetry] Use module capture for telemetry logging to prevent warnings.","ref":"changelog.html#2-10-0-2021-11-09","title":"Changelog - 2.10.0 — 2021-11-09","type":"extras"},{"doc":"[Oban] Loosen telemetry requirement to allow either 0.4 or 1.0 without forcing apps to use an override.","ref":"changelog.html#2-9-2-2021-09-27","title":"Changelog - 2.9.2 — 2021-09-27","type":"extras"},{"doc":"Fixed [Oban] Correctly handle prefix in cancel_job and cancel_all_jobs . [Oban] Safely guard against empty changeset lits passed to insert_all/2,4 .","ref":"changelog.html#2-9-1-2021-09-23","title":"Changelog - 2.9.1 — 2021-09-23","type":"extras"},{"doc":"Optionally Use Meta for Unique Jobs It's now possible to use the meta field for unique jobs. Unique jobs have always supported worker , queue , and args fields. That was flexible, but forced applications to put ad-hoc unique values in args when they should really be in meta . The meta field supports keys , just like args . That makes it possible to use highly efficient fingerprint style uniqueness (and possibly drop the index on args , if desired). Here's an example of using a single &quot;fingerprint&quot; key in meta for uniqueness: defmodule MyApp.FingerprintWorker do use Oban.Worker , unique : [ fields : [ :worker , :meta ] , keys : [ :fingerprint ] ] @impl Worker def new ( args , opts ) do fingerprint = :erlang . phash2 ( args ) super ( args , Keyword . put ( opts , :meta , %{ fingerprint : fingerprint } ) ) end end For backward compatiblity meta isn't included in unique fields by default. Expanded Start and Scale Options After extensive refactoring to queue option management and validation, now it's possible to start and scale queues with all supported options. Previously start/stop functions only supported the limit option for dynamic scaling, reducing runtime flexibility considerably. Now it's possible to start a queue in the paused state: Oban . start_queue ( queue : :dynamic , paused : true ) Even better, for apps that use an alternative engine like the SmartEngine from Oban Pro, it's possible to start a dynamic queue with options like global concurrency or rate limiting: Oban . start_queue ( queue : :dynamic , local_limit : 10 , global_limit : 50 ) All options are also passed through scale_queue , locally or globally, even allowing you to reconfigure a feature like rate limiting at runtime: Oban . scale_queue ( queue : :dynamic , rate_limit : [ allowed : 50 , period : 60 ] ) Added [Oban] Add Oban.cancel_all_jobs/1,2 to cancel multiple jobs at once, within an atomic transaction. The function accepts a Job query for complete control over which jobs are cancelled. [Oban] Add Oban.retry_all_jobs/1,2 to retry multiple jobs at once, within an atomic transaction. Like cancel_all_jobs , it accepts a query for fine-grained control. [Oban] Add with_limit option to drain_queue/2 , which controls the number of jobs that are fetched and executed concurrently. When paired with with_recursion this can drastically speed up interdependent job draining, i.e. workflows. [Oban.Telemetry] Add telemetry span events for all engine and notifier actions. Now all database operations are covered by spans. [Oban.Migrations] Add create_schema option to prevent automatic schema creation in migrations. Changed [Oban] Consistently include a :snoozed count in drain_queue/2 output. Previously the count was only included when there was at least one snoozed job. [Oban.Testing] Default to attempt: 1 for perform_job/3 , as a worker's perform/1 would never be called with attempt: 0 . Fixed [Oban.Queue.Supervisor] Change supervisor strategy to :one_for_all . Queue supervisors used a :rest_for_one strategy, which allowed the task supervisor to keep running when a producer crashed. That allowed duplicate long-lived jobs to run simultaneously, which is a bug in itself, but could also cause attempt &gt; max_attempts violations. [Oban.Plugins.Cron] Start step ranges from the minimum value, rather than for the entire set. Now the range 8-23/4 correctly includes [8, 12, 16, 20] . [Oban.Plugins.Cron] Correcly parse step ranges with a single value, e.g. 0 1/2 * * * [Oban.Telemetry] Comply with :telemetry.span/3 by exposing errors as reason in metadata","ref":"changelog.html#2-9-0-2021-09-15","title":"Changelog - 2.9.0 — 2021-09-15","type":"extras"},{"doc":"Time Unit Scheduling It's now possible to specify a unit for :schedule_in , rather than always assuming seconds. This makes it possible to schedule a job using clearer minutes, hours, days, or weeks, e.g. schedule_in: {1, :minute} or schedule_in: {3, :days} . Changed [Oban.Testing] Accept non-map args to perform_job/3 for compatibility with overridden Worker.new/2 callbacks. [Oban.Queue.Producer] Include some jitter when scheduling queue refreshes to prevent queues from refreshing simultaneously. In extreme cases, refresh contention could cause producers to crash. Fixed [Oban.Queue.Executor] Restore logged warnings for unexpected job results by retaining the safe flag during normal execution. [Oban.Plugins.Gossip] Catch and discard unexpected messages rather than crashing the plugin. [Oban.Testing] Eliminate dialyzer warnings by including repo option in the Oban.Testing.perform_job/3 spec.","ref":"changelog.html#2-8-0-2021-07-30","title":"Changelog - 2.8.0 — 2021-07-30","type":"extras"},{"doc":"Fixed [Oban.Plugins.Pruner] Consider cancelled_at or discarded_at timestamps when querying prunable jobs. The previous query required an attempted_at value, even for cancelled or discarded jobs. If a job was cancelled before it was attempted then it wouldn't ever be pruned. [Oban.Plugins.Gossip] Correct exit handling during safe checks. Occasionally, producer checks time out and the previous catch block didn't handle exits properly.","ref":"changelog.html#2-7-2-2021-06-10","title":"Changelog - 2.7.2 — 2021-06-10","type":"extras"},{"doc":"Fixed [Oban.Telemetry] Restore the :job key to [:oban, :job, :start] events. It was mistakenly removed during a refactoring.","ref":"changelog.html#2-7-1-2021-05-26","title":"Changelog - 2.7.1 — 2021-05-26","type":"extras"},{"doc":"Pluggable Notifier The PubSub functionality Oban relies on for starting, stopping, scaling, and pausing queues is now pluggable. The previous notifier was hard-coded to use Postgres for PubSub via LISTEN/NOTIFY . Unfortunately, LISTEN/NOTIFY doesn't work for PG Bouncer in &quot;Transaction Mode.&quot; While relatively few users run in &quot;Transaction Mode,&quot; it is increasingly common and deserved a work-around. Oban.Notifier defines a minimal behaviour and Oban ships with a default implementation, Oban.PostgresNotifier , based on the old LISTEN/NOTIFY module. You can specify a different notifier in your Oban configuration: config :my_app , Oban , notifier : MyApp.CustomNotifier , repo : MyApp.Repo , ... An alternate pg/pg2 based notifier is available in Oban Pro. Replacing Opts on Unique Conflict The replace_args option for updating args on unique conflict is replaced with a highly flexible replace option. Using replace , you can update any job field when there is a conflict. Replace works for any of the following fields: args , max_attempts , meta , priority , queue , scheduled_at , tags , worker . For example, to change the priority to 0 and increase max_attempts to 5 when there is a conflict: BusinessWorker . new ( args , max_attempts : 5 , priority : 0 , replace : [ :max_attempts , :priority ] ) Another example is bumping the scheduled time to 1 second in the future on conflict. Either scheduled_at or schedule_in values will work, but the replace option is always scheduled_at . UrgentWorker . new ( args , schedule_in : 1 , replace : [ :scheduled_at ] ) Added [Oban.Job] A new conflict? field indicates whether a unique job conflicted with an existing job on insert. [Oban.Telemetry] Always populate the unsaved_error field for error or discard events. Previously, the field was empty for discard events. [Oban.Queue.Engine] Define a cancel_job/2 callback in the engine and move cancellation from the query module to the BasicEngine . Changed [Oban.Testing] Thanks to a slight refactoring, the perform_job/3 helper now emits the same telemetry events that you'd have in production. The refactoring also exposed and fixed an inconsistency around invalid snooze handlers. [Oban.Testing] Expose Testing.all_enqueued/0 , an alternate clause that doesn't require any options. This new clause makes it possible to check for any enqueued jobs without filters. [Oban.Plugins.Stager] Limit the number of jobs staged at one time to prevent timeout errors while staging a massive number of jobs. Fixed [Oban.Repo] Respect ongoing transactions when using get_dynamic_repo . Prior to this fix, calls that use Oban.Repo , such as Oban.insert/2 , could use a different repo than the one used in the current transaction.","ref":"changelog.html#2-7-0-2021-05-25","title":"Changelog - 2.7.0 — 2021-05-25","type":"extras"},{"doc":"Fixed [Oban.Drainer] Always use the BasicEngine for draining, regardless of the currently configured engine.","ref":"changelog.html#2-6-1-2021-04-02","title":"Changelog - 2.6.1 — 2021-04-02","type":"extras"},{"doc":"🌟 Web and Pro users should check the v2.6 upgrade guide for a complete walkthrough. Pluggable Queue Engines Queue producers now use an &quot;engine&quot; callback module to manage demand and work with the database. The engine provides a clean way to expand and enhance the functionality of queue producers without modifying the solid foundation of queue supervision trees. Engines make enhanced functionality such as global concurrency, local rate limiting and global rate limiting possible! The BasicEngine is the default, and it retains the exact functionality of previous Oban versions. To specify the engine explicitly, or swap in an alternate engine, set it in your configuration: config :my_app , Oban , engine : Oban.Queue.BasicEngine , ... See the v2.6 upgrade guide for instructions on swapping in an alternate engine. Recursive Queue Draining The ever-handy Oban.drain_queue/1,2 function gained a new with_recursion option, which makes it possible to test jobs that insert more jobs when they execute. When with_recursion is enabled the queue will keep executing until no jobs are available. It even composes with with_scheduled ! In practice, this is especially useful for testing dependent workflows. Gossip Plugin for Queue Monitoring The new gossip plugin uses PubSub to efficiently exchange queue state information between all interested nodes. This allows Oban instances to broadcast state information regardless of which engine they are using, and without storing anything in the database. See the v2.6 upgrade guide for details on switching to the gossip plugin. Added [Oban.Job] Inject the current conf into the jobs struct as virtual field, making the complete conf available within perform/1 . [Oban.Notifier] Add unlisten/2 , used to stop a process from receiving notifications for one or more channels. Changed [Oban.Telemetry] Stop emitting circuit events for queue producers. As producers no longer poll, the circuit breaker masked real errors more than it guarded against sporatic issues. [Oban.Telemetry] Delay [:oban, :supervisor, :init] event until the complete supervision tree finishes initialization. [Oban.Migration] Stop creating an oban_beats table entirely. Fixed [Oban.Plugins.Cron] Prevent schedule overflow right before midnight","ref":"changelog.html#2-6-0-2021-04-02","title":"Changelog - 2.6.0 — 2021-04-02","type":"extras"},{"doc":"Top of the Minute Cron Rather than scheduling cron evaluation 60 seconds after the server starts, evaluation is now scheduled at the top of the next minute. This yields several improvements: More predictable timing for cron jobs now that they are inserted at the top of the minute. Note that jobs may execute later depending on how busy queues are. Minimize contention between servers inserting jobs, thanks to the transaction lock acquired by each plugin. Prevent duplicate inserts for jobs that omit the completed state (when server start time is staggered the transaction lock has no effect) Repeater Plugin for Transaction Pooled Databases Environments that can't make use of PG notifications, i.e. because they use PgBouncer with transaction pooling, won't process available jobs reliably. The new Repeater plugin provides a workaround that simulates polling functionality for producers. Simply include the Repeater in your plugin list: config :my_app , Oban , plugins : [ Oban.Plugins.Pruner , Oban.Plugins.Stager , Oban.Plugins.Repeater ] , ... Include Perform Result in Job Telemetry Metadata The metadata for [:oban, :job, :stop] telemetry events now include the job's perform/1 return value. That makes it possible to extract job output from other processes: def handle_event ( [ :oban , :job , :stop ] , _timing , meta , _opts ) do IO . inspect ( meta . result , label : &quot;return from \#{ meta . job . worker } &quot; ) :ok end Added [Oban] Support a changeset function in addition to a changeset in insert_all/4 . Inserting jobs within a multi is even more powerful. [Oban.Queue.Executor] Stash a timed job's pid to enable inner process messaging, notably via telemetry events. Changed [Oban] Upgrade minimum Elixir version from 1.8 to 1.9. [Oban.Plugins.Cron] Individually validate crontab workers and options, providing more descriptive errors at the collection and entry level. [Oban] Simplify the job cancelling flow for consistency. Due to race conditions it was always possible for a job to complete before it was cancelled, now that flow is much simpler. [Oban.Queue.Producer] Replace dispatch cooldown with a simpler debounce mechanism. [Oban.Plugins.Stager] Limit the number of notify events to one per queue, rather than one per available job. Fixed [Oban.Queue.Producer] Handle unclean exits without an error and stack, which prevents &quot;zombie&quot; jobs. This would occur after multiple hot upgrades when the worker module changed, as the VM would forcibly terminate any processes running the old modules. [Oban] Specify that the changeset_wrapper type allows keys other than :changesets , fixing a dialyzer type mismatch. Removed [Oban] Remove the undocumented version/0 function in favor of using the standard Application.spec/2","ref":"changelog.html#2-5-0-2021-02-26","title":"Changelog - 2.5.0 — 2021-02-26","type":"extras"},{"doc":"Fixed [Oban.Telemetry] Use conf rather than config in meta for :job and :circuit telemetry events.","ref":"changelog.html#2-4-3-2021-02-07","title":"Changelog - 2.4.3 — 2021-02-07","type":"extras"},{"doc":"Fixed [Oban.Plugins.Stager] Notify queues of all available jobs, not only jobs that were most recently staged. In some circumstances, such as reboot or retries, jobs are available without previously being scheduled.","ref":"changelog.html#2-4-2-2021-01-28","title":"Changelog - 2.4.2 — 2021-01-28","type":"extras"},{"doc":"Fixed [Oban.Migrations] Correctly migrate up between sequential versions. The V10 migration adds check constraints, which can't be re-run safely. An attempted work-around prevented re-runs, but broke sequential upgrades like 9-&gt;10. [Oban.Queue.Producer] Kick off an initial dispatch to start processing available jobs on startup, to compensate for a lack of incoming jobs or scheduling. [Oban.Plugins.Stager] Keep rescheduling staging messages after the initial run.","ref":"changelog.html#2-4-1-2021-01-27","title":"Changelog - 2.4.1 — 2021-01-27","type":"extras"},{"doc":"Centralized Stager Plugin Queue producers no longer poll every second to stage scheduled jobs. Instead, the new Oban.Plugins.Stager plugin efficiently handles staging from a single location. This reduces overall load on the BEAM and PostgreSQL, allowing apps to easily run hundreds of queues simultaneously with little additional overhead. In a test with a single Ecto connection and 512 queues the CPU load went from 60.0% BEAM / 21.5% PostgreSQL in v2.3.4 to 0.5% BEAM / 0.0% PostgreSQL. The stager plugin is automatically injected into Oban instances and there isn't any additional configuration necessary. However, if you've set a poll_interval for Oban or an individual queue you can safely remove it. Overhauled Cron The CRON parser is entirely rewritten to be simpler and dramatically smaller. The previous parser was built on nimble_parsec and while it was fast, the compiled parser added ~5,500LOC to the code base. Thanks to a suite of property tests we can be confident that the new parser behaves identically to the previous one, and has much clearer error messages when parsing fails. Along with a new parser the crontab functionality is extracted into the Oban.Plugins.Cron plugin. For backward compatibility, top-level crontab and timezone options are transformed into a plugin declaration. If you'd like to start configuring the plugin directly change your config from: config :my_app , Oban , crontab : [ { &quot;* * * * *&quot; , MyApp.Worker } ] , timezone : &quot;America/Chicago&quot; ... To this: config :my_app , Oban , plugins : [ { Oban.Plugins.Cron , crontab : [ { &quot;* * * * *&quot; , MyApp.Worker } ] , timezone : &quot;America/Chicago&quot; } ] The plugin brings a cleaner implementation, simplified supervision tree, and eliminates the need to set crontab: false in test configuration. Improved Indexes for Unique Jobs Applications may have thousands to millions of completed jobs in storage. As the table grows the performance of inserting unique jobs can slow drastically. A new V10 migration adds necessary indexes, and paired with improved query logic it alleviates insert slowdown entirely. For comparison, a local benchmark showed that in v2.3.4 inserting a unique job into a table of 1 million jobs runs around 4.81 IPS. In v2.4.0 it runs at 925.34 IPS, nearly 200x faster . The V10 migration is optional. If you decide to migrate, first create a new migration: mix ecto.gen.migration upgrade_oban_jobs_to_v10 Next, call Oban.Migrations in the generated migration: defmodule MyApp.Repo.Migrations.UpdateObanJobsToV10 do use Ecto.Migration def up do Oban.Migrations . up ( version : 10 ) end def down do Oban.Migrations . down ( version : 9 ) end end Added [Oban] Add [:oban, :supervisor, :init] event emitted when an Oban supervisor starts. [Oban.Telemetry] Wrap built-in plugins with telemetry spans and consistently include conf in all telemetry events. [Oban.Config] Improve the error messages raised during initial validation. Also, the Config module is now public with light documentation. [Oban.Testing] Support specifying a default prefix for all test assertions. This comes with improved assertion messages that now include the prefix. [Oban.Repo] Add delete/3 as a convenient wrapper around Ecto.Repo.delete/2 . [Oban.Job] New check constraints prevent inserting jobs with an invalid priority , negative max_attempts or an attempt number beyond the maximum. Changed [Oban.Telemetry] Deprecate and replace span/3 usage in favor of the official :telemetry.span/3 , which wasn't available when span/3 was implemented. Fixed [Oban] Inspect the error reason for a failed insert!/2 call before it is raised as an error. When insert!/2 was called in a transaction the error could be :rollback , which wasn't a valid error.","ref":"changelog.html#2-4-0-2021-01-26","title":"Changelog - 2.4.0 — 2021-01-26","type":"extras"},{"doc":"Fixed [Oban.Worker] Update from_string/1 to correctly work with modules that weren't loaded.","ref":"changelog.html#2-3-4-2020-12-02","title":"Changelog - 2.3.4 — 2020-12-02","type":"extras"},{"doc":"Changed [Oban.Migration] Conditionally skip altering oban_job_state if the cancelled state is already present. This allows for a smoother upgrade path for users running PG &lt; 12. See the notes for v2.3.0 for more details.","ref":"changelog.html#2-3-3-2020-11-10","title":"Changelog - 2.3.3 — 2020-11-10","type":"extras"},{"doc":"Fixed [Oban.Migration] Restore indexes possibly removed by changing oban_job_state . This only applies to PG older than version 12. [Oban.Plugins.Pruner] Prune cancelled jobs along with completed or discarded .","ref":"changelog.html#2-3-2-2020-11-06","title":"Changelog - 2.3.2 — 2020-11-06","type":"extras"},{"doc":"Changed [Oban.Migration] Conditionally alter oban_job_state if the PG version is 12 or greater. This is vastly faster than the renaming, adding and dropping required for older PG versions.","ref":"changelog.html#2-3-1-2020-11-06","title":"Changelog - 2.3.1 — 2020-11-06","type":"extras"},{"doc":"Migration Required (V9) Migration V9 brings the new cancelled state, a cancelled_at column, and job meta . Older versions of PostgreSQL, prior to version 12, don't allow altering an enum within a transaction. If you're running an older version and want to prevent a table rewrite you can add the new cancelled at state before running the V9 migration. Create a migration with @disable_ddl_transaction true declared and run the command ALTER TYPE oban_job_state ADD VALUE IF NOT EXISTS 'cancelled' . The V9 migration will see that the cancelled value exists and won't attempt to modify the enum. After you've sorted adding the cancelled state (or ignored the issue, because you're either running PG &gt;= 12 or don't have many jobs retained), generate a new migration: mix ecto.gen.migration upgrade_oban_jobs_to_v9 Next, call Oban.Migrations in the generated migration: defmodule MyApp.Repo.Migrations.UpdateObanJobsToV9 do use Ecto.Migration def up do Oban.Migrations . up ( version : 9 ) end def down do Oban.Migrations . down ( version : 8 ) end end Oban will manage upgrading to V9 regardless of the version your application is currently using, and it will roll back a single version. Added [Oban.Job] Add new meta field for storing arbitrary job data that isn't appropriate as args . [Oban.Job] Introduce a cancelled state, along with a new cancelled_at timestamp field. Cancelling a job via Oban.cancel_job (or via Oban Web) now marks the job as cancelled rather than discarded . [Oban.Worker] Add from_string/1 for improved worker module resolution. [Oban.Telemetry] Pass the full job schema in telemetry metadata, not only select fields. Individual fields such as args , worker , etc. are still passed for backward compatibility. However, their use is deprecated and they are no longer documented. Fixed [Oban.Notifier] Fix resolution of Oban.Notifier child process in Oban.Notifier.listen/2 . [Oban.Queue.Producer] Fix cancelling jobs without a supervised process. In some circumstances, namely a hot code upgrade, the job's process could terminate without the producer tracking it and leave the job un-killable. [Oban] Only convert invalid changesets into ChangesetError from insert! . This prevents unexpected errors when insert! is called within a transaction after the transaction has rolled back.","ref":"changelog.html#2-3-0-2020-11-06","title":"Changelog - 2.3.0 — 2020-11-06","type":"extras"},{"doc":"Added [Oban] Replace local dynamically composed names with a registry. This dramatically simplifies locating nested children, avoids unnecessary atom creation at runtime and improves the performance of config lookups. [Oban.Repo] The new Oban.Repo module wraps interactions with underlying Ecto repos. This ensures consistent prefix and log level handling, while also adding full dynamic repo support. The Oban.Repo wrapper should be used when authoring plugins. [Oban.Job] Augment the unique keys option with replace_args , which allows enqueuing a unique job and replacing the args subsequently. For example, given a job with these args: %{ some_value : 1 , id : 123 } Attempting to insert a new job: %{ some_value : 2 , id : 123 } |&gt; MyJob . new ( schedule_in : 10 , replace_args : true unique : [ keys : [ :id ] ] ) |&gt; Oban . insert ( ) Will result in a single job with the args: %{ some_value : 2 , id : 123 } [Oban] Add Oban.check_queue/1,2 for runtime introspection of a queue's state. It displays a compactly formatted breakdown of the queue's configuration, a list of running jobs, and static identity information. [Oban] Add Oban.retry_job/1,2 , used to manually retry a discarded or retryable job. [Oban.Telemetry] Include tags as part of the metadata for job events. Changed [Oban.Worker] The default backoff algorithm now includes a small amount of jitter. The jitter helps prevent jobs that fail simultaneously from repeatedly retrying together. Fixed [Oban.Job] Don't include discarded state by default when accounting for uniqueness. [Oban.Cron] Fully support weekday ranges in crontab expressions. Previously, given a weekday range like MON-WED the parser would only retain the MON .","ref":"changelog.html#2-2-0-2020-10-12","title":"Changelog - 2.2.0 — 2020-10-12","type":"extras"},{"doc":"Fixed [Oban.Testing] Modify the behaviour checking in perform_job/3 so that it considers all behaviours. [Oban.Plugins.Pruner] Use distinct names when running multiple Oban instances in the same application. Added [Oban] In addition to the standard queue-name / limit keyword syntax it is now possible to define per-queue options such as :poll_interval , :dispatch_cooldown , :paused , and even override the :producer module. [Oban] Cancelling a running job now records an error on the job if it was running at the time it was cancelled. [Oban.Job] Allow using :discarded as a unique state and expose all possible states through Oban.Job.states/0 . [Oban.Worker] Allow returning {:discard, reason} from perform/1 , where the reason is recorded in the job's errors array. If no reason is provided then a default of &quot;None&quot; is used. All discarded jobs will have an error now, whether discarded manually or automatically. [Oban.Job] Add support for uniqueness across sub-args with the :keys attribute. This makes it possible to only use a few values from a job to determine uniqueness. For example, a job that includes a lot of metadata but must be considered unique based on a :url field can specify keys: [:url] . Changed [Oban.Worker] Wrap {:error, reason} and {:discard, reason} in a proper Oban.PerformError exception with a customized message. This ensures that the :error value passed to telemetry handlers is an exception and not a raw term. [Oban.Worker] Wrap job timeouts in Oban.TimeoutError with a customized message indicating the worker and timeout value. This replaces the raw :timeout atom that was reported before. [Oban.Worker] Wrap caught exits and throws in Oban.CrashError with a formatted message. This means the :error value passed to telemetry is always a proper exception and easier to report. [Oban.Worker] Stop reporting internal stacktraces for timeouts, discards or error tuples. The stacktrace was useless and potentially misleading as it appeared that the error originated from Oban rather than the worker module.","ref":"changelog.html#2-1-0-2020-08-21","title":"Changelog - 2.1.0 — 2020-08-21","type":"extras"},{"doc":"No changes from 2.0.0-rc.3 .","ref":"changelog.html#2-0-0-2020-07-10","title":"Changelog - 2.0.0 — 2020-07-10","type":"extras"},{"doc":"Fixed [Oban.Crontab.Cron] Do not raise an ArgumentError exception when the crontab configuration includes a step of 1, which is a valid step value. [Oban.Telemetry] Correctly record duration and queue_time using native time units, but log them using microseconds. Previously they used a mixture of native and microseconds, which yielded inconsistent values. Added [Oban.Telemetry] Include job queue_time in the default logger output. Changed [Oban.Plugins.Pruner] The FixedPruner is renamed to Pruner and allows users to configure the max_age value. This partially reverses the breaking change imposed by the move from prune to the FixedPruner plugin, though there isn't any support for max_len or dynamic pruning.","ref":"changelog.html#2-0-0-rc-3-2020-07-01","title":"Changelog - 2.0.0-rc.3 — 2020-07-01","type":"extras"},{"doc":"Breaking Changes [Oban] The interface for pause_queue/2 , resume_queue/2 and scale_queue/3 now matches the recently changed start_queue/2 and stop_queue/2 . All queue manipulation functions now have a consistent interface, including the ability to work in :local_only mode. See the notes around start_queue/2 and stop_queue/2 in 2.0.0-rc.1 for help upgrading. [Oban] Replace drain_queue/3 with drain_queue/2 , which now has an interface consistent with the other *_queue/2 operations. Where you previously called drain_queue/2,3 like this: Oban . drain_queue ( :myqueue , with_safety : false ) You'll now it with options, like this: Oban . drain_queue ( queue : :myqueue , with_safety : false ) Fixed [Oban.Breaker] Prevent connection bomb when the Notifier experiences repeated disconnections. [Oban.Queue.Executor] Fix error reporting when a worker fails to resolve. [Oban.Telemetry] Stop logging the :error value for circuit trip events. The error is a struct that isn't JSON encodable. We include the normalized Postgrex / DBConnection message already, so the error is redundant. Added [Oban.Crontab] Add @reboot special case to crontab scheduling [Oban.Telemetry] Add new :producer events for descheduling and dispatching jobs from queue producers. Removed [Oban] Removed kill_job/2 , which wasn't as flexible as Oban.cancel_job/2 . Use Oban.cancel_job/2 instead to safely discard scheduled jobs or killing executing jobs.","ref":"changelog.html#2-0-0-rc-2-2020-06-23","title":"Changelog - 2.0.0-rc.2 — 2020-06-23","type":"extras"},{"doc":"Breaking Changes [Oban.Config] The :verbose setting is renamed to :log . The setting started off as a simple boolean, but it has morphed to align with the log values accepted by calls to Ecto.Repo . To migrate, replace any :verbose declarations: config :my_app , Oban , verbose : false , ... With use of :log instead: config :my_app , Oban , log : false , ... [Oban] The interface for start_queue/3 is replaced with start_queue/2 and stop_queue/2 no longer accepts a queue name as the second argument. Instead, both functions now accept a keyword list of options. This enables the new local_only flag, which allows you to dynamically start and stop queues only for the local node. Where you previously called start_queue/2,3 or stop_queue/2 like this: :ok = Oban . start_queue ( :myqueue , 10 ) :ok = Oban . stop_queue ( :myqueue ) You'll now call them with options, like this: :ok = Oban . start_queue ( queue : :myqueue , limit : 10 ) :ok = Oban . stop_queue ( queue : :myqueue ) Or, to only control the queue locally: :ok = Oban . start_queue ( queue : :myqueue , limit : 10 , local_only : true ) :ok = Oban . stop_queue ( queue : :myqueue , local_only : true )","ref":"changelog.html#2-0-0-rc-1-2020-06-12","title":"Changelog - 2.0.0-rc.1 — 2020-06-12","type":"extras"},{"doc":"Breaking Changes [Oban.Worker] The perform/2 callback is replaced with perform/1 , where the only argument is an Oban.Job struct. This unifies the interface for all Oban.Worker callbacks and helps to eliminate confusion around pattern matching on arguments. To migrate change all worker definitions from accepting an args map and a job struct: def perform ( %{ &quot;id&quot; =&gt; id } , _job ) , do : IO . inspect ( id ) To accept a single job struct and match on the args key directly: def perform ( % Job { args : %{ &quot;id&quot; =&gt; id } } ) , do : IO . inspect ( id ) [Oban.Worker] The backoff/1 callback now expects a job struct instead of an integer. That allows applications to finely control backoff based on more than just the current attempt number. Use of backoff/1 with an integer is no longer supported. To migrate change any worker definitions that used a raw attempt like this: def backoff ( attempt ) , do : attempt * 60 To match on a job struct instead, like this: def backoff ( % Job { attempt : attempt } ) , do : attempt * 60 [Oban.Telemetry] The format for telemetry events has changed to match the new telemetry span convention. This listing maps the old event to the new one: [:oban, :started] -&gt; [:oban, :job, :start] [:oban, :success] -&gt; [:oban, :job, :stop] [:oban, :failure] -&gt; [:oban, :job, :exception] [:oban, :trip_circuit] -&gt; [:oban, :circuit, :trip] [:oban, :open_circuit] -&gt; [:oban, :circuit, :open] In addition, for exceptions the stacktrace meta key has changed from :stack to the standardized :stacktrace . [Oban.Prune] Configurable pruning is no longer available. Instead, pruning is handled by the new plugin system. A fixed period pruning module is enabled as a default plugin. The plugin always retains &quot;prunable&quot; (discarded or complete) jobs for 60 seconds. Remove any :prune , :prune_interval or prune_limit settings from your config. To disable the pruning plugin in test mode set plugins: false instead. [Oban.Beat] Pulse tracking and periodic job rescue are no longer available. Pulse tracking and rescuing will be handled by an external plugin. This is primarily an implementation detail, but it means that jobs may be left in the executing state after a crash or forced shutdown. Remove any :beats_maxage , :rescue_after or :rescue_interval settings from your config. Fixed [Oban.Scheduler] Ensure isolation between transaction locks in different prefixes. A node with multiple prefix-isolated instances (i.e. &quot;public&quot; and &quot;private&quot;) would always attempt to schedule cron jobs at the same moment. The first scheduler would acquire a lock and block out the second, preventing the second scheduler from ever scheduling jobs. [Oban.Query] Correctly prefix unprepared unique queries. Unique queries always targeted the &quot;public&quot; prefix, which either caused incorrect results when there were both &quot;public&quot; and an alternate prefix. In situations where there wasn't a public oban_jobs table at all it would cause cryptic transaction errors. [Oban.Query] Wrap all job fetching in an explicit transaction to enforce FOR UPDATE SKIP LOCKED semantics. Prior to this it was possible to run the same job at the same time on multiple nodes. [Oban.Crontab] Fix weekday matching for Sunday, which is represented as 0 in crontabs. Added [Oban] Bubble up errors and exits when draining queues by passing with_safety: false as an option to Oban.drain_queue/2 . [Oban] Add Oban.cancel_job/2 for safely discarding scheduled jobs or killing executing jobs. This deprecates kill_job/2 , which isn't as flexible. [Oban.Worker] Support returning {:snooze, seconds} from perform/1 to re-schedule a job some number of seconds in the future. This is useful for recycling jobs that aren't ready to run yet, e.g. because of rate limiting. [Oban.Worker] Support returning :discard from perform/1 to immediately discard a job. This is useful when a job encounters an error that won't resolve with time, e.g. invalid arguments or a missing record. [Oban.Job] Introduce a virtual unsaved_error field, which is populated with an error map after failed execution. The unsaved_error field is set before any calls to the worker's backoff/1 callback, allowing workers to calculate a custom backoff depending on the error that failed the job. [Oban.Worker] Add :infinity option for unique period. [Oban.Telemetry] Add span/3 for reporting normalized :start , :stop and :exception events with timing information. [Oban.Telemetry] Include the configured prefix in all event metadata. This makes it possible to identify which schema prefix a job ran with, which is useful for differentiating errors in a multi-tenant system. [Oban.Telemetry] Include queue_time as a measurement with stop and exception events. This is a measurement in milliseconds of the amount of time between when a job was scheduled to run and when it was last attempted. [Oban.Testing] Add perform_job/2,3 helper to automate validating, normalizing and performing jobs while unit testing. This is now the preferred way to unit test workers. To update your tests replace any calls to perform/1,2 with the new Oban.Testing.perform_job/2,3 helper: defmodule MyApp.WorkerTest do use MyApp.DataCase , async : true use Oban.Testing , repo : MyApp.Repo alias MyApp.Worker test &quot;doing business in my worker&quot; do assert :ok = perform_job ( Worker , %{ id : 1 } ) end end The perform_job/2,3 helper will verify the worker, the arguments and any provided options. It will then verify that your worker returns a valid result and return the value for you to assert on. [Oban.Crontab] Add support for non-standard expressions such as @daily , @hourly , @midnight , etc. [Oban.Crontab] Add support for using step values in conjunction with ranges, enabling expressions like 10-30/2 , 15-45/3 , etc. Changed [Oban.Notifier] Make the module public and clean up the primary function interfaces. Listening for and delivering notifications is simplified and no longer requires macros for pattern matching. Notifier dispatching performance is slightly improved as well. It is now a no-op if no processes are listening to a notification's channel. [Oban.Query] The completed_at timestamp is no longer set for failed jobs, whether they are put in the discarded or retryable state. However, the information is still available and is recorded in the errors array as the at value with the error for that attempt. This corrects a long standing inconsistency between discarding a job manually or automatically when it exhausts retries. [Oban.Producer] Stop dispatching jobs immediately on queue startup. Instead, only dispatch on the first poll. This makes it possible to send the producer a message or allow sandboxed connection access before the initial dispatch. [Oban.Worker] Limit default backoff calculations to 20 attempts, or roughly 24 days. The change addresses an issue with snoozing, which can increase a job's attempts into the hundreds or thousands. In this situation the algorithm calculates the backoff using a ratio of attempts to max attempts, but is still limited to roughly 24 days. For changes prior to 2.0 see the 1.2 branch","ref":"changelog.html#2-0-0-rc-0-2020-06-03","title":"Changelog - 2.0.0-rc.0 — 2020-06-03","type":"extras"},{"doc":"Oban is published on Hex . Add it to your list of dependencies in mix.exs : def deps do [ { :oban , &quot;~&gt; 2.10&quot; } ] end Then run mix deps.get to install Oban and its dependencies, including Ecto , Jason and Postgrex . After the packages are installed you must create a database migration to add the oban_jobs table to your database: mix ecto.gen.migration add_oban_jobs_table Open the generated migration in your editor and call the up and down functions on Oban.Migrations : defmodule MyApp.Repo.Migrations.AddObanJobsTable do use Ecto.Migration def up do Oban.Migrations . up ( ) end # We specify `version: 1` in `down`, ensuring that we&#39;ll roll all the way back down if # necessary, regardless of which version we&#39;ve migrated `up` to. def down do Oban.Migrations . down ( version : 1 ) end end This will run all of Oban's versioned migrations for your database. Migrations between versions are idempotent and will never change after a release. As new versions are released you may need to run additional migrations. Now, run the migration to create the table: mix ecto.migrate Next see Usage for how to integrate Oban into your application and start defining jobs!","ref":"installation.html","title":"Installation","type":"extras"},{"doc":"","ref":"troubleshooting.html","title":"Troubleshooting","type":"extras"},{"doc":"No Notifications with Transaction Pooling Using PG Bouncer's &quot;Transaction Pooling&quot; setup disables all of PostgreSQL's LISTEN and NOTIFY activity. Some functionality, such as triggering job execution, scaling queues, canceling jobs, etc. rely on those notifications. To ensure full functionality you must use a Repo that connects directly to the database, or use another mode like &quot;Session Pooling&quot; if possible. If you must use &quot;Transaction Pooling&quot; you can use the Repeater plugin to ensure that queues keep processing jobs: config :my_app , Oban , plugins : [ Oban.Plugins.Pruner , Oban.Plugins.Stager , Oban.Plugins.Repeater ] , ... Note: The Repeater plugin keeps jobs processing, it will not facilitate other notification based functionality, e.g. scaling queues.","ref":"troubleshooting.html#pg-bouncer","title":"Troubleshooting - PG Bouncer","type":"extras"},{"doc":"Elixir and Erlang Versions If your app crashes on launch, be sure to confirm you are running the correct version of Elixir and Erlang ( view requirements ). If using the hashnuke/elixir buildpack, you can update the elixir_buildpack.config file in your application's root directory to something like: # Elixir version elixir_version = 1.9 . 0 # Erlang version erlang_version = 22.0 . 3 Available Erlang versions are available here . Database Connections Make sure that you have enough available database connections when running on Heroku. Oban uses a database connection in order to listen for Pub/Sub notifications. This is in addition to your Ecto Repo pool_size setting. Heroku's Hobby tier Postgres plans have a maximum of 20 connections, so if you're using one of those plan accordingly.","ref":"troubleshooting.html#heroku","title":"Troubleshooting - Heroku","type":"extras"},{"doc":"While having the same Oban configuration for every environment might be fine, there are certainly times you might want to make changes for a specific environment. For example, you may want to increase or decrease a queue's concurrency.","ref":"release_configuration.html","title":"Release Configuration","type":"extras"},{"doc":"If you are using Elixir Releases, this is straight forward to do using Module Config Providers : defmodule MyApp.ConfigProvider do @moduledoc &quot;&quot;&quot; Provide release configuration for Oban Queue Concurrency &quot;&quot;&quot; @behaviour Config.Provider def init ( path ) when is_binary ( path ) , do : path def load ( config , path ) do case parse_json ( path ) do nil -&gt; config queues -&gt; Config.Reader . merge ( config , ingestion : [ { Oban , [ queues : queues ] } ] ) end end defp parse_json ( path ) do { :ok , _ } = Application . ensure_all_started ( :jason ) if File . exists? ( path ) do path |&gt; File . read! ( ) |&gt; Jason . decode! ( ) |&gt; Map . fetch! ( &quot;queues&quot; ) |&gt; Keyword . new ( fn { key , value } -&gt; { String . to_atom ( key ) , value } end ) end end end Our config provider ensures that the Jason app is loaded so that we can parse a JSON configuration file. Once the JSON is loaded we must extract the queues map and convert it to a keyword list where all of the keys are atoms. The use of String.to_atom/1 is safe because all of our queues names are already defined. Then you include this in your mix.exs file, where your release is configured: releases : [ umbrella_app : [ version : &quot;0.0.1&quot; , applications : [ child_app : :permanent ] , config_providers : [ { Path.To.ConfigProvider , &quot;/etc/config.json&quot; } ] ] ] Then when you release your app, you ensure that you have a JSON file mounted at whatever path you specified above and that it contains all of your desired queues: {&quot;queues&quot;: {&quot;special&quot;: 1, &quot;default&quot;: 10, &quot;events&quot;: 20}}","ref":"release_configuration.html#using-config-providers","title":"Release Configuration - Using Config Providers","type":"extras"},{"doc":"Oban supports the use of plugins to extend its base functionality. A plugin is any module that begins a process and exposes a start_link/1 function. That means a plugin may be a GenServer , an Agent , a Task , or any other OTP behaviour that manages a process. Realistically you'll want a long lived process to complement Oban's behaviour, which makes a GenServer or GenStateMachine ideal. Upon startup, Oban dynamically injects each plugin into its supervision tree and passes a few base options along with any custom configuration for the plugin. Every plugin receives these base options: :conf — An Oban.Config struct, which contains all of the options provided to Oban.start_link/1 in a validated and normalized form. :name — A unique via tuple scoped to the current supervision tree. The name may be used later to support helper/interface functions.","ref":"writing_plugins.html","title":"Writing Plugins","type":"extras"},{"doc":"Let's look at a tiny example plugin to get a feel for how options are passed in and how they run. Our plugin will periodically generate a table of counts for each queue / state combination and then print it out. It isn't an amazingly useful plugin, but it demonstrates how to handle options, work with the Oban.Config struct, and periodically interact with the oban_jobs table. defmodule MyApp.Plugins.Breakdown do use GenServer import Ecto.Query , only : [ group_by : 3 , select : 3 ] def start_link ( opts ) do name = Keyword . get ( opts , :name , __MODULE__ ) GenServer . start_link ( __MODULE__ , opts , name : name ) end @impl GenServer def init ( opts ) do state = Map . new ( opts ) { :ok , schedule_poll ( state ) } end @impl GenServer def handle_info ( :poll , %{ conf : conf } = state ) do breakdown = Oban.Repo . all ( conf , Oban.Job |&gt; group_by ( [ j ] , [ j . queue , j . state ] ) |&gt; select ( [ j ] , { j . queue , j . state , count ( j . id ) } ) ) IO . inspect ( breakdown ) { :noreply , schedule_poll ( state ) } end defp schedule_poll ( %{ interval : interval } ) do Process . send_after ( self ( ) , :poll , interval ) state end end The plugin's start_link/1 function expects a keyword list with :name , :conf , and :interval values. After extracting the :name for process registration, it passes the options through to init/1 . The init function converts the keyword list of options into a map for easier access and then begins a polling loop. Each iteration of the loop will query the oban_jobs table and print out a list of {queue, state, count} tuples like this: [ { &quot;default&quot; , &quot;executing&quot; , 8 } , { &quot;default&quot; , &quot;retryable&quot; , 1 } , { &quot;default&quot; , &quot;completed&quot; , 3114 } , { &quot;default&quot; , &quot;discarded&quot; , 1 } , { &quot;events&quot; , &quot;scheduled&quot; , 3 } , { &quot;events&quot; , &quot;executing&quot; , 21 } , { &quot;events&quot; , &quot;retryable&quot; , 2 } , { &quot;events&quot; , &quot;completed&quot; , 1783 } , { &quot;events&quot; , &quot;discarded&quot; , 1 } , ] Now all that is left is to adding the Breakdown module to Oban's plugin list: config :my_app , Oban , plugins : [ { MyApp.Plugins.Breakdown , interval : :timer . seconds ( 10 ) } ] ... In the configuration we're only providing the :interval value. Oban injects the :name and :conf automatically.","ref":"writing_plugins.html#example-plugin","title":"Writing Plugins - Example Plugin","type":"extras"},{"doc":"Plugins are named dynamically using via tuples, which is an effective way to manage process registration for multiple unique Oban instances. However, it makes writing interface functions for plugins a little more complicated. The solution is to make use of the Oban.Registry for process discovery. Imagine adding a pause interface function to the Breakdown plugin we built above: alias Oban.Registry def pause ( oban_name \\\\ Oban ) do oban_name |&gt; Registry . whereis ( { :plugin , __MODULE__ } ) |&gt; GenServer . call ( :pause ) end The function accepts an oban_name argument with a default of Oban , which is the default name for an Oban supervision tree. It then calls whereis/2 with a {:plugin, plugin_name} tuple and uses the returned pid to call the plugin process. You can then call pause/1 from elsewhere in the application: MyApp.Plugins.Breakdown . pause ( ) # or MyApp.Plugins.Breakdown . pause ( MyApp.OtherOban )","ref":"writing_plugins.html#calling-interface-functions","title":"Writing Plugins - Calling Interface Functions","type":"extras"},{"doc":"Plugins run directly within Oban's supervision tree. A badly behaving plugin, e.g. a plugin that crashes repeatedly, may bring down the entire supervision tree. Be sure that your plugin has safety mechanisms in place to prevent repeated crashes during startup.","ref":"writing_plugins.html#caveats","title":"Writing Plugins - Caveats","type":"extras"},{"doc":"Recursive jobs, like recursive functions, call themselves after they have executed. Except unlike recursive functions, where recursion happens in a tight loop, a recursive job enqueues a new version of itself and may add a slight delay to alleviate pressure on the queue. Recursive jobs are a great way to backfill large amounts of data where a database migration or a mix task may not be suitable. Here are a few reasons that a recursive job may be better suited for backfilling data: Data can't be backfilled with a database migration, it may require talking to an external service A task may fail partway through execution; resuming the task would mean starting over again, or tracking progress manually to resume where the failure occurred A task may be computationally intensive or put heavy pressure on the database A task may run for too long and would be interrupted by code releases or other node restarts A task may interface with an external service and require some rate limiting A job can be used directly for new records and to backfill existing records Let's explore recursive jobs with a use case that builds on several of those reasons.","ref":"recursive-jobs.html","title":"Recursive Jobs","type":"extras"},{"doc":"Consider a worker that queries an external service to determine what timezone a user resides in. The external service has a rate limit and the response time is unpredictable. We have a lot of users in our database missing timezone information, and we need to backfill. Our application has an existing TimezoneWorker that accepts a user's id , makes an external request and then updates the user's timezone. We can modify the worker to handle backfilling by adding a new clause to perform/1 . The new clause explicitly checks for a backfill argument and will enqueue the next job after it executes: defmodule MyApp.Workers.TimezoneWorker do use Oban.Worker import Ecto.Query alias MyApp . { Repo , User } @backfill_delay 1 @impl true def perform ( %{ args : %{ &quot;id&quot; =&gt; id , &quot;backfill&quot; =&gt; true } } ) do with :ok &lt;- perform ( %{ args : %{ &quot;id&quot; =&gt; id } } ) do case fetch_next ( id ) do next_id when is_integer ( next_id ) -&gt; %{ id : next_id , backfill : true } |&gt; new ( schedule_in : @backfill_delay ) |&gt; Oban . insert ( ) nil -&gt; :ok end end end def perform ( %{ args : %{ &quot;id&quot; =&gt; id } } ) do update_timezone ( id ) end defp fetch_next ( current_id ) do User |&gt; where ( [ u ] , is_nil ( u . timezone ) ) |&gt; where ( [ u ] , u . id &gt; ^ current_id ) |&gt; order_by ( asc : :id ) |&gt; limit ( 1 ) |&gt; select ( [ u ] , u . id ) |&gt; Repo . one ( ) end defp update_timezone ( _id ) , do : Enum . random ( [ :ok , { :error , :reason } ] ) end There is a lot happening in the worker module, so let's unpack it a little bit. There are two clauses for perform/1 , the first only matches when a job is marked as &quot;backfill&quot; =&gt; true , the second does the actual work of updating the timezone. The backfill clause checks that the timezone update succeeds and then uses fetch_next/1 to look for the id of the next user without a timezone. When another user needing a backfill is available it enqueues a new backfill job with a one second delay. With the new perform/1 clause in place and our code deployed we can kick off the recursive backfill. Assuming the id of the first user is 1 , you can start the job from an iex console: iex&gt; %{ id : 1 , backfill : true } |&gt; MyApp.Workers.TimezoneWorker . new ( ) |&gt; Oban . insert ( ) Now the jobs will chug along at a steady rate of one per second until the backfill is complete (or something fails). If there are any errors the backfill will pause until the failing job completes: especially useful for jobs relying on flaky external services. Finally, when there aren't any more user's without a timezone, the backfill is complete and recursion will stop.","ref":"recursive-jobs.html#use-case-backfilling-timezone-data","title":"Recursive Jobs - Use Case: Backfilling Timezone Data","type":"extras"},{"doc":"This was a relatively simple example, and hopefully it illustrates the power and flexibility of recursive jobs. Recursive jobs are a general pattern and aren't specific to Oban. In fact, aside from the use Oban.Worker directive there isn't anything specific to Oban in the recipe!","ref":"recursive-jobs.html#building-on-recursive-jobs","title":"Recursive Jobs - Building On Recursive Jobs","type":"extras"},{"doc":"A common variant of recursive jobs are &quot;scheduled jobs&quot;, where the goal is for a job to repeat indefinitely with a fixed amount of time between executions. The part that makes it &quot;reliable&quot; is the guarantee that we'll keep retrying the job's business logic when the job retries, but we'll only schedule the next occurrence once . In order to achieve this guarantee we'll make use of the perform function to receive a complete Oban.Job struct. Time for illustrative example!","ref":"reliable-scheduling.html","title":"Reliable Scheduled Jobs","type":"extras"},{"doc":"When a new user signs up to use our site we need to start sending them daily digest emails. We want to deliver the emails around the same time a user signed up, repeating every 24 hours. It is important that we don't spam them with duplicate emails, so we ensure that the next email is only scheduled on our first attempt. defmodule MyApp.Workers.ScheduledWorker do use Oban.Worker , queue : :scheduled , max_attempts : 10 alias MyApp.Mailer @one_day 60 * 60 * 24 @impl true def perform ( %{ args : %{ &quot;email&quot; =&gt; email } = args , attempt : 1 } ) do args |&gt; new ( schedule_in : @one_day ) |&gt; Oban . insert! ( ) Mailer . deliver_email ( email ) end def perform ( %{ args : %{ &quot;email&quot; =&gt; email } } ) do Mailer . deliver_email ( email ) end end You'll notice that the first perform/1 clause only matches a job struct on the first attempt. When it matches, the first clause schedules the next iteration immediately, before attempting to deliver the email. Any subsequent retries fall through to the second perform/1 clause, which only attempts to deliver the email again. Combined, the clauses get us close to at-most-once semantics for scheduling , and at-least-once semantics for delivery .","ref":"reliable-scheduling.html#use-case-delivering-daily-digest-emails","title":"Reliable Scheduled Jobs - Use Case: Delivering Daily Digest Emails","type":"extras"},{"doc":"Delivering around the same time using cron-style scheduling would need extra book-keeping to check when a user signed up, and then only deliver to those users that signed up within that window of time. The recursive scheduling approach is more accurate and entirely self contained—when and if the digest interval changes the scheduling will pick it up automatically once our code deploys. An extensive discussion on the Oban issue tracker prompted this example along with the underlying feature that made it possible.","ref":"reliable-scheduling.html#more-flexible-than-cron-scheduling","title":"Reliable Scheduled Jobs - More Flexible Than CRON Scheduling","type":"extras"},{"doc":"Most applications provide some way to generate an artifact—something that may take the server a long time to accomplish. If it takes several minutes to render a video, crunch some numbers or generate an export, users may be left wondering whether your application is working. Providing periodic updates to end users assures them that the work is being done and keeps the application feeling responsive. Reporting progress is something that any background job processor with unlimited execution time can do! Naturally, we'll look at an example built on Oban.","ref":"reporting-progress.html","title":"Reporting Job Progress","type":"extras"},{"doc":"Users of our site can export a zip of all the files they have uploaded. A zip file (no, not a tar, our users don't have neck-beards) is generated on the fly, when the user requests it. Lazily generating archives is great for our server's utilization, but it means that users may wait a while when there are many files. Fortunately, we know how many files will be included in the zip and we can use that information to send progress reports! We will compute the archive's percent complete as each file is added and push a message to the user. Before We Start In the forum question that prompted this guide the work was done externally by a port process. Working with ports is well outside the scope of this guide, so I've modified it for the sake of simplicity. The result is slightly contrived as it puts both processes within the same module, which isn't necessary if the only goal is to broadcast progress. This guide is ultimately about coordinating processes to report progress from a background job, so that's what we'll focus on (everything else will be rather hand-wavy ).","ref":"reporting-progress.html#use-case-exporting-a-large-zip-file","title":"Reporting Job Progress - Use Case: Exporting a Large Zip File","type":"extras"},{"doc":"Our worker, the creatively titled ZippingWorker , handles both building the archive and reporting progress to the client. Showing the entire module at once felt distracting, so we'll start with only the module definition and the perform/1 function: defmodule MyApp.Workers.ZippingWorker do use Oban.Worker , queue : :exports , max_attempts : 1 alias MyApp . { Endpoint , Zipper } def perform ( % _ { args : %{ &quot;channel&quot; =&gt; channel , &quot;paths&quot; =&gt; paths } } ) do build_zip ( paths ) await_zip ( channel ) end # ... end The function accepts an Oban Job with a channel name and a list of file paths, which it immediately passes on to the private build_zip/1 : defp build_zip ( paths ) do job_pid = self ( ) Task . async ( fn -&gt; zip_path = Zipper . new ( ) paths |&gt; Enum . with_index ( 1 ) |&gt; Enum . each ( fn { path , index } -&gt; :ok = Zipper . add_file ( zip_path , path ) send ( job_pid , { :progress , trunc ( index / length ( paths ) * 100 ) } ) end ) send ( job_pid , { :complete , zip_path } ) end ) end The function grabs the current pid, which belongs to the job, and kicks off an asynchronous task to handle the zipping. With a few calls to a fictional Zipper module the task works through each file path, adding it to the zip. After adding a file the task sends a :progress message with the percent complete back to the job. Finally, when the zip finishes, the task sends a :complete message with a path to the archive. The asynchronous call spawns a separate process and returns immediately. In order for the task to finish building the zip we need to wait on it. Typically we'd use Task.await/2 , but we'll use a custom receive loop to track the task's progress: defp await_zip ( channel ) do receive do { :progress , percent } -&gt; Endpoint . broadcast ( channel , &quot;zip:progress&quot; , percent ) await_zip ( channel ) { :complete , zip_path } -&gt; Endpoint . broadcast ( channel , &quot;zip:complete&quot; , zip_path ) after 30_000 -&gt; Endpoint . broadcast ( channel , &quot;zip:failed&quot; , &quot;zipping failed&quot; ) raise RuntimeError , &quot;no progress after 30s&quot; end end The receive loop blocks execution while it waits for :progress or :complete messages. When a message comes in it broadcasts to the provided channel and the client receives an update (this example uses Phoenix Channels , but any other PubSub type mechanism would work). As a safety mechanism we have an after clause that will timeout after 30 seconds of inactivity. If the receive block times out we notify the client and raise an error, failing the job.","ref":"reporting-progress.html#coordinating-processes","title":"Reporting Job Progress - Coordinating Processes","type":"extras"},{"doc":"Reporting progress asynchronously works in Oban because anything that blocks a worker's perform/1 function will keep the job executing. Jobs aren't executed inside of a transaction, which alleviates any limitations on how long a job can run. This technique is suitable for any single long running job where an end user is waiting on the results. You can read the &quot;Batching Jobs for Monitoring&quot; guide if you need to combine multiple jobs into a single output by creating batch jobs .","ref":"reporting-progress.html#made-possible-by-unlimited-execution","title":"Reporting Job Progress - Made Possible by Unlimited Execution","type":"extras"},{"doc":"Reporting job errors by sending notifications to an external service is essential to maintaining application health. While reporting is essential, noisy reports for flaky jobs can become a distraction that gets ignored. Sometimes we expect that a job will error a few times. That could be because the job relies on an external service that is flaky, because it is prone to race conditions, or because the world is a crazy place. Regardless of why a job fails, reporting every failure may be undesirable.","ref":"expected-failures.html","title":"Handling Expected Failures","type":"extras"},{"doc":"One solution for reducing noisy error notifications is to start reporting only after a job has failed several times. Oban uses Telemetry to make reporting errors and exceptions a simple matter of attaching a handler function. In this example we will extend Honeybadger reporting from the Oban.Telemetry documentation, but account for the number of processing attempts. To start, we'll define a Reportable protocol with a single reportable?/2 function: defprotocol MyApp.Reportable do @fallback_to_any true def reportable? ( worker , attempt ) end defimpl MyApp.Reportable , for : Any do def reportable? ( _worker , _attempt ) , do : true end The Reportable protocol has a default implementation which always returns true , meaning it reports all errors. Our application has a FlakyWorker that's known to fail a few times before succeeding. We don't want to see a report until after a job has failed three times, so we'll add an implementation of Reportable within the worker module: defmodule MyApp.Workers.FlakyWorker do use Oban.Worker defstruct [ ] defimpl MyApp.Reportable do @threshold 3 def reportable? ( _worker , attempt ) , do : attempt &gt; @threshold end @impl true def perform ( %{ args : %{ &quot;email&quot; =&gt; email } } ) do MyApp.ExternalService . deliver ( email ) end end Note that we've also used defstruct [] to make our worker a viable struct. This is necessary for our protocol to dispatch correctly, as protocols consider all modules to be a plain atom . The final step is to call reportable?/2 from our application's error reporter, passing in the worker module and the attempt number: defmodule MyApp.ErrorReporter do alias MyApp.Reportable def handle_event ( _ , _ , meta , _ ) do worker_struct = maybe_get_worker_struct ( meta . job . worker ) if Reportable . reportable? ( worker_struct , meta . job . attempt ) do context = Map . take ( meta . job , [ :id , :args , :queue , :worker ] ) Honeybadger . notify ( meta . reason , context , meta . stacktrace ) end end def maybe_get_worker_struct ( worker ) do try do { :ok , module } = Oban.Worker . from_string ( ) struct ( module ) rescue UndefinedFunctionError -&gt; worker end end end Attach the failure handler somewhere in your application.ex module: :telemetry . attach ( &quot;oban-errors&quot; , [ :oban , :job , :exception ] , &amp; ErrorReporter . handle_event / 4 , nil ) With the failure handler attached you will start getting error reports only after the third error . Giving Time to Recover If a service is especially flaky you may find that Oban's default backoff strategy is too fast. By defining a custom backoff function on the FlakyWorker we can set a linear delay before retries: # inside of MyApp.Workers.FlakyWorker @impl true def backoff ( attempt , base_amount \\\\ 60 ) do attempt * base_amount end Now the first retry is scheduled 60s later, the second 120s later, and so on. Building Blocks Elixir's powerful primitives of behaviours, protocols and event handling make flexible error reporting seamless and extensible. While our Reportable protocol only considered the number of attempts, this same mechanism is suitable for filtering by any other meta value.","ref":"expected-failures.html#use-case-silencing-initial-notifications-for-flaky-services","title":"Handling Expected Failures - Use Case: Silencing Initial Notifications for Flaky Services","type":"extras"},{"doc":"Running every job queue on every node isn't always ideal. Imagine that your application has some CPU intensive jobs that you'd prefer not to run on nodes that serve web requests. Perhaps you start temporary nodes that are only meant to insert jobs but should never execute any. Fortunately, we can control this by configuring certain node types, or even single nodes, to run only a subset of queues .","ref":"splitting-queues.html","title":"Splitting Queues Between Nodes","type":"extras"},{"doc":"One notorious type of CPU intensive work is video processing. When our application is transcoding multiple videos simultaneously it is a major drain on system resources and may impact response times. To avoid this we can run dedicated worker nodes that don't serve any web requests and handle all of the transcoding. While it's possible to separate our system into web and worker apps within an umbrella, that wouldn't allow us to dynamically change queues at runtime. Let's look at an environment variable based method for dynamically configuring queues at runtime. Within config.exs our application is configured to run three queues: default , media and events : config :my_app , Oban , repo : MyApp.Repo , queues : [ default : 15 , media : 10 , events : 25 ] We will use an OBAN_QUEUES environment variable to override the queues at runtime. For illustration purposes the queue parsing all happens within the application module, but it would work equally well in releases.exs . defmodule MyApp.Application do @moduledoc false use Application def start ( _type , _args ) do children = [ MyApp.Repo , MyApp.Endpoint , { Oban , oban_opts ( ) } ] Supervisor . start_link ( children , strategy : :one_for_one , name : MyApp.Supervisor ) end defp oban_opts do env_queues = System . get_env ( &quot;OBAN_QUEUES&quot; ) :my_app |&gt; Application . get_env ( Oban ) |&gt; Keyword . update ( :queues , [ ] , &amp; queues ( env_queues , &amp;1 ) ) end defp queues ( &quot;*&quot; , defaults ) , do : defaults defp queues ( nil , defaults ) , do : defaults defp queues ( _ , false ) , do : false defp queues ( values , _defaults ) when is_binary ( values ) do values |&gt; String . split ( &quot; &quot; , trim : true ) |&gt; Enum . map ( &amp; String . split ( &amp;1 , &quot;,&quot; , trim : true ) ) |&gt; Keyword . new ( fn [ queue , limit ] -&gt; { String . to_existing_atom ( queue ) , String . to_integer ( limit ) } end ) end end The queues function's first three clauses ensure that we can fall back to the queues specified in our configuration (or false , for testing). The fourth clause is much more involved, and that is where the environment parsing happens. It expects the OBAN_QUEUES value to be a string formatted as queue,limit pairs and separated by spaces. For example, to run only the default and media queues with a limit of 5 and 10 respectively, you would pass the string default,5 media,10 . Note that the parsing clause has a couple of safety mechanisms to ensure that only real queues are specified: It automatically trims while splitting values, so extra whitespace like won't break parsing (i.e. default,3 ) It only converts the queue string to an existing atom , hopefully preventing typos that would start a random queue (i.e. defalt ) Usage Examples In development (or when using mix rather than releases) we can specify the environment variable inline: OBAN_QUEUES=&quot;default,10 media,5&quot; mix phx.server # default: 10, media: 5 We can also explicitly opt in to running all of the configured queues: OBAN_QUEUES=&quot;*&quot; mix phx.server # default: 15, media: 10, events: 25 Finally, without OBAN_QUEUES set at all it will implicitly fall back to the configured queues: mix phx.server # default: 15, media: 10, events: 25","ref":"splitting-queues.html#use-case-isolating-video-processing-intensive-jobs","title":"Splitting Queues Between Nodes - Use Case: Isolating Video Processing Intensive Jobs","type":"extras"},{"doc":"This environment variable based solution is more flexible than running separate umbrella apps because we can reconfigure at any time. In a limited environment, like staging, we can run all the queues on a single node using the exact same code we use in production. In the future, if other workers start to utilize too much CPU or RAM we can shift them to the worker node without any code changes . This guide was prompted by an inquiry on the Oban issue tracker.","ref":"splitting-queues.html#flexible-across-all-environments","title":"Splitting Queues Between Nodes - Flexible Across all Environments","type":"extras"},{"doc":"This information is extracted and expanded from the CHANGELOG . This is a big release with a ton of new features and fixes, but also a few breaking changes. This guide will walk you through the upgrade process.","ref":"v2-0.html","title":"Upgrading to v2.0","type":"extras"},{"doc":"Update Oban to the latest version: defp deps do [ { :oban , &quot;~&gt; 2.0.0&quot; } ... ] end If you're an Oban Web+Pro user you'll also need to bump Oban.Web and add Oban.Pro: defp deps do [ { :oban , &quot;~&gt; 2.0.0&quot; } , { :oban_web , &quot;~&gt; 2.0.0&quot; , organization : &quot;oban&quot; } , { :oban_pro , &quot;~&gt; 0.3.0&quot; , organization : &quot;oban&quot; } ... ] end","ref":"v2-0.html#bump-your-deps","title":"Upgrading to v2.0 - Bump Your Deps","type":"extras"},{"doc":"The perform/2 callback is replaced with perform/1 , where the only argument is an Oban.Job struct. This unifies the interface for all Oban.Worker callbacks and helps to eliminate confusion around pattern matching on arguments. Change all worker definitions from accepting args and a job to only accept a job and match on the nested args key: - def perform(%{&quot;id&quot; =&gt; id}, _job) + def perform(%Job{args: %{&quot;id&quot; =&gt; id}}) For finer control of job backoff the backoff/1 callback now expects a job struct instead of an integer. Change any workers that expect an attempt number to match on the full job struct: - def backoff(attempt) + def backoff(%Job{attempt: attempt})","ref":"v2-0.html#oban-worker-changes","title":"Upgrading to v2.0 - Oban.Worker Changes","type":"extras"},{"doc":"The :verbose setting was renamed to :log to better align with values accepted by Ecto.Repo . config :my_app, Oban, - verbose: false, + log: false ... Pruning is now handled by the new plugin system. Replace :prune in your config and pass a :max_age value to the plugin: config :my_app, Oban, - prune: {:max_age, 60}, - prune_interval: 30_000, - prune_limit: 5000, + plugins: [{Oban.Plugins.Pruner, max_age: 60}] ... In test mode you can disable pruning by setting plugins to false instead: config :my_app, Oban, - prune: false, + plugins: false ... 🌟 Oban.Pro users may opt to use the DynamicPruner instead for finer control. For example, to set per-state retention periods: config :my_app , Oban , plugins : [ { Oban.Pro.Plugins.DynamicPruner , state_overrides : [ completed : { :max_age , { 1 , :day } } , discarded : { :max_age , { 1 , :month } } ] } ] ... Pulse tracking and orphaned job rescue are removed from base Oban. This change means that you will need to manually rescue any jobs left executing state after a crash or forced shutdown. Remove any :beats_maxage , :rescue_after or :rescue_interval settings from your config: config :my_app, Oban, - beats_maxage: 10, - rescue_after: 30_000, - rescue_interval: 5_000 ... 🌟 Oban.Pro users may use the Lifeline plugin to retain automatic orphaned job rescue with lightweight heartbeat recording: config :my_app , Oban , plugins : [ Oban.Pro.Plugins.Lifeline ] ...","ref":"v2-0.html#update-your-config","title":"Upgrading to v2.0 - Update Your Config","type":"extras"},{"doc":"The new perform_job test helper automates validating, normalizing and perform jobs in unit tests. To update your tests replace any calls to perform with the new helper: - assert :ok = MyApp.Worker.perform(%{id: 1}, %Oban.Job{}) + assert :ok = perform_job(MyApp.Worker, %{id: 1}) The perform_job helper will verify the worker, the arguments and any provided options. It will then verify that your worker returns a valid result and return the value for you to assert on. Within integration tests replace drain_queue/3 with drain_queue/2 : - Oban.drain_queue(:myqueue, with_safety: false) + Oban.drain_queue(queue: :myqueue, with_safety: false)","ref":"v2-0.html#update-your-tests","title":"Upgrading to v2.0 - Update Your Tests","type":"extras"},{"doc":"Telemetry event names have changed, along with some of the metadata and timing units. The new event names are consistent and align with the now-standard telemetry:span conventions. Update handlers to and match on the new event names: - def handle_event([:oban, :failure], measure, meta, _) do + def handle_event([:oban, :job, :exception], measure, meta, _) do And update the attach calls: - :telemetry.attach(&quot;oban&quot;, [[:oban, :failure]], &amp;handle_event/4, %{}) + :telemetry.attach(&quot;oban&quot;, [[:oban, :job, :exception]], &amp;handle_event/4, %{}) Finally, replace references to meta.stack with meta.stacktrace : - Sentry.capture_exception(meta.error, stacktrace: meta.stack, extra: extra) + Sentry.capture_exception(meta.error, stacktrace: meta.stacktrace, extra: extra) Here is a conversion chart for any other handlers you may have: [:oban, :started] -&gt; [:oban, :job, :start] [:oban, :success] -&gt; [:oban, :job, :stop] [:oban, :failure] -&gt; [:oban, :job, :exception] [:oban, :trip_circuit] -&gt; [:oban, :circuit, :trip] [:oban, :open_circuit] -&gt; [:oban, :circuit, :open]","ref":"v2-0.html#update-telemetry","title":"Upgrading to v2.0 - Update Telemetry","type":"extras"},{"doc":"Oban.Web is streamlined to share configuration with Oban and use the new plugin system. Remove any Oban.Web specific configuration: - config :my_app, Oban.Web, repo: MyApp.Repo Next, add the necessary plugins to the Oban config: config :my_app , Oban , plugins : [ Oban.Pro.Plugins.DynamicPruner , Oban.Pro.Plugins.Lifeline , Oban.Web.Plugins.Stats ] , ... Now remove ObanWeb from your application's supervision tree: children = [ MyApp.Repo, MyApp.Endpoint, {Oban, oban_opts}, - {ObanWeb, oban_web_opts} ] And delete any existing ObanWeb migrations: - defdelegate up, to: ObanWeb.Migrations - defdelegate down, to: ObanWeb.Migrations Finally, use the oban_dashboard macro to easily mount in your router: + import Oban.Web.Router scope &quot;/&quot; do pipe_through :browser - live &quot;/oban&quot;, ObanWeb.DashboardLive, layout: {ObanWeb.LayoutView, &quot;app.html&quot;} + oban_dashboard(&quot;/oban&quot;) end","ref":"v2-0.html#update-oban-web-optional","title":"Upgrading to v2.0 - Update Oban.Web (Optional)","type":"extras"},{"doc":"For Oban OSS users the v2.6 upgrade is a drop in replacement—there isn't anything to do! However, Web+Pro users will need to make some changes to unlock the goodness of engines.","ref":"v2-6.html","title":"Upgrading to v2.6","type":"extras"},{"doc":"Update Oban, Web and Pro to the latest versions: [ { :oban , &quot;~&gt; 2.6&quot; } , { :oban_web , &quot;~&gt; 2.6&quot; , organization : &quot;oban&quot; } , { :oban_pro , &quot;~&gt; 0.7&quot; , organization : &quot;oban&quot; } ... ] Be sure to specify both :oban_web and :oban_pro if you use them both. There aren't any dependencies between Web and Pro now. That means you're free to use Pro for workers and only include Web for Phoenix servers, etc.","ref":"v2-6.html#bump-your-deps","title":"Upgrading to v2.6 - Bump Your Deps","type":"extras"},{"doc":"The SmartEngine uses centralized records to track and exchange state globally, enabling features such as global concurrency. First, create a migration to add the new oban_producers table: $ mix ecto.gen.migration add_oban_producers Within the migration module: use Ecto.Migration defdelegate change , to : Oban.Pro.Migrations.Producers If you have multiple Oban instances or use prefixes, you can specify the prefix and create multiple tables in one migration: use Ecto.Migration def change do Oban.Pro.Migrations.Producers . change ( ) Oban.Pro.Migrations.Producers . change ( prefix : &quot;special&quot; ) Oban.Pro.Migrations.Producers . change ( prefix : &quot;private&quot; ) end Next, update your config to use the SmartEngine : config :my_app , Oban , engine : Oban.Pro.Queue.SmartEngine , ... If you have multiple Oban instances you need to configure each one to use the SmartEngine , otherwise they'll default to the BasicEngine .","ref":"v2-6.html#switch-to-the-smartengine","title":"Upgrading to v2.6 - Switch to the SmartEngine","type":"extras"},{"doc":"Oban Pro no longer writes heartbeat records to oban_beats . Instead, any Oban instance that runs queues must use the Gossip plugin to broadcast status via PubSub. To start, include the Gossip plugin in your Oban config: config :my_app , Oban , plugins : [ Oban.Plugins.Gossip ... With the default configuration the plugin will broadcast every 1 second. If that is too frequent you can configure the interval: plugins : [ { Oban.Plugins.Gossip , interval : :timer . seconds ( 5 ) } ...","ref":"v2-6.html#start-gossiping","title":"Upgrading to v2.6 - Start Gossiping","type":"extras"},{"doc":"Due to an improvement in how configuration is passed to workers the WorkflowManager plugin is no longer needed. You can remove it from your list of plugins: plugins: [ Oban.Pro.Plugins.Lifeline, - Oban.Pro.Plugins.WorkflowManager, ...","ref":"v2-6.html#remove-the-workflow-manager","title":"Upgrading to v2.6 - Remove the Workflow Manager","type":"extras"},{"doc":"The Lifeline plugin is simplified and doesn't accept as many configuration options. If you previously configured the record_interval or delete_interval you can remove them: plugins: [{ Oban.Pro.Plugins.Lifeline, - delete_interval: :timer.minutes(10), - record_interval: :timer.seconds(10), rescue_interval: :timer.minutes(5) }]","ref":"v2-6.html#remove-extra-lifeline-options","title":"Upgrading to v2.6 - Remove Extra Lifeline Options","type":"extras"},{"doc":"Once you've rolled out the switch to producer records, the smart engine and the gossip plugin you are free to remove the oban_beats table at your discretion (preferrably in a follow up release, to prevent errors): $ mix ecto.gen.migration drop_oban_beats Within the generated migration module: use Ecto.Migration def up do drop_if_exists table ( &quot;oban_beats&quot; ) drop_if_exists table ( &quot;oban_beats&quot; , prefix : &quot;private&quot; ) # If you have any prefixes: end def down do # No going back! end","ref":"v2-6.html#drop-the-beats-table","title":"Upgrading to v2.6 - Drop the Beats Table","type":"extras"},{"doc":"🌟 Oban.Pro is a collection of plugins , workers and extensions that improve Oban's reliability even further and make difficult workflows possible. It is provided as a separate package and is available through a paid license. Get started with installation .","ref":"pro_overview.html","title":"Overview","type":"extras"},{"doc":"Smart Engine — A queue engine that enables truly global concurrency and global rate limiting.","ref":"pro_overview.html#extensions","title":"Overview - Extensions","type":"extras"},{"doc":"PG Notifier — An alternate notifier implementation using process groups and built on Distributed Erlang.","ref":"pro_overview.html#notifiers","title":"Overview - Notifiers","type":"extras"},{"doc":"DynamicPruner — Delete completed or exhausted jobs based on configurable limits with detailed overrides. DynamicCron — Configure cron scheduling at runtime, globally, across an entire cluster. Lifeline — Rescue orphaned jobs and periodically prune outdated producer records. Relay — Insert and await the results of jobs across any number of nodes, i.e. persistent distributed tasks. Reprioritizer — Prevents queue starvation by automatically adjusting priorities to ensure all jobs are eventually processed.","ref":"pro_overview.html#plugins","title":"Overview - Plugins","type":"extras"},{"doc":"Batch — Process jobs in groups while tracking overall progress across all nodes and executing optional callbacks. Chunk — Process jobs atomically in groups based on size or timeout, with robust error handling semantics. Workflow — Compose workers together with arbitrary dependencies between jobs, allowing sequential, fan-out and fan-in execution.","ref":"pro_overview.html#workers","title":"Overview - Workers","type":"extras"},{"doc":"All notable changes to Oban.Pro are documented here.","ref":"pro-changelog.html","title":"Changelog","type":"extras"},{"doc":"Fixed [SmartEngine] Verify the presence of a rate-limit period before calculating window time. This fixes situations where a producer record for the same queue existed, but lacked a rate limit structure. [Producer] Improve legacy Ecto support by lazily calculating the local limit, without assuming the params are coerced into a map.","ref":"pro-changelog.html#v0-9-2-2021-10-13","title":"Changelog - v0.9.2 — 2021-10-13","type":"extras"},{"doc":"Fixed [SmartEngine] Respect configured log level when performing transaction lock queries. Previously, the Smart Engine could log every time it made an advisory lock query with a global_limit or rate_limit set.","ref":"pro-changelog.html#v0-9-1-2021-09-26","title":"Changelog - v0.9.1 — 2021-09-26","type":"extras"},{"doc":"Partitioned Rate Limiting Driven by popular demand, v0.9 brings partitions to the Smart Engine's rate limiter. With partitions, rate limits are applied per-worker, on args, or on a subset of args fields rather than across the entire queue. This enables your application to enforce limits per-customer or respect external throttling, without splitting jobs into multiple queues. rate_limit : [ allowed : 100 , period : 5 , partition : [ fields : [ :worker ] ] Alternatively, you can partition by the job's account_id field: rate_limit : [ allowed : 100 , period : 5 , partition : [ fields : [ :args ] , keys : [ :account_id ] ] ] Naturally, a combination of worker , args , and any number of keys works. Check out the SmartEngine Guide for options and details. Batch Worker Improvements Batches are the oldest worker in Pro, and as such, they existed prior to the meta field. Finally, that difference is rectified. Building batches on meta enables a handful of ergonomic improvements and new functionality: Forwarding args through the batch_callback_args option Heterogeneous batches with an alternative callback module through the batch_callback_worker option Fetch all jobs in a batch with stream_batch_jobs/2 for map/reduce processing Check out the updated Batch Guide for more examples, configuration options, use cases, and typespecs. IMPORTANT : For in-flight batch callback jobs to run after an upgrade you'll need to migrate batch_id and callback into meta . Run the following SQL in a migration immediately prior to, or following, the upgrade. UPDATE oban_jobs SET meta = jsonb_set_lax( jsonb_set(meta, '{batch_id}', args-&gt;'batch_id'), '{callback}', args-&gt;'callback', true, 'return_target' ) WHERE state NOT IN ('cancelled', 'completed', 'discarded') AND args ? 'batch_id' Workflow Worker Improvements Workflows got a little ergonomic love, too. Now you can dynamically extend workflows at runtime with the new append_workflow function: def process ( % Job { } = job ) do jobs = job |&gt; append_workflow ( check_deps : false ) |&gt; add ( :d , WorkerD . new ( %{ } ) , deps : [ :a ] ) |&gt; add ( :e , WorkerE . new ( %{ } ) , deps : [ :b ] ) |&gt; add ( :f , WorkerF . new ( %{ } ) , deps : [ :c ] ) |&gt; Oban . insert_all ( ) { :ok , jobs } end Note the use of check_deps: false to prevent dependency validation. To be safe and check jobs while appending, we'll use the new stream_workflow_jobs/1 function to load all of the previous jobs and feed them in: def process ( % Job { } = job ) do { :ok , jobs } = MyApp.Repo . transaction ( fn -&gt; job |&gt; stream_workflow_jobs ( ) |&gt; Enum . to_list ( ) end ) jobs |&gt; append_workflow ( ) |&gt; add ( :d , WorkerD . new ( %{ } ) , deps : [ :a ] ) |&gt; add ( :e , WorkerE . new ( %{ } ) , deps : [ :b ] ) |&gt; add ( :f , WorkerF . new ( %{ } ) , deps : [ :c ] ) |&gt; Oban . insert_all ( ) :ok end Check out the updated Workflow Guide for details on appending to workflows, job streaming, and complete typespecs. Changed [Oban] Require Oban ~&gt; v2.9 to support the new cancel_all_jobs engine callback. [Oban.Pro.Queue.SmartEngine] Make operations like refresh more efficient (less data transport) and more failure tolerant to prevent producer crashes.","ref":"pro-changelog.html#v0-9-0-2021-09-15","title":"Changelog - v0.9.0 — 2021-09-15","type":"extras"},{"doc":"Fixed [Workflow] Ensure a minimum snooze when all deps are scheduled. When an upstream dependency was scheduled to run right now, or sometime in the past, it would return an invalid value like {:snooze, 0} . In that case the result was considered a success and the job wouldn't run as expected. [DynamicCron] Prevent crashing when a cron's worker module is missing. This could happen during rolling or blue/green deploys. [PG] Correct ident check when receiving scoped messages for pause, resume, or scale type actions.","ref":"pro-changelog.html#v0-8-2-2021-07-30","title":"Changelog - v0.8.2 — 2021-07-30","type":"extras"},{"doc":"Fixed [SmartEngine] Correctly track the rate limit window between days. Rate limited queues could get stuck when fetch invocations straddled midnight. In that scenario the queue wouldn't start fetching more until the clock time for the current day exceeded the last recorded time from the previous day. [SmartEngine] Silence operations in multi transactions. The log level wasn't respected for some frequently used operations. [Relay] Remove &quot;local&quot; optimization that could incorrectly prevent results from getting back to the calling node.","ref":"pro-changelog.html#v0-8-1-2021-06-01","title":"Changelog - v0.8.1 — 2021-06-01","type":"extras"},{"doc":"Added [PG] An alternate Oban.Notifier implementation based on pg/pg2 for use with Distributed Erlang. [DynamicCron] Store timezone overrides for individual cron entries. Along with the a timezone for the DynamicCron plugin, individual cron entries now support a timezone override. The timezone can be set when inserting or updating cron entries. Changed [Relay] Decode relay results using the :safe option to prevent possible atom table exhaustion or remote code injection. Fixed [SmartEngine] Implementing a cancel_job/2 callback within the engine makes cancelling jobs safe. Previously cancelling a job could leave the job id in the list of running jobs, which could eventually exhaust the concurrency limit and prevent a queue from running new jobs.","ref":"pro-changelog.html#v0-8-0-2021-05-25","title":"Changelog - v0.8.0 — 2021-05-25","type":"extras"},{"doc":"Changed [SmartEngine] Use database connection safety for all SmartEngine calls. Database connections may be unstable for many reasons: pool contention, database overload, crashes. We shouldn't crash the producer due to a flickering database call. Fixed [DynamicCron] Explicitly declare cron timestamps as DateTime to compensate for apps that configure Ecto to use NaiveDateTime by default.","ref":"pro-changelog.html#v0-7-3-2021-05-19","title":"Changelog - v0.7.3 — 2021-05-19","type":"extras"},{"doc":"Changed [Lifeline] The plugin will only run along with the SmartEngine. Running with the BasicEngine is flawed and will cause inadvertent orphan rescues. You can either switch to the SmartEngine or remove Lifeline from your plugins. Fixed [SmartEngine] Consider the configured prefix for all ack and update options.","ref":"pro-changelog.html#v0-7-2-2021-04-11","title":"Changelog - v0.7.2 — 2021-04-11","type":"extras"},{"doc":"Changed [Relay] Restore the ability to use Elixir 1.9 by removing use of the is_map_key guard that wasn't introduced until Elixir 1.10. [SmartEngine] Apply jitter and more selective use of the mutex used for global coordination. The engine also avoids touching or reloading producer records to minimize connection use. [Workflow] Improve deps check performance by using an index powered operator. On instances with millions of jobs this may improve performance ~600x.","ref":"pro-changelog.html#v0-7-1-2021-04-07","title":"Changelog - v0.7.1 — 2021-04-07","type":"extras"},{"doc":"Pluggable Smart Engine Pro's SmartEngine enables truly global concurrency and global rate limiting by handling demand at the queue level, without any conflict or churn. The engine uses centralized producer records to coordinate with minimal load on the database. config :my_app , Oban , engine : Oban.Pro.Queue.SmartEngine , queues : [ alpha : 1 , gamma : [ global_limit : 1 ] , delta : [ local_limit : 2 , global_limit : 5 ] , kappa : [ local_limit : 5 , rate_limit : [ allowed : 30 , period : { 1 , :minute } ] ] , omega : [ global_limit : 1 , rate_limit : [ allowed : 500 , period : { 1 , :hour } ] ] ] , ... Check out the SmartEngine Guide for a walkthrough and details. Relay Plugin The Relay plugin lets you insert and await the results of jobs locally or remotely, across any number of nodes, i.e. persistent distributed tasks. Once the plugin is running, you can seamlessly distribute oban jobs and await the results synchronously: alias Oban.Pro.Plugins.Relay 1 .. 3 |&gt; Enum . map ( &amp; DoubleWorker . new ( %{ int : &amp;1 } ) ) |&gt; Enum . map ( &amp; Relay . async / 1 ) |&gt; Relay . await_many ( timeout : :timer . seconds ( 1 ) ) # [{:ok, 2}, {:ok, 4}, {:ok, 6}] See the Relay Plugin Guide for usage and details. Chunk Worker Process jobs &quot;broadway style&quot;, in groups based on size or a timeout, but with the robust error handling semantics of Oban. Chunking operates at the worker level, allowing many chunks to run in parallel within the same queue. defmodule MyApp.ChunkWorker do use Oban.Pro.Workers.Chunk , queue : :alpha , size : 10 , timeout : :timer . seconds ( 5 ) @impl Chunk def process ( jobs ) do jobs |&gt; Enum . map ( &amp; &amp;1 . args ) |&gt; Business . process_messages ( ) :ok end end See the Chunk Worker Guide for details on chunk managment, error handling, and more. Pro Worker Testing Module Testing batch, workflows and chunk workers can be tricky due to how they override the perform/1 function. The new Oban.Pro.Testing module provides a process_job/2,3 function that mirrors the functionality of perform_job/2,3 in the base Oban.Testing module. With process_job/2,3 you can test pro workers in isolation, without worrying about whether a manager is running or the state of a workflow. To start using it, import it into your test case or test file: import Oban.Pro.Testing Then test worker logic in isolation: assert :ok = process_job ( MyApp.ProWorker , %{ &quot;id&quot; =&gt; 1 } ) Changed [Oban.Pro.Plugins.WorkflowManager] This plugin is no longer necessary and can be safely removed from your configuration. Bump Oban dependency to ~&gt; 2.6","ref":"pro-changelog.html#v0-7-0-2021-04-02","title":"Changelog - v0.7.0 — 2021-04-02","type":"extras"},{"doc":"Changed [Oban.Pro.Plugins.BatchManager] Drastically speed up batch state checks to prevent timeouts with large batch sizes. The updated query is up to 40x faster under normal workloads. [Oban.Pro.Plugins.DynamicCron] Inject a dynamic cron's name as &quot;cron_name&quot; into inserted job meta to aid in searching, and eventually linking, jobs back to the dynamic cron record. [Oban.Pro.Plugins.DynamicCron] Schedule cron insertion at the top of the next minute, identically to recent changes in Oban.Plugins.Cron . Loosen Oban constraint to allow v2.5.0 Fixed [Oban.Plugins.Reprioritizer] Rename config to conf in event meta so that it matches all other plugins. Specify a minimum of Elixir v1.9 and avoid features from later Elixir versions, namely Keyword.pop! which was added in Elixir v1.10.","ref":"pro-changelog.html#v0-6-2-2021-03-02","title":"Changelog - v0.6.2 — 2021-03-02","type":"extras"},{"doc":"Fixed [Oban.Pro.Plugins.Lifeline] Consider fully qualified node identities when rescuing potential orphans. The previous version could erroneously rescue executing jobs.","ref":"pro-changelog.html#v0-6-1-2021-01-28","title":"Changelog - v0.6.1 — 2021-01-28","type":"extras"},{"doc":"Changed [Oban.Pro.Plugins.Lifeline] Record fully qualified node identities (combining the Oban supervisor name and the host name). This allows Oban Web to track beats recorded by the same host but different instances. Enhance all plugin span metadata with details such as how many jobs were pruned, inserted, reprioritized, etc. Telemetry events and metadata for all Pro plugins now matches the events from Oban OSS. Bump Oban dependency to ~&gt; 2.4.0 due to CRON parser changes and improved telemetry span events. Fixed [Oban.Pro.Plugins.Workflow] Correct typespecs for new_workflow/1 and libgraph related functions. [Oban.Pro.Plugins.DynamicCron] Support inserting and retrieving cron records with a custom prefix. All operations are wrapped in a Multi, which ignored the outer transaction's prefix.","ref":"pro-changelog.html#v0-6-0-2021-01-26","title":"Changelog - v0.6.0 — 2021-01-26","type":"extras"},{"doc":"Fixed [Oban.Pro.Plugins.DynamicPruner] Consider cancelled jobs that were never attempted when pruning.","ref":"pro-changelog.html#v0-5-3-2020-12-09","title":"Changelog - v0.5.3 — 2020-12-09","type":"extras"},{"doc":"Fixed [Oban.Pro.Plugins.BatchManager] Check for the correct output from Worker.from_string , fixing callback insertion.","ref":"pro-changelog.html#v0-5-2-2020-11-27","title":"Changelog - v0.5.2 — 2020-11-27","type":"extras"},{"doc":"[Oban.Pro.Plugins.DynamicPruner] Allow state overrides when pruning cancelled jobs.","ref":"pro-changelog.html#v0-5-1-2020-11-06","title":"Changelog - v0.5.1 — 2020-11-06","type":"extras"},{"doc":"[Oban.Pro.Workers.Workflow] A new worker that composes arbitrary jobs together for directed execution. This allows sequential, fan-out and fan-in workflows that are horizontally distributed and fault tolerant. See the Workflow docs for setup and a walk-through with use-cases. Bump Oban dependency to ~&gt; 2.3.0 due to the workflow worker's use of the new meta field in Oban.Job .","ref":"pro-changelog.html#v0-5-0-2020-11-06","title":"Changelog - v0.5.0 — 2020-11-06","type":"extras"},{"doc":"[Oban.Pro.Plugins.DynamicPruner] Always delete oldest job when pruning in max_len mode. In max_len mode, when job insertion exceeded job pruning the oldest jobs would persist between prunes.","ref":"pro-changelog.html#v0-4-2-2020-10-19","title":"Changelog - v0.4.2 — 2020-10-19","type":"extras"},{"doc":"Changed Upgrade Oban dependency to ~&gt; 2.2.0 along with fixes for the move to Oban.Registry and Oban.Repo .","ref":"pro-changelog.html#v0-4-1-2020-10-11","title":"Changelog - v0.4.1 — 2020-10-11","type":"extras"},{"doc":"Added [Oban.Pro.Plugins.DynamicCron] The DynamicCron plugin is a replacement for Oban's built in cron scheduler. It supports adding, updating, deleting, pausing and resuming periodic jobs at runtime and boot time. It is an ideal solution for applications that must dynamically start and manage scheduled tasks at runtime. See the DynamicCron installation docs for instructions on getting started. [Oban.Pro.Plugins.Lifeline] Add :retry_exhausted option, which will retry exhausted jobs rather than discard them while rescuing. Breaking Change Emit standardized [:oban, :plugin, :start | :stop | :exception] events for all plugins. Previously, each plugin would emit its own event, which required extensive :telemetry.attach_many calls. Now all plugins have a standard event name with differentiated metatada, including a :plugin key with the module name responsible for emitting the event. To upgrade replace any events that used :prune , :lifeline or :reprioritizer with :plugin . For example, [:oban, :prune, :start] is now [:oban, :plugin, :start] with the metadata %{plugin: DynamicPruner} .","ref":"pro-changelog.html#v0-4-0-2020-10-05","title":"Changelog - v0.4.0 — 2020-10-05","type":"extras"},{"doc":"[Oban.Pro.Plugins.DynamicPruner] Correctly prune discarded jobs that were cancelled before they were ever attempted.","ref":"pro-changelog.html#v0-3-2-2020-08-28","title":"Changelog - v0.3.2 — 2020-08-28","type":"extras"},{"doc":"Fixed [Oban.Pro.Workers.Batch] Only invoke batch callbacks that are defined for the current worker.","ref":"pro-changelog.html#v0-3-1-2020-08-05","title":"Changelog - v0.3.1 — 2020-08-05","type":"extras"},{"doc":"Fixed [Oban.Pro.Plugins.Lifeline] Ignore unknown messages to the plugin to prevent crashes. Changed Bump Oban dependency to 2.0.0","ref":"pro-changelog.html#v0-3-0-2020-07-10","title":"Changelog - v0.3.0 — 2020-07-10","type":"extras"},{"doc":"Added Add handle_exhausted/1 callback to Oban.Pro.Workers.Batch worker. This new callback is triggered when all jobs are either completed or discarded .","ref":"pro-changelog.html#v0-2-1-2020-07-01","title":"Changelog - v0.2.1 — 2020-07-01","type":"extras"},{"doc":"Fixed Conditional improvements to the BatchManager plugin Changed Bump Oban dependency to 2.0.0-rc.1","ref":"pro-changelog.html#v0-2-0-2020-06-12","title":"Changelog - v0.2.0 — 2020-06-12","type":"extras"},{"doc":"Initial release including these plugins: Oban.Pro.Plugins.BatchManager Oban.Pro.Plugins.DynamicPruner Oban.Pro.Plugins.Lifeline Oban.Pro.Plugins.Reprioritizer As well as the Oban.Pro.Workers.Batch worker.","ref":"pro-changelog.html#v0-1-0-2020-06-03","title":"Changelog - v0.1.0 — 2020-06-03","type":"extras"},{"doc":"Before continuing, be sure you have Oban up and running in your app! Oban.Pro is delivered as a hex package named oban_pro , which is published privately to our self-hosted package repository. Before you can pull the package into your application you need to add a new oban hex repo. First, grab the OBAN_KEY_FINGERPRINT and OBAN_LICENSE_KEY from your account page. Then, ensure hex is up-to-date: mix local.hex --force Finally, run the following mix hex.repo command: mix hex.repo add oban https://getoban.pro/repo \\ --fetch-public-key $OBAN_KEY_FINGERPRINT \\ --auth-key $OBAN_LICENSE_KEY ⚠️ You'll also need to authenticate on any other development machines, build servers and CI/CD instances. There are also guides to help with building Docker Images , authenticating on Gigalixir and Heroku . Now that you're authenticated you're ready to add oban_pro as a dependency for your application. Open mix.exs and add the following line: { :oban_pro , &quot;~&gt; 0.9&quot; , repo : &quot;oban&quot; } Now fetch your dependencies: $ mix deps.get There isn't any direct configuration for Oban.Pro . Instead, you configure Oban to run an engine or plugins directly and use the various workers. For example, to use the Lifeline plugin you'd add it to the Oban config within config.exs : config :my_app , Oban , engine : Oban.Pro.Queue.SmartEngine , repo : MyApp.Repo , queues : [ alpha : 10 , gamma : 10 , delta : 10 ] , plugins : [ Oban.Pro.Plugins.Lifeline ] Now you're ready to start using the smart engine , various plugins and workers !","ref":"pro_installation.html","title":"Installation","type":"extras"},{"doc":"Building Docker images with Pro and Web requires an additional step to add the private hex repo. Passing the fingerprint and password in as build arguments isn't secure, and you should use buildkit's secret support instead if possible. Reference the secrets within in your Dockerfile by mounting it within a RUN command: RUN --mount=type=secret,id=oban_key_fingerprint \\ --mount=type=secret,id=oban_license_key \\ mix hex.repo add oban https://getoban.pro/repo \\ --fetch-public-key &quot;$(cat /run/secrets/oban_key_fingerprint)&quot; \\ --auth-key &quot;$(cat /run/secrets/oban_license_key)&quot; Then, at build time, pass the fingerprint and license through using the --secret flag: docker build \\ --secret id=oban_key_fingerprint,env=OBAN_KEY_FINGERPRINT \\ --secret id=oban_license_key,env=OBAN_LICENSE_KEY .","ref":"pro_installation.html#building-docker-images","title":"Installation - Building Docker Images","type":"extras"},{"doc":"If your app runs on Gigalixir you can use a buildpack hook to authenticate with the private oban repo. First, set your license key on Gigalixir: gigalixir config:set OBAN_LICENSE_KEY=&quot;YOUR OBAN LICENSE KEY&quot; gigalixir config:set OBAN_KEY_FINGERPRINT=&quot;THE PUBLIC KEY FINGERPRINT&quot; Then add the hook to the end of your elixir_buildpack.config : hook_pre_fetch_dependencies = &quot;mix hex.repo add oban https://getoban.pro/repo \\ --fetch-public-key ${OBAN_KEY_FINGERPRINT} \\ --auth-key ${OBAN_LICENSE_KEY}&quot;","ref":"pro_installation.html#authorizing-on-gigalixir","title":"Installation - Authorizing on Gigalixir","type":"extras"},{"doc":"If your app runs on Heroku using the Elixir Buildpack (rather than Docker) you'll need to use compilation hook to authorize hex before fetching dependencies. First, set your license key on Heroku: heroku config:set OBAN_LICENSE_KEY=&quot;YOUR OBAN LICENSE KEY&quot; heroku config:set OBAN_KEY_FINGERPRINT=&quot;THE PUBLIC KEY FINGERPRINT&quot; Next, add a small shell script to your application in ./bin/predeps : #!/bin/bash mix hex.repo add oban https://getoban.pro/repo \\ --fetch-public-key ${OBAN_KEY_FINGERPRINT} \\ --auth-key ${OBAN_LICENSE_KEY} Finally, set the predeps script within elixir_buildpack.config : hook_pre_fetch_dependencies = &quot;./bin/predeps&quot;","ref":"pro_installation.html#authorizing-on-heroku","title":"Installation - Authorizing on Heroku","type":"extras"},{"doc":"🌟 This engine is available through Oban.Pro The SmartEngine enables truly global concurrency and global rate limiting by handling demand at the queue level, without any conflict or churn. The engine uses centralized producer records to coordinate with minimal load on the database. Engines are callback modules that coordinate how queue producers handle demand and interact the database.","ref":"smart_engine.html","title":"Smart Engine","type":"extras"},{"doc":"The SmartEngine relies on producer records, stored in the oban_producers table. To start, create a migration to create an oban_producers table: $ mix ecto.gen.migration add_oban_producers Within the migration module: use Ecto.Migration defdelegate change , to : Oban.Pro.Migrations.Producers If you have multiple Oban instances or use prefixes, you can specify the prefix and create multiple tables in one migration: use Ecto.Migration def change do Oban.Pro.Migrations.Producers . change ( ) Oban.Pro.Migrations.Producers . change ( prefix : &quot;special&quot; ) Oban.Pro.Migrations.Producers . change ( prefix : &quot;private&quot; ) end The Producers migration also exposes up/0 and down/0 functions if change/0 doesn't fit your usecase. Next, update your config to use the SmartEngine : config :my_app , Oban , engine : Oban.Pro.Queue.SmartEngine , ... If you have multiple Oban instances you'll need to configure each one to use the SmartEngine , otherwise they'll default to the BasicEngine .","ref":"smart_engine.html#installation","title":"Smart Engine - Installation","type":"extras"},{"doc":"With the SmartEngine in place you can define global concurrency limits and rate limits at the queue level. The engine supports the basic queue: limit format as well as the expanded format. Let's look at a few examples: queues : [ alpha : 1 , gamma : [ global_limit : 1 ] , delta : [ local_limit : 2 , global_limit : 5 ] , kappa : [ local_limit : 5 , rate_limit : [ allowed : 30 , period : { 1 , :minute } ] ] , omega : [ global_limit : 1 , rate_limit : [ allowed : 500 , period : { 1 , :hour } ] ] ] local_limit — This limits the number of concurrent jobs that run within a single node. Typically the global concurrency is local_limit * num_nodes . For example, with three nodes and a local limit of 10, you'll have a global limit of 30. global_limit — This limits the number of concurrent jobs that run across all nodes. The queue will process at most the local limit or global limit, whichever is lower. The only way to guarantee that all connected nodes will run exactly one job concurrently is to set global_limit: 1 . rate_limit — This limits the number of jobs that are processed within a sliding window of time, across all nodes. Rate limiting uses counts from all other producer records for the same queue in the cluster. The limiter uses a sliding window over the configured period to accurately approximate a limit. Without a modifier the rate_limit period is defined in seconds. You can provide a :second , :minute , :hour or :day modifier to use more intuitive values. Here are a few examples: period: 30 — 30 seconds period: {1, :minute} — 60 seconds period: {2, :minutes} — 120 seconds period: {1, :hour} — 3,600 seconds period: {1, :day} —86,400 seconds All limits stack and the queue will only fetch the minimum demand.","ref":"smart_engine.html#usage-and-configuration","title":"Smart Engine - Usage and Configuration","type":"extras"},{"doc":"In addition to queue-wide rate limiting, you can also rate-limit by worker and/or args within a queue. This is called &quot;partitioned&quot; rate-limiting, and it's configured through the partition option. For example, to allow one job per-worker, every five seconds, across every instance of the alpha queue in your cluster: queues : [ alpha : [ local_limit : 10 , rate_limit : [ allowed : 1 , period : 5 , partition : [ fields : [ :worker ] ] ] ] ] The following example partitions by the user_id of args instead: rate_limit : [ allowed : 1 , period : 5 , partition : [ fields : [ :args ] , keys : [ :user_id ] ] Here are a few more examples of viable partitioning schemes: fields: [:args] — partition by the entire args map, not a few select keys fields: [:worker, :args] — partition by both worker and all args fields: [:args], keys: [:id, :account_id] — partition by the id and account_ids keys","ref":"smart_engine.html#partitioned-rate-limiting","title":"Smart Engine - Partitioned Rate-Limiting","type":"extras"},{"doc":"The Lifeline plugin cleans up producers that haven't been updated for 1 minute or more. When an event doesn't have any jobs to produce it will automatically &quot;touch&quot; the producer record to prevent deletion. Queue state is stored in a producer record's meta column in the database, where you can easily check values such as the current rate limit counts. Rate limiting uses an efficient sliding window algorithm to safely handle large rate periods, e.g. multiple hours or days. Partitioned rate limiting tracks each partition separately as a &quot;window&quot; within the producer record. Minimize partition cardinality by using keys whenever possible.","ref":"smart_engine.html#implementation-notes","title":"Smart Engine - Implementation Notes","type":"extras"},{"doc":"🌟 This notifier is available through Oban.Pro The PG Notifier is a notifier implementation that runs with Distributed Erlang. Out of the box, Oban uses PostgreSQL's LISTEN/NOTIFY for PubSub. For most applications, that is fine, but Postgres-based PubSub isn't sufficient in some circumstances.","ref":"pg.html","title":"PG Notifier","type":"extras"},{"doc":"Note: You must be using Distributed Erlang to use the PG notifier. Specify the PG notifier in your Oban configuration: config :my_app , Oban , notifier : Oban.Pro.Notifiers.PG , ...","ref":"pg.html#usage-and-configuration","title":"PG Notifier - Usage and Configuration","type":"extras"},{"doc":"The notifier will use pg if available (OTP 23+) or fall back to pg2 for older OTP releases. Like the Postgres implementation, notifications are namespaced by prefix . For compatbility, message payloads are always serialized to JSON before broadcast and deserialized before relay to local processes.","ref":"pg.html#implementation-notes","title":"PG Notifier - Implementation Notes","type":"extras"},{"doc":"🌟 This plugin is available through Oban.Pro The DynamicCron plugin enhances Oban's built in cron scheduler by making it configurable at runtime, globally, across your entire cluster. DynamicCron supports adding, updating, deleting, and pausing cron entries at boot time or runtime. It is an ideal solution for applications that must dynamically start and manage scheduled tasks at runtime.","ref":"dynamic_cron.html","title":"Dynamic Cron Plugin","type":"extras"},{"doc":"Before running the DynamicCron plugin you must run a migration to add the oban_cron table to your database. mix ecto.gen.migration add_oban_cron Open the generated migration in your editor and call the change function on Oban.Pro.Migrations.DynamicCron : defmodule MyApp.Repo.Migrations.AddObanCron do use Ecto.Migration defdelegate change , to : Oban.Pro.Migrations.DynamicCron end As with the base Oban tables you can optionally provide a prefix to &quot;namespace&quot; the table within your database. Here we specify a &quot;private&quot; prefix: defmodule MyApp.Repo.Migrations.AddObanCron do use Ecto.Migration def change , do : Oban.Pro.Migrations.DynamicCron . change ( &quot;private&quot; ) end Run the migration to create the table: mix ecto.migrate Now we can use the DynamicCron plugin and start scheduling periodic jobs!","ref":"dynamic_cron.html#installation","title":"Dynamic Cron Plugin - Installation","type":"extras"},{"doc":"To begin using DynamicCron , add the module to your list of Oban plugins in config.exs : config :my_app , Oban , plugins : [ Oban.Pro.Plugins.DynamicCron ] ... By itself, without providing a crontab or dynamically inserting cron entries, the plugin doesn't have anything to schedule. To get scheduling started, provide a list of {cron, worker} or {cron, worker, options} tuples to the plugin. The syntax is identical to Oban's built in :crontab option, which means you can copying an existing standard :crontab list into the plugin's :crontab . plugins : [ { Oban.Pro.Plugins.DynamicCron , timezone : &quot;America/Chicago&quot; , crontab : [ { &quot;* * * * *&quot; , MyApp.MinuteJob } , { &quot;0 * * * *&quot; , MyApp.HourlyJob , queue : :scheduled } , { &quot;0 0 * * *&quot; , MyApp.DailyJob , max_attempts : 1 } , { &quot;0 12 * * MON&quot; , MyApp.MondayWorker , tags : [ &quot;scheduled&quot; ] } , { &quot;@daily&quot; , MyApp.AnotherDailyWorker } ] } ] For more details about periodic jobs and cron expressions see the documentation on Periodic Jobs . Now, when the dynamic pruner initializes, it will persist those cron entries to the database and start scheduling them according to their CRON expression. The plugin's crontab format is nearly identical to Oban's standard crontab, with a few important enhancements we'll look at soon. Each of the crontab entries are persisted to the database and referenced globally, by all the other connected Oban instances. That allows us to insert, update, or delete cron entries at any time. In fact, changing the schedule or options of an entry in the crontab provided to the plugin will automatically update the persisted entry. To demonstrate, let's modify the MinuteJob we specified so that it runs every other minute in the :scheduled queue: crontab : [ { &quot;*/2 * * * *&quot; , MyApp.MinuteJob , queue : :scheduled } , ... ] Now it isn't really a &quot;minute job&quot; any more, and the name is no longer suitable. However, we didn't provide a name for the entry and it's using the module name instead. To provide more flexibility we can add a :name overrride, then we can update the worker's name as well: crontab : [ { &quot;*/2 * * * *&quot; , MyApp.FrequentJob , name : &quot;frequent&quot; , queue : :scheduled } , ... ] All entries are referenced by name, which defaults to the worker's name and must be unique. You may define the same worker multiple times as long as you provide a name override: crontab : [ { &quot;*/3 * * * *&quot; , MyApp.BasicJob , name : &quot;client-1&quot; , args : %{ client_id : 1 } } , { &quot;*/3 * * * *&quot; , MyApp.BasicJob , name : &quot;client-2&quot; , args : %{ client_id : 2 } } , ... ] To temporarily disable scheduling jobs you can set the paused flag: crontab : [ { &quot;* * * * *&quot; , MyApp.BasicJob , paused : true } , ... ] To resume the job you must supply paused: false (or use update/2 to resume it manually), simply removing the paused option will have no effect. crontab : [ { &quot;* * * * *&quot; , MyApp.BasicJob , paused : false } , ... ] It is also possible to delete a persisted entry during initialization by passing the :delete option: crontab : [ { &quot;* * * * *&quot; , MyApp.MinuteJob , delete : true } , ... ] One or more entries can be deleted this way. Deleting entries is idempotent, nothing will happen if no matching entry can be found. In the next section we'll look at how to list, insert, update and delete jobs dynamically at runtime.","ref":"dynamic_cron.html#using-and-configuring","title":"Dynamic Cron Plugin - Using and Configuring","type":"extras"},{"doc":"Dynamic cron entries are persisted to the database, making it easy to manipulate them through typical CRUD operations. The DynamicCron plugin provides convenience functions to simplify working those operations. In this section we'll walk through each of the available functions and look at some examples. Typespecs 📚 In order to bridge the gap between module level docs and a guide, each section includes a typespec for the corresponding function. The snippet below defines the types listed in each section. @type cron_expr :: binary ( ) @type cron_name :: binary ( ) | atom ( ) @type cron_opt :: { :args , Oban.Job . args ( ) } | { :expression , cron_expr ( ) } | { :max_attempts , pos_integer ( ) } | { :paused , boolean ( ) } | { :priority , 0 .. 3 } | { :name , cron_name ( ) } | { :queue , atom ( ) | binary ( ) } | { :tags , Oban.Job . tags ( ) } | { :timezone , String . t ( ) } @type cron_input :: { cron_expr ( ) , module ( ) } | { cron_expr ( ) , module ( ) , [ cron_opt ] } Inserting Entries @spec insert ( [ cron_input ( ) ] ) :: { :ok , [ CronEntry . t ( ) ] } | { :error , Ecto.Changeset . t ( ) } The insert/1 function takes a list of one or more tuples with the same {expression, worker} or {expression, worker, options} format as the plugin's crontab option: DynamicCron . insert ( [ { &quot;0 0 * * *&quot; , MyApp.GenericWorker } , { &quot;* * * * *&quot; , MyApp.ClientWorker , name : &quot;client-1&quot; , args : %{ client_id : 1 } } , { &quot;* * * * *&quot; , MyApp.ClientWorker , name : &quot;client-2&quot; , args : %{ client_id : 2 } } , { &quot;* * * * *&quot; , MyApp.ClientWorker , name : &quot;client-3&quot; , args : %{ client_id : 3 } } ] ) Be aware that insert/1 acts like an &quot;upsert&quot;, making it possible to modify existing entries if the worker or name matches. Still, it is better to use update/2 to make targeted updates. Updating Entries @spec update ( cron_name ( ) , [ cron_opt ( ) ] ) ) :: { :ok , CronEntry . t ( ) } | { :error , Ecto.Changeset . t ( ) } The update/2 function updates a single cron entry, as identified by the worker or name. Any option available when specifying an entry in the crontab list or when calling insert/2 can be updated—that includes the cron expression and the worker . The following call demonstrates updating every possible option: { :ok , _ } = DynamicCron . update ( &quot;cron-1&quot; , expression : &quot;1 * * * *&quot; , max_attempts : 10 , name : &quot;special-cron&quot; , paused : false , priority : 0 , queue : &quot;dedicated&quot; , tags : [ &quot;client&quot; , &quot;scheduled&quot; ] , worker : Other.Worker ) Naturally, individual options may be updated instead. For example, set paused: true to pause an entry: { :ok , _ } = DynamicCron . update ( MyApp.ClientWorker , paused : true ) Since update/2 operates on a single entry at a time, it is possible to rename an entry without doing a delete / insert dance: # With the worker as the entry name { :ok , _ } = DynamicCron . update ( MyApp.ClientWorker , name : &quot;client-worker&quot; ) # With a custom entry name already set { :ok , _ } = DynamicCron . update ( &quot;cron-1&quot; , name : &quot;special-cron&quot; ) Deleting Entries @spec delete ( cron_name ( ) ) :: { :ok , CronEntry . t ( ) } | { :error , Ecto.Changeset . t ( ) } The delete/1 function operates on individual entries, by worker or name. You can use it to delete entries at runtime, rather than hard-coding the :delete flag into the crontab list at compile time. # With the worker as the entry name { :ok , _ } = DynamicCron . delete ( Worker ) # With a custom name { :ok , _ } = DynamicCron . delete ( &quot;cron-1&quot; ) Listing Entries Use all/0 to retrieve all persisted cron entries: entries = DynamicCron . all ( ) This returns a list of Oban.Pro.Cron schemas with raw attributes. The all/0 function is provided as a convenience to inspect persisted entries. As Oban.Pro.Cron is an Ecto schema you're free to query the table however you wish using Ecto.Query . For example, you can list all of the entries with a name like &quot;client-&quot;: import Ecto.Query , only : [ where : 3 ] Oban.Pro.Cron |&gt; where ( [ c ] , ilike ( c . name , &quot;client-%&quot; ) ) |&gt; MyApp.Repo . all ( ) You can use functions like update or update_all to modify cron jobs in place, but it is highly recommended that you use update/2 to ensure that options are set correctly and to prevent breakage. Overriding the Timezone Without any configuration the default timezone is Etc/UTC . You can override that for all cron entries by passing a timezone option to the plugin: plugins : [ { Oban.Pro.Plugins.DynamicCron , timezone : &quot;America/Chicago&quot; , # ... You can also override the timezone for individual entries by passing it as an option to the crontab list or to DynamicCron.insert/1 : DynamicCron . insert ( [ { &quot;0 0 * * *&quot; , MyApp.Pinger , name : &quot;oslo&quot; , timezone : &quot;Europe/Oslo&quot; } , { &quot;0 0 * * *&quot; , MyApp.Pinger , name : &quot;chicago&quot; , timezone : &quot;America/Chicago&quot; } , { &quot;0 0 * * *&quot; , MyApp.Pinger , name : &quot;zagreb&quot; , timezone : &quot;Europe/Zagreb&quot; } ] )","ref":"dynamic_cron.html#runtime-updates","title":"Dynamic Cron Plugin - Runtime Updates","type":"extras"},{"doc":"All DynamicCron functions have an alternate clause that accepts an Oban instance name as the first argument. This is in line with base Oban functions such as Oban.insert/2 , which allow you to seamlessly work with multiple Oban instances and across multiple database prefixes. For example, you can use list/1 to list all cron entries for the instance named ObanPrivate : entries = DynamicCron . all ( ObanPrivate ) Likewise, to insert a new entry using the configuration associated with the ObanPrivate instance: { :ok , _ } = DynamicCron . insert ( ObanPrivate , [ { &quot;* * * * *&quot; , PrivateWorker } ] )","ref":"dynamic_cron.html#isolation-and-namespacing","title":"Dynamic Cron Plugin - Isolation and Namespacing","type":"extras"},{"doc":"The DynamicCron plugin adds the following metadata to the [:oban, :plugin, :stop] event: :jobs - a list of jobs that were inserted into the database See the docs on Plugin Events for details.","ref":"dynamic_cron.html#instrumenting-with-telemetry","title":"Dynamic Cron Plugin - Instrumenting with Telemetry","type":"extras"},{"doc":"🌟 This plugin is available through Oban.Pro The DynamicPruner plugin enhances the default FixedPruner plugin's behaviour by allowing you to customize how long or how many jobs are retained in the jobs table. Where the FixedPruner will only retain jobs for 60 seconds, with the DynamicPruner you can specify either a maximum age or a maximum length and provide custom rules for specific queues, workers and job states.","ref":"dynamic_pruner.html","title":"Dynamic Pruner Plugin","type":"extras"},{"doc":"To start using the DynamicPruner add the module to your list of Oban plugins in config.exs : config :my_app , Oban , plugins : [ Oban.Pro.Plugins.DynamicPruner ] ... Without any additional options the pruner operates in maximum length mode ( max_len ) and retains a conservative 1,000 completed or discarded jobs. To increase the number of jobs retained you can provide your own mode configuration: plugins : [ { Oban.Pro.Plugins.DynamicPruner , mode : { :max_len , 50_000 } } ] Now the pruner will retain the most recent 50,000 jobs instead. A fixed limit on the number of jobs isn't always ideal. Often you want to retain jobs based on their age instead. For example, if you your application needs to ensure that a duplicate job hasn't been enqueued within the past 24 hours you need to retain jobs for at least 24 hours; a fixed limit simply won't work. For that we can use maximum age ( max_age ) mode instead: plugins : [ { Oban.Pro.Plugins.DynamicPruner , mode : { :max_age , 60 * 60 * 48 } } ] Here we've specified max_age using seconds, where 60 * 60 * 48 is the number of seconds in two days. Calculating the number of seconds in a period isn't especially readable, particularly when you have numerous max_age declarations in overrides (see below). For clarity you can specify the age's time unit as :second , :minute , :hour , :day or :month . Here is the same 48 hour configuration from above, but specified in terms of days: plugins : [ { Oban.Pro.Plugins.DynamicPruner , mode : { :max_age , { 2 , :days } } } ] Now you can tell exactly how long jobs should be retained, without reverse calculating how many seconds an expression represents.","ref":"dynamic_pruner.html#using-and-configuring","title":"Dynamic Pruner Plugin - Using and Configuring","type":"extras"},{"doc":"The mode option is indiscriminate when determining which jobs to prune. It pays no attention to which queue they are in, what worker the job is for, or which state they landed in. The DynamicPruner allows you to specify per-queue, per-worker and per-state overrides that fine tune pruning. We'll start with a simple example of limiting the total number of retained jobs in the events queue: plugins : [ { Oban.Pro.Plugins.DynamicPruner , mode : { :max_age , { 7 , :days } } , queue_overrides : [ events : { max_len : 1_000 } ] } ] With this configuration most jobs will be retained for seven days, but we'll only keep the latest 1,000 jobs in the events queue. We can extend this further and override all of our queues (and omit the default mode entirely): plugins : [ { Oban.Pro.Plugins.DynamicPruner , queue_overrides : [ default : { :max_age , { 6 , :hours } } , analysis : { :max_age , { 1 , :day } } , events : { :max_age , { 10 , :minutes } } , mailers : { :max_age , { 2 , :weeks } } , media : { :max_age , { 2 , :months } } ] } ] When pruning by queue isn't granular enough you can provide overrides by worker instead: plugins : [ { Oban.Pro.Plugins.DynamicPruner , worker_overrides : [ &quot;MyApp.BusyWorker&quot; : { :max_age , { 1 , :day } } , &quot;MyApp.SecretWorker&quot; : { :max_age , { 1 , :second } } , &quot;MyApp.HistoricWorker&quot; : { :max_age , { 1 , :month } } ] } ] You can also override by state, which allows you to keep discarded jobs for inspection while quickly purging cancelled or successfully completed jobs: plugins : [ { Oban.Pro.Plugins.DynamicPruner , state_overrides : [ cancelled : { :max_age , { 1 , :hour } } , completed : { :max_age , { 1 , :day } } , discarded : { :max_age , { 1 , :month } } ] } ] Naturally you can mix and match overrides to finely control job retention: plugins : [ { Oban.Pro.Plugins.DynamicPruner , mode : { :max_age , { 7 , :days } } , queue_overrides : [ events : { :max_age , { 10 , :minutes } } ] , worker_overrides : [ &quot;MyApp.SecretWorker&quot; : { :max_age , { 1 , :second } } ] , state_overrides : [ discarded : { :max_age , { 2 , :days } } ] } ]","ref":"dynamic_pruner.html#providing-overrides","title":"Dynamic Pruner Plugin - Providing Overrides","type":"extras"},{"doc":"Overrides are applied sequentially, in this order: Queue State Worker Default Using the example above, jobs in the events queue are deleted first, followed by jobs in the discarded state, then the MyApp.SecretWorker worker, and finally any other jobs older than 7 days that weren't covered by any overrides .","ref":"dynamic_pruner.html#override-precedence","title":"Dynamic Pruner Plugin - Override Precedence","type":"extras"},{"doc":"With the default settings the DynamicPruner will only delete 10,000 jobs each time it prunes. The limit exists to prevent connection timeouts and excessive table locks. A busy system can easily insert more than 10,000 jobs per minute during standard operation. If you find that jobs are accumulating despite active pruning you can override the limit . Here we set the delete limit to 25,000: plugins : [ { Oban.Pro.Plugins.DynamicPruner , mode : { :max_len , 100_000 } , limit : 25_000 } ] Deleting in PostgreSQL is very fast, and the 10k default is rather conservative. Feel free to increase the limit to a number that your system can handle.","ref":"dynamic_pruner.html#keeping-up-with-inserts","title":"Dynamic Pruner Plugin - Keeping Up With Inserts","type":"extras"},{"doc":"Some additional notes about pruning in general and nuances of the DynamicPruner plugin: Pruning is best-effort and performed out-of-band. This means that all limits are soft; jobs beyond a specified age may not be pruned immediately after jobs complete. Pruning is only applied to jobs that are completed , cancelled or discarded (has reached the maximum number of retries or has been manually killed). It'll never delete a new, scheduled, or retryable job. Only a single node will prune at any given time, which prevents potential deadlocks between transactions.","ref":"dynamic_pruner.html#implementation-notes","title":"Dynamic Pruner Plugin - Implementation Notes","type":"extras"},{"doc":"The DynamicPruner plugin adds the following metadata to the [:oban, :plugin, :stop] event: :pruned_count - the total jobs that were pruned See the docs on Plugin Events for details.","ref":"dynamic_pruner.html#instrumenting-with-telemetry","title":"Dynamic Pruner Plugin - Instrumenting with Telemetry","type":"extras"},{"doc":"🌟 This plugin is available through Oban.Pro The Lifeline plugin uses producer records to periodically rescues orphaned jobs, i.e. jobs that are stuck in the executing state because the node was shut down before the job could finish. Without the Lifeline plugin you may need to manually rescue jobs stuck in the executing state.","ref":"lifeline.html","title":"Lifeline Plugin","type":"extras"},{"doc":"To use the Lifeline plugin add the module to your list of Oban plugins in config.exs : config :my_app , Oban , plugins : [ Oban.Pro.Plugins.Lifeline ] ... There isn't any configuration necessary. By default, the plugin will delete outdated producer records and rescue orphaned jobs every 1 minute. If necessary you can configure the rescue interval: plugins : [ { Oban.Pro.Plugins.Lifeline , rescue_interval : :timer . minutes ( 5 ) } ] Note that rescuing orphans relies on producer records as used by the SmartEngine .","ref":"lifeline.html#using-and-configuring","title":"Lifeline Plugin - Using and Configuring","type":"extras"},{"doc":"When a job's attempt matches its max_attempts its retries are considered &quot;exhausted&quot;. Normally, the Lifeline plugin transitions exhausted jobs to the discarded state and they won't be retried again. It does this for a couple of reasons: To ensure at-most-once semantics. If a long running job interacted with a non idempotent service and was shut down while waiting for a reply you may not want that jot to retry. To prevent infinitely crashing BEAM nodes. Poorly behaving jobs may crash the node (through NIFs, memory exhaustion, etc.) We don't want to repeatedly rescue and rerun a job that repeatedly crashes the entire node. Discarding exhausted jobs may not always be desired. Use the retry_exhausted option if you'd prefer to retry exhausted jobs when they are rescued, rather than discarding them: plugins : [ { Oban.Pro.Plugins.Lifeline , retry_exhausted : true } ] During rescues, with retry_exhausted: true , a job's max_attempts is incremented and it is moved back to the available state.","ref":"lifeline.html#rescuing-exhausted-jobs","title":"Lifeline Plugin - Rescuing Exhausted Jobs","type":"extras"},{"doc":"Some additional notes about how Lifeline operates: Orphan rescuing is guaranteed to only rescue jobs that belong to dead queue processes or nodes. Only a single node will rescue orphans at any given time, which prevents potential deadlocks and churn.","ref":"lifeline.html#implementation-notes","title":"Lifeline Plugin - Implementation Notes","type":"extras"},{"doc":"The Lifeline plugin adds the following metadata to the [:oban, :plugin, :stop] event: :action — :rescue . :deleted_count — the number of producers deleted :rescued_count — the number of jobs rescued See the docs on Plugin Events for details.","ref":"lifeline.html#instrumenting-with-telemetry","title":"Lifeline Plugin - Instrumenting with Telemetry","type":"extras"},{"doc":"🌟 This plugin is available through Oban.Pro The Relay plugin lets you insert and await the results of jobs locally or remotely, across any number of nodes, i.e. persistent distributed tasks. Once the plugin is running, you can seamlessly distribute jobs and await the results synchronously.","ref":"relay.html","title":"Relay Plugin","type":"extras"},{"doc":"The Relay plugin must be started to coordinate events and messages for async/await functions. Start by adding it to your plugins: plugins : [ Oban.Pro.Plugins.Relay ] There isn't any extra configuration necessary (or possible).","ref":"relay.html#installation","title":"Relay Plugin - Installation","type":"extras"},{"doc":"Typespecs 📚 In order to bridge the gap between module level docs and a guide, each section includes a typespec for the corresponding function. The snippet below defines the types listed in each section. @type t :: %{ job : Job . t ( ) , pid : pid ( ) , ref : UUID . t ( ) } @type await_result :: :ok | { :ok , term ( ) } | { :error , :discarded } | { :error , :result_too_large } | { :error , :snoozed } | { :error , :timeout } | { :error , Exception . t ( ) } Async @spec async ( Job . changeset ( ) ) :: t ( ) | { :error , Job . changeset ( ) } @spec async ( Oban . name ( ) , Job . changeset ( ) ) :: t ( ) | { :error , Job . changeset ( ) } Use Relay.async/1,2 to insert a job for asynchronous execution. The single arity version takes a job changeset and inserts it: relay = %{ id : 123 } |&gt; MyApp.Worker . new ( ) |&gt; Relay . async ( ) When the Oban instance has a custom name, or an app has multiple Oban instances, you can use the two arity version to select an instance: changeset = MyApp.Worker . new ( %{ id : 123 } ) relay = Relay . async ( MyOban , changeset ) The returned map contains the caller's pid and a unique ref that is used to await the results. Await @spec await ( relay :: t ( ) ) :: await_result ( ) @spec await ( relay :: t ( ) , timeout ( ) ) :: await_result ( ) After inserting a job and constructing a relay, use Relay.await/1,2 to await the job's execution and return the result: { :ok , result } = %{ :ok , 4 } |&gt; MyApp.Worker . new ( ) |&gt; Relay . async ( ) |&gt; Relay . await ( ) By default, await will timeout after 5 seconds and return an {:error, :timeout} tuple. The job itself may continue to run, only the local process stops waiting on it. The second argument to await/2 is a timeout value: Relay . await ( relay , :timer . seconds ( 30 ) ) When the executed job fails, crashes, snoozes or is discarded the result comes back as an error tuple with :discarded , :snoozed , or the exception. Note that results are encoded using term_to_binary/2 with compression and then decoded locally using binary_to_term/2 with safety enabled to prevent leaking atoms. Any value smaller than 8kb when compressed is can be returned. Values larger than 8kb will return {:error, :result_too_large} tuple. Await Many @spec await_many ( relays :: [ t ( ) ] ) :: [ await_result ( ) | nil ] @spec await_many ( relays :: [ t ( ) ] , timeout ( ) ) :: [ await_result ( ) | nil ] Use Relay.await_many/1,2 to await replies from multiple relays and return the results. It returns a list of the results in the same order as the relays supplied as the first argument. relayed = 1 .. 3 |&gt; Enum . map ( &amp; DoubleWorker . new ( %{ int : &amp;1 } ) ) |&gt; Enum . map ( &amp; Relay . async ( &amp;1 ) ) |&gt; Relay . await_many ( ) # [{:ok, 2}, {:ok, 4}, {:ok, 6}] Unlike Task.await_many or Task.yield_many , Relay.await_many may return partial results when the timeout is reached. When a job hasn't finished executing the value will be a timeout error tuple. relayed = [ 1 , 2 , 300_000_000 ] |&gt; Enum . map ( &amp; SlowWorker . new ( %{ int : &amp;1 } ) ) |&gt; Enum . map ( &amp; Relay . async ( &amp;1 ) ) |&gt; Relay . await_many ( 100 ) # [{:ok, 2}, {:ok, 4}, {:error, :timeout}]","ref":"relay.html#usage","title":"Relay Plugin - Usage","type":"extras"},{"doc":"Some additional notes and requirements to be aware of: Relay uses PubSub for to transmit results. That means it will work without Erlang distribution or clustering, but it does require functional PostgreSQL based PubSub. It doesn't matter where a job is executed, the result will still be broadcast back provided that the Oban instance which processed it is running the Relay plugin as well. Results are encoded using :erlang.term_to_binary/2 and decoded using :erlang.binary_to_term/2 using the :safe option to prevent the creation of new atoms or function references. If you're returning results with atoms you must be sure that those atoms are defined locally, where the await/2 or await_many/2 function is called.","ref":"relay.html#implementation-notes","title":"Relay Plugin - Implementation Notes","type":"extras"},{"doc":"🌟 This plugin is available through Oban.Pro Using mixed priorities in a queue ensures that certain jobs are executed before others. For example, a queue that processes jobs from various customers may prioritize customers that are in a higher tier or plan. All high priority ( 0 ) jobs are guaranteed to run before any with lower priority ( 1..3 ), which is wonderful for the higher tier customers but can lead to resource starvation. When there is a constant flow of high priority jobs the lower priority jobs will never get the chance to run. The Reprioritizer plugin automatically adjusts lower job's priorities so that all jobs are eventually processed.","ref":"reprioritizer.html","title":"Reprioritizer Plugin","type":"extras"},{"doc":"To use the Reprioritizer plugin add the module to your list of Oban plugins in config.exs : config :my_app , Oban , plugins : [ Oban.Pro.Plugins.Reprioritizer ] ... Without any additional options the plugin will automatically increase the priority of any jobs that are available for 5 minutes or more. To reprioritize after less time waiting you can configure the :after time: plugins : [ { Oban.Pro.Plugins.Reprioritizer , after : :timer . minutes ( 2 ) } ] Now lower job priorities are bumped after 2 minutes of waiting, and every minute after that. To lower the reprioritization frequency you can change the :interval along with the :after time: plugins : [ { Oban.Pro.Plugins.Reprioritizer , after : :timer . minutes ( 10 ) , interval : :timer . minutes ( 5 ) } ] Here we've specified that job priority will increase every 5 minutes after the first 10 minutes of waiting.","ref":"reprioritizer.html#using-and-configuring","title":"Reprioritizer Plugin - Using and Configuring","type":"extras"},{"doc":"The after option applies to jobs for any workers across all queues. The Reprioritizer plugin allows you to specify per-queue and per-worker overrides that fine tune reprioritization. Let's configure reprioritization for the analysis queue so that it nudges jobs after only 1 minute: plugins : [ { Oban.Pro.Plugins.Reprioritizer , queue_overrides : [ analysis : :timer . minutes ( 1 ) ] } ] We can also effectively disable reprioritization for all other queues by setting the period to :infinity : plugins : [ { Oban.Pro.Plugins.Reprioritizer , after : :infinity , queue_overrides : [ analysis : :timer . minutes ( 1 ) ] } ] If per-queue overrides aren't enough we can override on a per-worker basis instead: plugins : [ { Oban.Pro.Plugins.Reprioritizer , interval : :timer . seconds ( 15 ) , worker_overrides : [ &quot;MyApp.HighSLAWorker&quot; : :timer . seconds ( 30 ) , &quot;MyApp.LowSLAWorker&quot; : :timer . minutes ( 10 ) ] } ] Naturally you can mix and match overrides to finely control reprioritization: plugins : [ { Oban.Pro.Plugins.Reprioritizer , interval : :timer . minutes ( 2 ) , after : :timer . minutes ( 5 ) , queue_overrides : [ media : :timer . minutes ( 10 ) ] , worker_overrides : [ &quot;MyApp.HighSLAWorker&quot; : :timer . seconds ( 30 ) ] } ]","ref":"reprioritizer.html#providing-overrides","title":"Reprioritizer Plugin - Providing Overrides","type":"extras"},{"doc":"The Reprioritizer plugin adds the following metadata to the [:oban, :plugin, :stop] event: :reprioritized_count — the number of jobs reprioritized See the docs on Plugin Events for details.","ref":"reprioritizer.html#instrumenting-with-telemetry","title":"Reprioritizer Plugin - Instrumenting with Telemetry","type":"extras"},{"doc":"🌟 This worker is available through Oban.Pro A Batch worker links the execution of many jobs as a group and runs optional callbacks after jobs are processed. This allows your application to coordinate the execution of tens, hundreds or thousands of jobs in parallel. It is built as an ergonomic abstraction over the top of the standard Oban.Worker .","ref":"batch.html","title":"Batch Worker","type":"extras"},{"doc":"While Batch workers are built within your application, they all rely on the BatchManager plugin from Oban.Pro . The BatchManager is responsible for tracking the execution of jobs within a batch and reliably enqueuing callback jobs. Start by adding BatchManager to your list of Oban plugins in config.exs : config :my_app , Oban , plugins : [ Oban.Pro.Plugins.BatchManager ] With the BatchManager plugin set to run we can create our batch workers. Let's define a worker that delivers daily emails in a large batch: defmodule MyApp.EmailBatch do use Oban.Pro.Workers.Batch , queue : :mailers @impl true def process ( % Job { args : %{ &quot;email&quot; =&gt; email } } ) do MyApp.Mailer . daily_update ( email ) end end Note that we define a process/1 callback instead of perform/1 because perform/1 is used behind the scenes to coordinate regular execution and callbacks within the same worker. The process/1 function receives an Oban.Job struct, just like perform/1 would and it should return the same accepted values, i.e. :ok , {:ok, value} , {:error, error} . The process/1 function above only looks for an &quot;email&quot; key in the job args, but a &quot;batch_id&quot; is also available in meta . We'll modify the function to extract the batch_id as well: def process ( % Job { args : %{ &quot;email&quot; =&gt; email } , meta : %{ &quot;batch_id&quot; =&gt; batch_id } } ) do { :ok , reply } = MyApp.Mailer . daily_update ( email ) track_delivery ( batch_id , reply ) :ok end Now the hypothetical track_delivery/2 function will store the delivery details for retrieval later, possibly by one of our handler callbacks . Typespecs 📚 In order to bridge the gap between module level docs and a guide, each section includes a typespec for the corresponding function. The snippet below defines the types listed in each section. @type args_or_jobs :: [ Job . t ( ) | Job . args ( ) ] @type batch_opts :: Job . option ( ) | { :batch_id , String . t ( ) } | { :batch_callback_args , map ( ) } | { :batch_callback_worker , module ( ) }","ref":"batch.html#usage","title":"Batch Worker - Usage","type":"extras"},{"doc":"@callback new_batch ( args_or_jobs ( ) , [ batch_opts ( ) ] ) :: [ Changeset . t ( ) ] Create batches with new_batch/1,2 by passing a list of args and options, or a list of heterogeneous jobs. A list of args and options is transformed into a list of jobs for that batch module. For example, this will build and insert two EmailBatch jobs: [ %{ email : &quot;foo@example.com&quot; } , %{ email : &quot;bar@example.com&quot; } ] |&gt; MyApp.EmailBatch . new_batch ( ) |&gt; Oban . insert_all ( ) To schedule a batch in the future, or override default options, you pass a list of options: [ %{ email : &quot;foo@example.com&quot; } , %{ email : &quot;bar@example.com&quot; } ] |&gt; MyApp.EmailBatch . new_batch ( schedule_in : 60 , priority : 1 , max_attempts : 3 ) |&gt; Oban . insert_all ( ) The new_batch/1,2 function automatically injects a unique batch_id into each job's meta. A Batch worker is a regular Oban.Worker under the hood, which means you can use new/2 to insert jobs as well, provided you use a deterministic batch id . Creating a heterogeneous batch is similar, though you may want to provide an explicit worker for [callbacks][#handler-callbacks]: mail_jobs = Enum . map ( mail_args , &amp; MyApp.MailWorker . new / 1 ) push_jobs = Enum . map ( push_args , &amp; MyApp.PushWorker . new / 1 ) MyApp.BatchWorker . new_batch ( mail_jobs ++ push_jobs )","ref":"batch.html#inserting-batches","title":"Batch Worker - Inserting Batches","type":"extras"},{"doc":"@callback gen_id ( ) :: String . t ( ) By default a batch_id is generated as a version 4 random UUID . UUIDs are more than sufficient to ensure that batches are unique between workers and nodes for any period. However, if you require control, you can override batch_id generation at the worker level or pass a value directly to the new_batch/2 function. To override the batch_id for a particular worker you override the gen_id callback: defmodule MyApp.BatchWorker do use Oban.Pro.Workers.Batch # Generate a 24 character long random string instead @impl Batch def gen_id do 24 |&gt; :crypto . strong_rand_bytes ( ) |&gt; Base . encode64 ( ) end end The gen_id/0 callback is suited for random/non-deterministic id generation. If you'd prefer to use a deterministic id instead you can pass the batch_id in as an option to new_batch/2 : MyApp.BatchWorker . new_batch ( list_of_args , batch_id : &quot;custom-batch-id&quot; ) Using this technique you can verify the batch_id in tests or append to the batch manually after it was originally created. For example, you can add to a batch that is scheduled for the future: batch_id = &quot;daily-batch- \#{ Date . utc_today ( ) } &quot; midnight = Date . utc_today ( ) |&gt; NaiveDateTime . new ( ~T[11:59:59] ) |&gt; elem ( 1 ) |&gt; DateTime . from_naive! ( &quot;Etc/UTC&quot; ) # Create the initial batch initial_args |&gt; MyApp.BatchWorker . new_batch ( batch_id : batch_id , schedule_at : midnight ) |&gt; Oban . insert_all ( ) # Add items to the batch later in the day %{ batch_id : batch_id , other_arg : &quot;other&quot; } |&gt; MyApp.BatchWorker . new ( schedule_at : midnight ) |&gt; Oban . insert ( ) When batch jobs execute at midnight they'll all be tracked together.","ref":"batch.html#generating-batch-ids","title":"Batch Worker - Generating Batch IDs","type":"extras"},{"doc":"@callback handle_attempted ( job :: Job . t ( ) ) :: :ok @callback handle_completed ( job :: Job . t ( ) ) :: :ok @callback handle_discarded ( job :: Job . t ( ) ) :: :ok @callback handle_exhausted ( job :: Job . t ( ) ) :: :ok After jobs in the batch are processed the BatchManager may insert callback jobs for a callback worker. There are four optional batch handler callbacks that a worker may define: handle_attempted — called after all jobs in the batch were attempted at least once, regardless of whether they succeeded or not. handle_completed — called after all jobs in the batch have a completed state. This handler may never be called if one or more jobs keep failing or any are discarded. handle_discarded — called after any jobs in the batch have a discarded state. handle_exhausted — called after all jobs in the batch have either a completed or discarded state. Each handler callback receives an Oban.Job struct with the batch_id in meta and should return :ok . The callbacks are executed as separate isolated jobs, so they may be retried or discarded like any other job. Here we'll implement each of the optional handler callbacks and have them print out the batch status along with the batch_id : defmodule MyApp.BatchWorker do use Oban.Pro.Workers.Batch @impl Batch def handle_attempted ( % Job { meta : %{ &quot;batch_id&quot; =&gt; batch_id } } ) do IO . puts ( &quot;Attempted \#{ batch_id } &quot; ) end @impl Batch def handle_completed ( % Job { meta : %{ &quot;batch_id&quot; =&gt; batch_id } } ) do IO . puts ( &quot;Completed \#{ batch_id } &quot; ) end @impl Batch def handle_discarded ( % Job { meta : %{ &quot;batch_id&quot; =&gt; batch_id } } ) do IO . puts ( &quot;Discarded \#{ batch_id } &quot; ) end @impl Batch def handle_exhausted ( % Job { meta : %{ &quot;batch_id&quot; =&gt; batch_id } } ) do IO . puts ( &quot;Exhausted \#{ batch_id } &quot; ) end end Forwarding Callback Args By default, callback jobs have an empty args map. With the :batch_callback_args option to new_batch/2 , you can pass custom args through to each callback. For example, here we're passing a webhook URLs in the args for use callbacks: MyBatch . new_batch ( jobs , batch_callback_args : %{ webhook : &quot;https://web.hook&quot; } ) Any JSON encodable map may be passed to callbacks, but note that the complete map is stored in each batch job's meta. Alternate Callback Workers For some batches, notably those with heterogeneous jobs, it's handy to specify a different worker for callbacks. That is easily accomplished by passing the :batch_callback_worker option to new_batch/2 : MyBatch . new ( jobs , batch_callback_worker : MyCallbackWorker ) The callback worker must be an Oban.Worker that defines one or more of the batch callback handlers.","ref":"batch.html#handler-callbacks","title":"Batch Worker - Handler Callbacks","type":"extras"},{"doc":"@callback stream_batch_jobs ( Job . t ( ) , Keyword . t ( ) ) :: Enum . t ( ) For map/reduce style workflows, or to pull more context from batch jobs, it's possible to load all jobs from the batch with stream_batch_jobs/1,2 . The function takes a single batch job and returns a stream of all non-callback jobs in the batch, which you can then operate on with Enum or Stream functions. As an example, imagine you have a batch that ran for a few thousand accounts and you'd like to notify admins that the batch is complete. defmodule MyApp.BatchWorker do use Oban.Pro.Workers.Batch @impl Batch def handle_completed ( % Job { } = job ) do { :ok , account_ids } = MyApp.Repo . transaction ( fn -&gt; job |&gt; stream_batch_jobs ( ) |&gt; Enum . map ( &amp; &amp;1 . args [ &quot;account_id&quot; ] ) end ) account_ids |&gt; MyApp.Accounts . all ( ) |&gt; MyApp.Mailer . notify_admins_about_batch ( ) end Streaming is provided by Ecto's Repo.stream , and it must take place within a transaction. While it may be overkill for small batches, for batches with tens or hundreds of thousands of jobs, it will prevent massive memory spikes or the database grinding to a halt!","ref":"batch.html#streaming-batch-jobs","title":"Batch Worker - Streaming Batch Jobs","type":"extras"},{"doc":"PostgreSQL's binary protocol has a limit of 65,535 parameters that may be sent in a single call. That presents an upper limit on the number of rows that may be inserted at one time, and therefor the number of jobs that may be inserted in a batch all at once. There is a simple workaround that will allow you to create arbitrarily large batches by inserting them in chunks. reducer = fn { changesets , index } , multi -&gt; Oban . insert_all ( multi , &quot;batch_ \#{ index } &quot; , changesets ) end list_of_args |&gt; MyApp.BatchWorker . new_batch ( ) |&gt; Enum . chunk_every ( 5_000 ) |&gt; Enum . with_index ( ) |&gt; Enum . reduce ( Ecto.Multi . new ( ) , reducer ) |&gt; MyApp.Repo . transaction ( ) A few words of explanation: This chunks every 5,000 jobs, you can use a smaller number but anything under 65,000 should work. The Enum.with_index/1 call is necessary to provide a unique name for each multi, without that you'll get an error about the batch names conflicting. The jobs are all inserted together within a transaction and won't start executing until they all go in.","ref":"batch.html#inserting-large-batches","title":"Batch Worker - Inserting Large Batches","type":"extras"},{"doc":"Callback jobs are only enqueued if your worker defines the corresponding callback, e.g. a worker that only defines handle_attempted/1 will only have a callback for that event. Callback jobs are unique, with an infinite period. The BatchManager uses debouncing to minimize queries and reduce overall load on your database.","ref":"batch.html#implementation-notes","title":"Batch Worker - Implementation Notes","type":"extras"},{"doc":"🌟 This worker is available through Oban.Pro A Chunk worker processes jobs &quot;broadway style&quot;, in groups based on size or a timeout, but with the robust error handling semantics of Oban. Chunking operates at the worker level, allowing many chunks to run in parallel within the same queue.","ref":"chunk.html","title":"Chunk Worker","type":"extras"},{"doc":"To declare a Chunk worker you use Oban.Pro.Workers.Chunk rather than the basic Oban.Worker . Let's define a worker that sends SMS messages in chunks, rather than individually: defmodule MyApp.ChunkWorker do use Oban.Pro.Workers.Chunk , queue : :messages , size : 100 , timeout : 1000 @impl true def process ( [ _ | _ ] = jobs ) do jobs |&gt; Enum . map ( &amp; &amp;1 . args ) |&gt; MyApp.Messaging . send_batch ( ) :ok end end Notice that we declared a size and a timeout along with a queue for the worker. If size or timeout are omitted they fall back to their defaults, 100 for size and 1000ms for timeout . To process larger batches less frequently, we can increase both values: use Oban.Pro.Workers.Chunk , size : 500 , timeout : :timer . seconds ( 5 ) Now chunks will run with up to 500 jobs or every 5 seconds, whichever comes first. Like other Pro workers, we define a process/1 callback rather than perform/1 . The Chunk worker's process/1 is a little different, as it receives a list of jobs rather than a single struct. Fittingly, the expected return values are different as well.","ref":"chunk.html#usage","title":"Chunk Worker - Usage","type":"extras"},{"doc":"Chunk worker's result type is tailored to handling multiple jobs at once. For reference, here are the types and callback definition for process/1 : @type jobs :: [ Job . t ( ) ] @type reason :: term ( ) @type result :: :ok | { :ok , term ( ) } | { :error , reason ( ) , jobs ( ) } | { :discard , reason ( ) , jobs ( ) } | [ { :error , { reason ( ) , jobs ( ) } } | { :discard , { reason ( ) , jobs ( ) } } ] @callback process ( jobs ( ) ) :: result ( ) The result types allow you to succeed an entire chunk or selectively fail parts of it. Here are each of the result options explained: :ok — The chunk succeeded and all jobs can be marked complete {:ok, result} — Like :ok , but with a result for testing {:error, reason, jobs} — One or more jobs in the chunk failed and may be retried, any unlisted jobs are a success {:discard, reason, jobs} — One or more jobs in the chunk should be discarded, any unlisted jobs are a success [{:error, {reason, jobs}} | {:discard, {reason, jobs}}] — Retry some jobs and discard some other jobs, leaving any jobs not in either list a success To illustrate using chunk results let's expand on the message processing example from earlier. We'll extend it to complete the whole batch when all messages are delivered or discard undeliverable messages: def process ( [ _ | _ ] = jobs ) do notifications = jobs |&gt; Enum . map ( &amp; &amp;1 . args ) |&gt; MyApp.Messaging . send_batch ( ) bad_token = fn %{ response : response } -&gt; response == :bad_token end if Enum . any? ( notifications , bad_token ) do discards = notifications |&gt; Enum . zip ( jobs ) |&gt; Enum . filter ( fn { notification , _job } -&gt; bad_token . ( notification ) end ) |&gt; Enum . map ( &amp; elem ( &amp;1 , 1 ) ) { :discard , :bad_token , discards } else { :ok , notifications } end end In the event of an ephemeral crash, like a network issue, the entire batch may be retried if there are additional attempts.","ref":"chunk.html#chunk-results-and-error-handling","title":"Chunk Worker - Chunk Results and Error Handling","type":"extras"},{"doc":"Chunks are ran by a leader job. When the leader executes it determines whether a complete chunk is available or if enough time has elapsed to run anyhow. If neither case applies then the leader will delay until the timeout elapsed and execute with as many jobs as it can find. Only the leader job may be cancelled as it is the only one tracked by a producer. Cancelling any other jobs in the chunk won't stop the chunk from running. Jobs are processed as a chunk and are tracked individually. If a single job keeps erroring between batches it will transition to discarded by itself.","ref":"chunk.html#implementation-notes","title":"Chunk Worker - Implementation Notes","type":"extras"},{"doc":"🌟 This worker is available through Oban.Pro Workflow workers compose together with arbitrary dependencies between jobs, allowing sequential, fan-out and fan-in execution workflows. Workflows are fault tolerant, can be homogeneous or heterogeneous, and scale horizontally across all available nodes. Workflows are unique to Oban—only possible due to historic job retention.","ref":"workflow.html","title":"Workflow Worker","type":"extras"},{"doc":"You build Workflow workers within your application. Each worker within a workflow must use Oban.Pro.Workers.Workflow rather than Oban.Worker , otherwise jobs will execute without any dependency management. Our example workflow will coordinate archiving an account. To properly archive the account we must perform the following workers: FinalReceipt — generate a final receipt and email it to the account owner EmailSubscriber — email each subscriber on the account BackupPost — up all important posts on the account to external storage DeleteAccount — delete the account from the database We need step 1 to run before step 2 or 3, but both steps 2 and 3 can run in parallel. After all of the jobs in steps 2 and 3 finish we can finally delete the account from the database. For demonstration purposes we'll define one of the workers with a stubbed-in process/1 function: defmodule MyApp.FinalReceipt do use Oban.Pro.Workers.Workflow , queue : :default @impl true def process ( % Job { } = _job ) do # Generate the final receipt. :ok end end Note that we define a process/1 callback instead of perform/1 because the perform/1 function coordinates the workflow. The process/1 function receives an Oban.Job struct, just like perform/1 and it should return similar values, i.e. :ok , {:ok, value} . Assuming that you've defined the other workers, we can construct a workflow to enforce sequential execution: alias MyApp . { BackupPost , DeleteAccount , EmailSubscriber , FinalReceipt } def archive_account ( acc_id ) do FinalReceipt . new_workflow ( ) |&gt; FinalReceipt . add ( :receipt , FinalReceipt . new ( %{ id : acc_id } ) ) |&gt; FinalReceipt . add ( :email , EmailSubscriber . new ( %{ id : acc_id } ) , deps : [ :receipt ] ) |&gt; FinalReceipt . add ( :backup , BackupPost . new ( %{ id : acc_id } ) , deps : [ :backup ] ) |&gt; FinalReceipt . add ( :delete , DeleteAccount . new ( %{ id : acc_id } ) , deps : [ :backup , :receipt ] ) |&gt; Oban . insert_all ( ) end Here we use new_workflow/0,1 to initialize a new workflow (later on we will look at how to customize workflow behaviour with various options). Jobs are then added to the workflow with add/3,4 , where the first argument is an existing workflow, the second argument is unique name, the third is a job changeset, and the fourth an optional list of dependencies. Finally, the workflow passes directly to Oban.insert_all/1,2 , which handles inserting each of the jobs into the database atomically. All workers that use Workflow have new_workflow and add functions defined, so it doesn't matter which module you use to initialize or to add new jobs. Dependency resolution guarantees that jobs execute in the order receipt -&gt; email -&gt; backup -&gt; delete , even if one of the jobs fails and needs to retry. Any job that defines a dependency will wait for each upstream dependency to complete before it starts. Recall how our original specification stated that there could be multiple email and backup jobs? The workflow we've built only handles one email or backup job at a time. Let's modify the workflow to fan-out to multiple email and backup jobs and then fan-in to the final delete job: def archive_account_workflow ( acc_id ) do subscribers = subscribers_for_account ( acc_id ) documents = documents_for_account ( acc_id ) email_deps = Enum . map ( subscribers , &amp; &quot;email_ \#{ &amp;1 . id } &quot; ) backup_deps = Enum . map ( documents , &amp; &quot;backup_ \#{ &amp;1 . id } &quot; ) delete_deps = email_deps ++ backup_deps FinalReceipt . new_workflow ( ) |&gt; FinalReceipt . add ( :receipt , FinalReceipt . new ( %{ id : acc_id } ) ) |&gt; add_email_jobs ( subscribers ) |&gt; add_backup_jobs ( documents ) |&gt; DeleteAccount . add ( :delete , DeleteAccount . new ( %{ id : acc_id } ) , deps : delete_deps ) end defp add_email_jobs ( workflow , subscribers ) do Enum . reduce ( subscribers , workflow , fn %{ id : id , email : email } , acc -&gt; EmailSubscriber . add ( acc , &quot;email_ \#{ id } &quot; , EmailSubscriber . new ( %{ email : email } ) , deps : [ :receipt ] ) end ) end defp add_backup_jobs ( workflow , subscribers ) do Enum . reduce ( subscribers , workflow , fn %{ id : id } , acc -&gt; BackupPost . add ( acc , &quot;backup_ \#{ id } &quot; , BackupPost . new ( %{ id : id } ) , deps : [ :receipt ] ) end ) end Now the workflow will run all of the email and backup jobs we need, before deleting the account. To confirm the flow without inserting and executing jobs we can visualize it. Typespecs 📚 In order to bridge the gap between module level docs and a guide, each section includes a typespec for the corresponding function. The snippet below defines the types listed in each section. @type name :: atom ( ) | String . t ( ) @type add_option :: { :deps , [ name ( ) ] } @type new_option :: { :ignore_cancelled , boolean ( ) } | { :ignore_discarded , boolean ( ) } | { :waiting_delay , pos_integer ( ) } | { :waiting_limit , pos_integer ( ) } | { :waiting_snooze , pos_integer ( ) } | { :workflow_id , String . t ( ) } @type append_option :: new_option ( ) | { :check_deps , boolean ( ) }","ref":"workflow.html#usage","title":"Workflow Worker - Usage","type":"extras"},{"doc":"@callback new_workflow ( [ new_option ( ) ] ) :: Workflow . t ( ) @callback add ( Workflow . t ( ) , name ( ) , Changeset . t ( ) , [ add_option ( ) ] ) :: Workflow . t ( ) Workflows use conservative defaults for safe, and relatively quick, dependency resolution. You can customize the safety checks, waiting times and resolution intensity by providing a few top-level options: ignore_cancelled — regard cancelled dependencies as completed rather than halting the workflow. Defaults to false . ignore_discarded — regard discarded dependencies as completed rather than halting the workflow. Defaults to false . waiting_delay — the number of milliseconds to wait between dependency checks. This value, combined with the waiting_limit , dictates how long a job will await upstream dependencies. Defaults to 1000ms . waiting_limit — the number of times to retry dependency checks. This value, combined with waiting_delay , dictates how long a job will await upstream dependencies. Defaults to 10 . waiting_snooze — the number of seconds a job will snooze after awaiting executing upstream dependencies. If upstream dependencies are scheduled or retryable then the job snoozes until the latest dependency's scheduled_at timestamp. The following example creates a workflow with all of the available options: workflow = MyWorkflow . new_workflow ( ignore_cancelled : true , ignore_discarded : true , waiting_delay : :timer . seconds ( 5 ) , waiting_limit : 5 , waiting_snooze : 10 ) Internally, the meta field stores options for each job. That makes it possible to set or override workflow options per-job. For example, configure a single job to ignore cancelled dependencies and another to snooze longer: MyWorkflow . new_workflow ( ) |&gt; MyWorkflow . add ( :a , MyWorkflow . new ( %{ } ) ) |&gt; MyWorkflow . add ( :b , MyWorkflow . new ( %{ } , meta : %{ ignore_cancelled : true } ) , deps : [ :a ] ) |&gt; MyWorkflow . add ( :c , MyWorkflow . new ( %{ } , meta : %{ waiting_snooze : 30 } ) , deps : [ :b ] )","ref":"workflow.html#creating-workflows","title":"Workflow Worker - Creating Workflows","type":"extras"},{"doc":"@callback gen_id ( ) :: String . t ( ) By default workflow_id is a version 4 random UUID . This is more than sufficient to ensure that workflows are unique for any period of time. However, if you require more control you can override workflow_id generation at the worker level, or pass a value directly to the new_workflow/2 function. To override the workflow_id for a particular workflow you override the gen_id callback: defmodule MyApp.Workflow do use Oban.Pro.Workers.Workflow # Generate a 24 character long random string instead @impl true def gen_id do 24 |&gt; :crypto . strong_rand_bytes ( ) |&gt; Base . encode64 ( ) end ... end The gen_id/0 callback works for random/non-deterministic id generation. If you'd prefer to use a deterministic id instead you can pass the workflow_id in as an option to new_workflow/2 : MyApp.Workflow . new_workflow ( workflow_id : &quot;custom-id&quot; ) Using this technique you can verify the workflow_id in tests or append to the workflow manually after it was originally created.","ref":"workflow.html#generating-workflow-ids","title":"Workflow Worker - Generating Workflow IDs","type":"extras"},{"doc":"@callback stream_workflow_jobs ( Job . t ( ) , Keyword . t ( ) ) :: Enum . t ( ) The stream_workflow_jobs/1,2 function makes it possible to lazily load all jobs in a workflow from within your workers. For example, to fetch all of the completed jobs in a workflow: defmodule MyApp.Workflow do use Oban.Pro.Workers.Workflow @impl Workflow def process ( % Job { } = job ) do { :ok , workflow_jobs } = MyApp.Repo . transaction ( fn -&gt; job |&gt; stream_workflow_jobs ( ) |&gt; Stream . filter ( &amp; &amp;1 . state == &quot;completed&quot; ) |&gt; Enum . to_list ( ) end ) do_things_with_jobs ( workflow_jobs ) :ok end end Note that streaming is provided by Ecto's Repo.stream , and it must take place within a transaction. Using a stream lets you control the number of jobs loaded from the database, minimizing memory usage for large workflows.","ref":"workflow.html#streaming-workflow-jobs","title":"Workflow Worker - Streaming Workflow Jobs","type":"extras"},{"doc":"@callback append_workflow ( Job . t ( ) | [ Job . t ( ) ] , [ append_option ( ) ] ) :: Workflow . t ( ) Sometimes all jobs aren't known when the workflow is created. In that case, you can add more jobs with optional dependency checking using append_workflow/2 . An appended workflow starts with one or more jobs, which reuses the original workflow id, and optionally builds a set of dependencies for checking. In this example we disable deps checking with check_deps: false : defmodule MyApp.WorkflowWorker do use Oban.Pro.Workers.Workflow @impl Workflow def process ( % Job { } = job ) do jobs = job |&gt; append_workflow ( check_deps : false ) |&gt; add ( :d , WorkerD . new ( %{ } ) , deps : [ :a ] ) |&gt; add ( :e , WorkerE . new ( %{ } ) , deps : [ :b ] ) |&gt; add ( :f , WorkerF . new ( %{ } ) , deps : [ :c ] ) |&gt; Oban . insert_all ( ) { :ok , jobs } end end The new jobs specify deps on preexisting jobs named :a , :b , and :c , but there isn't any guarantee those jobs actually exist. That could lead to an incomplete workflow where the new jobs may never complete. To be safe and check jobs while appending we'll fetch all of the previous jobs and feed them in: defmodule MyApp.WorkflowWorker do use Oban.Pro.Workers.Workflow @impl Workflow def process ( % Job { } = job ) do { :ok , jobs } = MyApp.Repo . transaction ( fn -&gt; job |&gt; stream_workflow_jobs ( ) |&gt; Enum . to_list ( ) end ) jobs |&gt; append_workflow ( ) |&gt; add ( :d , WorkerD . new ( %{ } ) , deps : [ :a ] ) |&gt; add ( :e , WorkerE . new ( %{ } ) , deps : [ :b ] ) |&gt; add ( :f , WorkerF . new ( %{ } ) , deps : [ :c ] ) |&gt; Oban . insert_all ( ) :ok end end Now there isn't any risk of an incomplete workflow, at the expense of loading some extraneous jobs.","ref":"workflow.html#appending-to-a-workflow","title":"Workflow Worker - Appending to a Workflow","type":"extras"},{"doc":"@callback to_dot ( Workflow . t ( ) ) :: String . t ( ) Workflows are a type of Directed Acyclic Graph , also known as a DAG. That means we can describe a workflow as a graph of jobs and dependencies, where execution flows between jobs. By converting the workflow into DOT notation, a standard graph description language, we can render visualizations! Dot generation relies on libgraph , which is an optional dependency. You'll need to specify it as a dependency before generating dot output: def deps do [ { :libgraph , &quot;~&gt; 0.7&quot; } ] end Once you've installed libgraph , we can use to_dot/1 to convert a workflow. As with new_workflow and add , all workflow workers define a to_dot/1 function that takes a workflow and returns a dot formatted string. For example, calling to_dot/1 with the account archiving workflow from above: FinalReceipt . to_dot ( archive_account_workflow ( 123 ) ) Generates the following dot output, where each vertex is a combination of the job's name in the workflow and its worker module: strict digraph { &quot;delete (MyApp.DeleteAccount)&quot; &quot;backup_1 (MyApp.BackupPost)&quot; &quot;backup_2 (MyApp.BackupPost)&quot; &quot;backup_3 (MyApp.BackupPost)&quot; &quot;receipt (MyApp.FinalReceipt)&quot; &quot;email_1 (MyApp.EmailSubscriber)&quot; &quot;email_2 (MyApp.EmailSubscriber)&quot; &quot;backup_1 (MyApp.BackupPost)&quot; -&gt; &quot;delete (MyApp.DeleteAccount)&quot; [weight=1] &quot;backup_2 (MyApp.BackupPost)&quot; -&gt; &quot;delete (MyApp.DeleteAccount)&quot; [weight=1] &quot;backup_3 (MyApp.BackupPost)&quot; -&gt; &quot;delete (MyApp.DeleteAccount)&quot; [weight=1] &quot;receipt (MyApp.FinalReceipt)&quot; -&gt; &quot;backup_1 (MyApp.BackupPost)&quot; [weight=1] &quot;receipt (MyApp.FinalReceipt)&quot; -&gt; &quot;backup_2 (MyApp.BackupPost)&quot; [weight=1] &quot;receipt (MyApp.FinalReceipt)&quot; -&gt; &quot;backup_3 (MyApp.BackupPost)&quot; [weight=1] &quot;receipt (MyApp.FinalReceipt)&quot; -&gt; &quot;email_1 (MyApp.EmailSubscriber)&quot; [weight=1] &quot;receipt (MyApp.FinalReceipt)&quot; -&gt; &quot;email_2 (MyApp.EmailSubscriber)&quot; [weight=1] &quot;email_1 (MyApp.EmailSubscriber)&quot; -&gt; &quot;delete (MyApp.DeleteAccount)&quot; [weight=1] &quot;email_2 (MyApp.EmailSubscriber)&quot; -&gt; &quot;delete (MyApp.DeleteAccount)&quot; [weight=1] } Now we can take that dot output and render it using a tool like graphviz . The following example function accepts a workflow and renders it out as an SVG: defmodule WorkflowRenderer do alias Oban.Pro.Workers.Workflow def render ( workflow ) do dot_path = &quot;workflow.dot&quot; svg_path = &quot;workflow.svg&quot; File . write! ( dot_path , Workflow . to_dot ( workflow ) ) System . cmd ( &quot;dot&quot; , [ &quot;-T&quot; , &quot;svg&quot; , &quot;-o&quot; , svg_path , dot_path ] ) end end With graphviz installed, that will generate a SVG of the workflow: delete (MyApp.DeleteAccount) backup_1 (MyApp.BackupPost) backup_2 (MyApp.BackupPost) backup_3 (MyApp.BackupPost) receipt (MyApp.FinalReceipt) email_1 (MyApp.EmailSubscriber) email_2 (MyApp.EmailSubscriber) Looking at the visualized graph we can clearly see how the workflow starts with a single render job, fans-out to multiple email and backup jobs, and finally fans-in to the delete job—exactly as we planned!","ref":"workflow.html#visualizing-workflows","title":"Workflow Worker - Visualizing Workflows","type":"extras"},{"doc":"Jobs aren't executed until all upstream dependencies have completed. This includes waiting on retries, scheduled jobs, or snoozing. Dependency resolution relies on jobs lingering in the database after execution. If your system prunes job dependencies then the workflow may never complete.","ref":"workflow.html#implementation-notes","title":"Workflow Worker - Implementation Notes","type":"extras"},{"doc":"🧭 Oban.Web is a view of Oban's inner workings that you host directly within your Phoenix application. Powered by Phoenix Live View, it is extremely lightweight and continuously updated. It builds on features provided by Oban.Pro and is available through a paid license. Get started with installation .","ref":"web_overview.html","title":"Overview","type":"extras"},{"doc":"Live Inspection — Monitor background job activity across all of your nodes in real time. Composable Filtering — Sift through jobs instantly with any combination of queue, state, node, worker and other metadata. Powerful Search — Intelligently search through job arguments instantly, with support for partial matches and auto-correction. Detailed Inspection — View job details including when, where and how it was ran (or how it failed to run). Batch Actions — Cancel, delete and retry selected jobs or all jobs matching the current filters. Queue Controls — Scale, pause, resume and stop queues across all running nodes with a couple of clicks. Access Controll — Allow admins to control queues and interract with jobs while restricting other users to read-only use of the dashboard. Action Logging — Use telemetry events to instrument and report all of a user's dashboard activity.","ref":"web_overview.html#features","title":"Overview - Features","type":"extras"},{"doc":"All notable changes to Oban.Web are documented here.","ref":"web-changelog.html","title":"Changelog","type":"extras"},{"doc":"Fixed Safely assign state from params when re-mounting to prevent a missing key error.","ref":"web-changelog.html#v2-7-4-2021-09-27","title":"Changelog - v2.7.4 — 2021-09-27","type":"extras"},{"doc":"Changed Expose documented, but unfortunately hidden, socket_path option. The socket path is no longer hard coded as /live . It is now possible to configure an alternate path when mounting the dashboard. Fixed Guard against nil global limit when calculating sidebar stats.","ref":"web-changelog.html#v2-7-3-2021-09-15","title":"Changelog - v2.7.3 — 2021-09-15","type":"extras"},{"doc":"Changed Upgrade minimum Phoenix Live View dependency to 0.16 .","ref":"web-changelog.html#v2-7-2-2021-08-13","title":"Changelog - v2.7.2 — 2021-08-13","type":"extras"},{"doc":"Fixed Guard agaisnt empty params on initial render when the dashboard is re-mounted. Added Add id: qualifier for id and multiple id based searching.","ref":"web-changelog.html#v2-7-1-2021-06-01","title":"Changelog - v2.7.1 — 2021-06-01","type":"extras"},{"doc":"Enhanced Search Syntax Job search has a new advanced syntax that lets you limit searches to particular fields, use negated queries, look for exact matches, and navigate into structured fields like args and meta . Here are a few examples to show what's possible: alpha in:worker — only look within worker for &quot;alpha&quot; alpha -omega in:tags,meta — look for &quot;alpha&quot; and not &quot;omega&quot; within tags or meta fields business in:args.account.plan — dig into the account and plan maps within args to look for &quot;business&quot; a1b2c3d4e5 in:meta.workflow_id — look for an exact workflow id See the new searching guide for more examples. Push State Navigation Navigation, filtering, and search now use push-state for navigation. This makes it possible to deep-link into job details, save searches for later, or to share filter combinations with co-workers.","ref":"web-changelog.html#v2-7-0-2021-05-25","title":"Changelog - v2.7.0 — 2021-05-25","type":"extras"},{"doc":"Fixed Strictly guard against missing stats ETS table During a restart it was possible for the dashboard to receive a refresh signal before the stats plugin had finished mounting. This introduces a stricter set of conditionals that will safely degrade stats results back to an empty state when the ets table isn't available. Increase the timeout while waiting for config on dashboard mount. The restart period for some apps exceeds 5s and mounting would still cause an error.","ref":"web-changelog.html#v2-6-2-2021-05-19","title":"Changelog - v2.6.2 — 2021-05-19","type":"extras"},{"doc":"Fixed Update stale sidebar state and queue counts when there aren't any jobs matching the given state or queue. When a queue/state combination had no values, the counts on the sidebar would &quot;stick&quot; and only show the previous non-zero value. This was most noticeable with the executing state as it worked through jobs quickly. Apply access controls on the server side as well as the client. It is no longer possible to use blocked actions by manually restoring button HTML in the browser.","ref":"web-changelog.html#v2-6-1-2021-04-23","title":"Changelog - v2.6.1 — 2021-04-23","type":"extras"},{"doc":"Remove Reliance on Beats Previously, node and queue information on the dashboard was powered by heartbeat records stored in the oban_beats table. That is no longer the case and the dashboard is entirely independent from beats and from Pro. Check out the Oban v2.6 upgrade guide for a walkthrough on using the Gossip plugin for realtime updates. Content Security Policy To secure the dashboard, or comply with an existing CSP within your application, you can specify nonce keys for images, scripts and styles. For example, to pull the asset nonce from a single :my_csp_nonce assignment: oban_dashboard ( &quot;/oban&quot; , csp_nonce_assign_key : :my_csp_nonce ) See the CSP install guide for details. Changed Remove dependency on oban_pro (again). The dashboard is driven by events and data provided by Oban alone. Bump the Oban dependency ~&gt; v2.6","ref":"web-changelog.html#v2-6-0-2021-04-02","title":"Changelog - v2.6.0 — 2021-04-02","type":"extras"},{"doc":"Changed Restore a dependency on oban_pro , as removal broke the install flow for many applications that had an implicit dependency. Bump the minimum oban version from v2.3 to v2.4 to enhance compatibility with oban_pro . Bump the minimum Elixir version from v1.8 to v1.9, matching recent oban and oban_pro releases.","ref":"web-changelog.html#v2-5-2-2021-03-04","title":"Changelog - v2.5.2 — 2021-03-04","type":"extras"},{"doc":"Fixed Conditionally increase max_attempts when bulk retrying jobs. The check constraint added by Oban 2.4.0 exposed a bug which made it impossible to retry jobs that had exhausted all available attempts.","ref":"web-changelog.html#v2-5-1-2021-01-28","title":"Changelog - v2.5.1 — 2021-01-28","type":"extras"},{"doc":"Web Resolver Behaviour A new Oban.Web.Resolver behaviour module allows users to resolve the current user when loading the dashboard, apply per-user access controls and set per-user refresh rates. defmodule MyApp.Resolver do @behaviour Oban.Web.Resolver @impl true def resolve_user ( conn ) do conn . private . current_user end @impl true def resolve_access ( user ) do if user . admin? do [ cancel_jobs : true , delete_jobs : true , retry_jobs : true ] else :read_only end end @impl true def resolve_refresh ( _user ) , do : 1 end Pass your Resolver callback to oban_dashboard in your router: scope &quot;/&quot; do pipe_through :browser oban_dashboard &quot;/oban&quot; , resolver : MyApp.Resolver end Viola, your dashboard now has per-user access controls! Now only admins can cancel, retry or delete jobs while other users can still monitor running jobs and check stats. See the new Customization guide for more examples and a rundown of available access controls. Telemetry Integration The Oban.Web.Telemetry module adds events for instrumentation, logging, error reporting and activity auditing. Action events are emitted whenever a user performs a write operation with the dashboard, e.g. pausing a queue, cancelling a job, etc. Web now ships with a log handler that you can attach to get full dashboard audit logging. Add this call at application start: Oban.Web.Telemetry . attach_default_logger ( :info ) It will output structured JSON logs matching the format that Oban uses, including the user that performed the action: { &quot;action&quot; :&quot;cancel_jobs&quot; , &quot;duration&quot; : 2544 , &quot;event&quot; :&quot;action:stop&quot; , &quot;job_ids&quot; : [ 290950 ] , &quot;oban_name&quot; :&quot;Oban&quot; , &quot;source&quot; :&quot;oban_web&quot; , &quot;user&quot; : 1818 } See the new Telemetry guide for event details! Other Improvements Remove a hard dependency on oban_pro . If you rely on any Pro plugins be sure to specify an oban_pro dependency in your mix.exs . Remove erroneous batch &quot;Delete&quot; action from states where it shouldn't apply, e.g. executing . Require confirmation before deleting jobs. Deleting is permanent and irreversible, unlike cancelling or other possible actions. Optimize the query that calculates queue and state counts in the sidebar. With millions of jobs the query could take longer than the refresh rate, leading to problems.","ref":"web-changelog.html#v2-5-0-2021-01-15","title":"Changelog - v2.5.0 — 2021-01-15","type":"extras"},{"doc":"Added Allow default_refresh option when mounting a dashboard in a router. The refresh rate controls how frequently the server pulls statistics from the database, and when data is pushed from the server. The default refresh rate is 1 second, but you can now customize it when mounting a dashboard. For example, to set the default refresh to 5 seconds: scope &quot;/&quot; do pipe_through :browser oban_dashboard &quot;/oban&quot; , default_refresh : 5 end This makes it easy to reduce the load on your database when a lot of users are viewing Oban dashboards.","ref":"web-changelog.html#v2-4-0-2020-12-11","title":"Changelog - v2.4.0 — 2020-12-11","type":"extras"},{"doc":"Changed Upgrade minimum Phoenix Live View dependency to 0.15 . Fixed Allow retrying or deleting cancelled jobs when they were never attempted.","ref":"web-changelog.html#v2-3-1-2020-11-27","title":"Changelog - v2.3.1 — 2020-11-27","type":"extras"},{"doc":"Added Cancelling a job transitions the job to the cancelled state rather than discarded . The discarded state is now reserved for jobs that exhaust retry attempts or are purposefully discarded through a {:discard, reason} tuple. Display meta in the job details view. Changed Upgrade Oban dependency to ~&gt; 2.3.0 to support the new cancelled state, and meta field.","ref":"web-changelog.html#v2-3-0-2020-11-06","title":"Changelog - v2.3.0 — 2020-11-06","type":"extras"},{"doc":"Changed Replace the queue scale slider with a number input and a submit button. Aside from how difficult it was to scale accurately with a slider, it would fire erroneous update events due to DOM changes.","ref":"web-changelog.html#v2-2-3-2020-10-15","title":"Changelog - v2.2.3 — 2020-10-15","type":"extras"},{"doc":"Changed Upgrade Oban dependency to ~&gt; 2.2.0 along with fixes for the move to Oban.Registry and Oban.Repo .","ref":"web-changelog.html#v2-2-2-2020-10-11","title":"Changelog - v2.2.2 — 2020-10-11","type":"extras"},{"doc":"Fixed Prevent layout jumb between initial render and subsequent mount. This was caused by the lack of a closing tag that was later injected on DOM update. Prevent bleeding events between nested components. The lack of a closing tag seemed to confuse event handling and allowed the wrong component to receive events. Correct touch event handling in the sidebar for mobile devices.","ref":"web-changelog.html#v2-2-1-2020-09-29","title":"Changelog - v2.2.1 — 2020-09-29","type":"extras"},{"doc":"Added Support multiple isolated dashboards by providing an oban_name option to oban_dashboard/2 calls in the Phoenix router. For example, to mount dashboards for multiple Oban instances in an umbrella: defmodule MyAppWeb.Router do use Phoenix.Router import Oban.Web.Router scope &quot;/&quot; , MyAppWeb do pipe_through [ :browser ] oban_dashboard &quot;/oban/web&quot; , oban_name : MyWeb.Oban oban_dashboard &quot;/oban/ops&quot; , oban_name : MyOps.Oban oban_dashboard &quot;/oban/ingestion&quot; , oban_name : MyIngestion.Oban end end Changed Clicking anywhere on a job row now opens the details view. Previously only clicking the rightmost &quot;expand&quot; icon opened details, and clicking on the worker name would search for other jobs with the same worker. Now the rightmost &quot;document search&quot; icon performs a related job search. Close the job details view when a user clicks a node name, job state, or queue name in the sidebar. Prevent ugly overflow issues when there are long node or queue names in the sidebar. Note that the full name is still visible on hover.","ref":"web-changelog.html#v2-2-0-2020-09-11","title":"Changelog - v2.2.0 — 2020-09-11","type":"extras"},{"doc":"Fixed Correct expected plugin naming scheme for compatibility with Oban &gt;= 2.1 . Increase stats collection and activation timeouts to compensate for spikes in table size or additional load on the database pool from other processes.","ref":"web-changelog.html#v2-1-1-2020-08-24","title":"Changelog - v2.1.1 — 2020-08-24","type":"extras"},{"doc":"Added Support alternate live socket transports, namely the &quot;longpoll&quot; transport. Users can specify either &quot;websocket&quot; or &quot;longpoll&quot; directly from the oban_dashboard/2 call in their Phoenix router (which avoids global configuration). Fixed Skip full text search for PostgreSQL versions &lt; 11.0 . Older PG versions don't have proper support for JSONB vectorization or web style search operations.","ref":"web-changelog.html#v2-1-0-2020-08-06","title":"Changelog - v2.1.0 — 2020-08-06","type":"extras"},{"doc":"Changed Upgrade to Oban 2.0.0 , LiveView ~&gt; 0.14 , and add a dependency on Oban Pro 0.3.0 . Add simple load less/load more pagination for browsing through jobs Move details view inline, eliminating modal sizing and scrolling issues Display queue_time and run_time in the job details view Restore worker filtering using auto-populated search terms Introduce Oban.Web.Router for more convenient and flexible dashboard routing. The new module provides an oban_dashboard/2 macro, which prevents scoping mistakes and provides the foundation of better route support. Here is how to mount a dashboard at /oban with the new macro: defmodule MyAppWeb.Router do use Phoenix.Router import Oban.Web.Router scope &quot;/&quot; , MyAppWeb do pipe_through [ :browser ] oban_dashboard &quot;/oban&quot; end end Add bulk actions for all jobs in table view. It's now possible to select one or more jobs in the table and then cancel, run, retry or delete them all. Add a dropdown that controls the refresh rate of jobs and stats in the dashboard. Lowering the refresh rate can reduce query overhead, or pause updates entirely. Expose pause/resume controls and a slider for scaling in the queue side bar. Eliminate migrations. Migrations are no longer necessary for updates or full text searching. You may safely undo previous migrations. Consider job tags along with args and worker name when searching. More intelligent search using &quot;and&quot;, &quot;or&quot; and respecting parenthesis. Consistently toggle filters on and off from the sidebar, rather than matching on a filter list in the header.","ref":"web-changelog.html#v2-0-0-2020-07-10","title":"Changelog - v2.0.0 — 2020-07-10","type":"extras"},{"doc":"Changed Upgrade to Phoenix ~&gt; 1.5 , LiveView ~&gt; 0.12 and PubSub ~&gt; 2.0 . None of these upgrades required changes to ObanWeb, they are meant to enable upgrades for host applications.","ref":"web-changelog.html#v1-5-0-2020-04-27","title":"Changelog - v1.5.0 — 2020-04-27","type":"extras"},{"doc":"Changed Upgrade to LiveView ~&gt; 0.10 along with requisite changes to use @inner_content in the layout template. This prevents the view from hanging with a blank screen on load.","ref":"web-changelog.html#v1-4-0-2020-03-24","title":"Changelog - v1.4.0 — 2020-03-24","type":"extras"},{"doc":"Fixed Prevent FunctionClauseError when closing the dashboard before it has finished mounting.","ref":"web-changelog.html#v1-3-1-2020-03-18","title":"Changelog - v1.3.1 — 2020-03-18","type":"extras"},{"doc":"Changed Upgrade to LiveView ~&gt; 0.9 along with the requisite changes to flash handling. Note, this requires pipelines in the host app to use fetch_live_flash instead of fetch_flash .","ref":"web-changelog.html#v1-3-0-2020-03-10","title":"Changelog - v1.3.0 — 2020-03-10","type":"extras"},{"doc":"Fixes Prevent flash map key issue on initial mount. The lack of a show key on the empty flash map caused runtime errors. Prevent losing stats cache on restart by shifting table ownership into the supervisor. Changed Clear completed_at and discarded_at timestamps when descheduling jobs. Upgrade to LiveView ~&gt; 0.7 . Added Display job priority value in the table and detail views. Display job tags in the detail view.","ref":"web-changelog.html#v1-2-0-2020-02-07","title":"Changelog - v1.2.0 — 2020-02-07","type":"extras"},{"doc":"Fixes Correct syntax used for descheduling and discarding jobs Set the discarded_at timestamp when discarding jobs","ref":"web-changelog.html#v1-1-2-2020-02-07","title":"Changelog - v1.1.2 — 2020-02-07","type":"extras"},{"doc":"Fixes Fix job detail refreshing on tick events","ref":"web-changelog.html#v1-1-1-2020-02-06","title":"Changelog - v1.1.1 — 2020-02-06","type":"extras"},{"doc":"Changes Add verbose setting to control log levels. This command mirrors the behavior in Oban and is respected by all queries. Deprecate stats configuration. The stats module is entirely overhauled so that it only refreshes when one or more users are connected. That prevents it from using any connections or performing any queries while testing, which renders the stats option pointless. Add stats_interval to control how often node and queue counts are refreshed. The default value is every 1s. Add tick_interval to control how often the jobs table and job details are refreshed. The default value is every 500ms.","ref":"web-changelog.html#v1-1-0-2020-02-06","title":"Changelog - v1.1.0 — 2020-02-06","type":"extras"},{"doc":"Fixes Display discarded_at time in detail timeline view. Prevent jitter when clearing stats cache.","ref":"web-changelog.html#v1-0-1-2020-01-29","title":"Changelog - v1.0.1 — 2020-01-29","type":"extras"},{"doc":"Changed Upgrade Oban and set dependency to ~&gt; 1.0.0 Update stats and the queries that power stats to rely on more frequent refreshes and fewer database calls. The average refresh time dropped from ~310ms to ~180ms .","ref":"web-changelog.html#v1-0-0-2020-01-29","title":"Changelog - v1.0.0 — 2020-01-29","type":"extras"},{"doc":"Changed Upgrade to LiveView ~&gt; 0.5 and test with 0.6 .","ref":"web-changelog.html#v0-8-0-2020-01-23","title":"Changelog - v0.8.0 — 2020-01-23","type":"extras"},{"doc":"Added Add detail modal for inspecting job arguments, errors and state timings.","ref":"web-changelog.html#v0-7-0-2020-01-08","title":"Changelog - v0.7.0 — 2020-01-08","type":"extras"},{"doc":"Changed Display duration or distance in words depending on job state. This clarifies timestamp information. Relative timestamps have an formatted absolute value as the &quot;title&quot; attribute as well. Improved ordering for various job states. Many states were reversed, making the table view seem broken. Fixes Ensure all assigns are available on render when a disconnected node reconnects. This fixes an issue when viewing the UI in development.","ref":"web-changelog.html#v0-6-3-2019-12-15","title":"Changelog - v0.6.3 — 2019-12-15","type":"extras"},{"doc":"Changed Add support for explicitly disabling stats rather than inferring based on queue configuration. This prevents issues in production environments where certain nodes (i.e. web) aren't processing any queues, but the the UI is accessed. ObanWeb configuration now differs from Oban configuration and you'll need to specify it separately. Specify the repo in config.exs : config :my_app , ObanWeb , repo : MyApp.Repo Disable the stats tracker in the test environment: config :my_app , ObanWeb , stats : false","ref":"web-changelog.html#v0-6-2-2019-12-05","title":"Changelog - v0.6.2 — 2019-12-05","type":"extras"},{"doc":"Fixes Ignore the stats server when queues is empty.","ref":"web-changelog.html#v0-6-1-2019-11-22","title":"Changelog - v0.6.1 — 2019-11-22","type":"extras"},{"doc":"Before continuing, be sure you have Oban up and running in your app! Oban.Web is built with Phoenix Live View and it relies on a working installation of it in your application. If you don't have Live View installed, follow these instructions to get started. Oban.Web is delivered as a hex package named oban_web , which is published privately to our self-hosted package repository. The package is entirely self contained—it doesn't hook into your asset pipeline at all. Before you can pull the package into your application you need to add a new oban hex repo. First, grab the OBAN_KEY_FINGERPRINT and OBAN_LICENSE_KEY from your account page. Then, ensure hex is up-to-date: mix local.hex --force Finally, run the following mix hex.repo command: mix hex.repo add oban https://getoban.pro/repo \\ --fetch-public-key $OBAN_KEY_FINGERPRINT \\ --auth-key $OBAN_LICENSE_KEY ⚠️ You'll also need to authenticate on any other development machines, build servers and CI instances. There are also guides to help with authenticating on Gigalixir and Heroku . Now that you're authenticated you're ready to add oban_web as a dependency for your application. Open mix.exs and add the following line: { :oban_web , &quot;~&gt; 2.7&quot; , repo : &quot;oban&quot; } Now fetch your dependencies: $ mix deps.get This will fetch both oban_web and oban_pro , if you haven't already installed oban_pro . The Gossip plugin and the Stats plugin are necessary for the dashboard to function properly. Add them to your Oban configuration in config.exs : config :my_app , Oban , repo : MyApp.Repo , queues : [ alpha : 10 , gamma : 10 , delta : 10 ] , plugins : [ Oban.Plugins.Gossip , Oban.Web.Plugins.Stats ] After configuration you can mount the dashboard within your application's router.ex : # lib/my_app_web/router.ex use MyAppWeb , :router import Oban.Web.Router ... scope &quot;/&quot; do pipe_through :browser oban_dashboard &quot;/oban&quot; end Here we're using &quot;/oban&quot; as the mount point, but it can be anywhere you like. After you've verified that the dashboard is loading you'll probably want to restrict access to the dashboard via authentication, e.g. with Basic Auth . Installation is complete and you're all set! Start your Phoenix server, point your browser to where you mounted Oban and start monitoring your jobs.","ref":"web_installation.html","title":"Installation","type":"extras"},{"doc":"Applications that run multiple Oban instances can mount a dashboard for each instance. Set the mounted dashboard's :oban_name to match the corresponding supervision tree's name. For example, given two configured Oban instances, Oban and MyAdmin.Oban : config :my_app , Oban , repo : MyApp.Repo , name : Oban , ... config :my_admin , Oban , repo : MyAdmin.Repo , name : MyAdmin.Oban , ... You can then mount both dashboards in your router: scope &quot;/&quot; do pipe_through :browser oban_dashboard &quot;/oban&quot; , oban_name : Oban oban_dashboard &quot;/oban/admin&quot; , oban_name : MyAdmin.Oban end Note that the default name is Oban , setting oban_name: Oban in the example above was purely for demonstration purposes.","ref":"web_installation.html#running-multiple-dashboards","title":"Installation - Running Multiple Dashboards","type":"extras"},{"doc":"Applications that use a live socket other than &quot;/live&quot; can override the default socket path in the router. For example, if your live socket is hosted at /oban_live : socket &quot;/oban_live&quot; , Phoenix.LiveView.Socket scope &quot;/&quot; do pipe_through :browser oban_dashboard &quot;/oban&quot; , socket_path : &quot;/oban_live&quot; end If your application is hosted in an environment that doesn't support websockets you can use longpolling as an alternate transport. To start, make sure that your live socket is configured for longpolling: socket &quot;/live&quot; , Phoenix.LiveView.Socket , longpoll : [ connect_info : [ session : @session_options ] , log : false ] Then specify &quot;longpoll&quot; as your transport: scope &quot;/&quot; do pipe_through :browser oban_dashboard &quot;/oban&quot; , transport : &quot;longpoll&quot; end","ref":"web_installation.html#customizing-the-socket-connection","title":"Installation - Customizing the Socket Connection","type":"extras"},{"doc":"To secure the dashboard, or comply with an existing CSP within your application, you can specify nonce keys for images, scripts and styles. You'll configure the CSP nonce assign key in your router, where the dashboard is mounted. For example, to use a single nonce for all three asset types: oban_dashboard ( &quot;/oban&quot; , csp_nonce_assign_key : :my_csp_nonce ) That instructs the dashboard to extract a generated nonce from the assigns map on the plug connection, at the :my_csp_nonce key. Instead, you can specify different keys for each asset type: oban_dashboard ( &quot;/oban&quot; , csp_nonce_assign_key : %{ img : :img_csp_nonce , style : :style_csp_nonce , script : :script_csp_nonce } ) Note that using the CSP is entirely optional.","ref":"web_installation.html#content-security-policy","title":"Installation - Content Security Policy","type":"extras"},{"doc":"Implementing a Oban.Web.Resolver callback module allows you to customize the dashboard per-user, i.e. setting access controls or the default refresh rate. As a simple example, let's define a module that makes the dashboard read only: defmodule MyApp.Resolver do @behaviour Oban.Web.Resolver @impl true def resolve_access ( _user ) , do : :read_only end Then specify MyApp.Resolver as your resolver: scope &quot;/&quot; do pipe_through :browser oban_dashboard &quot;/oban&quot; , resolver : MyApp.Resolver end See Customizing the Dashboard for details on the Resolver behaviour.","ref":"web_installation.html#customizing-with-a-resolver-callback-module","title":"Installation - Customizing with a Resolver Callback Module","type":"extras"},{"doc":"Oban Web uses Telemetry to provide instrumentation and to power logging of dashboard activity. See the Telemetry guide for a breakdown of emitted events and how to use the default logger. Trouble installing? Have questions? Take a look at the troubleshooting guide to see if your issue is covered. If not, or if you need any help, stop by the #oban channel in Elixir Slack .","ref":"web_installation.html#integrating-with-telemetry","title":"Installation - Integrating with Telemetry","type":"extras"},{"doc":"Dashboard customization is done through a callback module that implements the Oban.Web.Resolver behaviour. Each of the callback functions are optional and will fall back to the default implementation, which is shown here as a starting point: defmodule MyApp.Resolver do @behaviour Oban.Web.Resolver @impl true def resolve_user ( _conn ) , do : nil @impl true def resolve_access ( _user ) , do : :all @impl true def resolve_refresh ( _user ) , do : 1 end Jump to details for each of the callbacks: resolve_user/1 resolve_access/1 resolve_refresh/1","ref":"web_customizing.html","title":"Customizing the Dashboard","type":"extras"},{"doc":"With the resolve_user/1 callback you can extract the current user from a Plug.Conn when the dashboard mounts. The extracted user is passed to all of the other callback functions, allowing you to customize the dashboard per user or role. In a typical plug based auth system the current user is stashed the private map: @impl true def resolve_user ( conn ) do conn . private . current_user end The resolve_user/1 callback is expected to return nil , a map or a struct. However, the resolved user is only passed to other functions in the Resolver and as part of the metadata for audit events, so you're free to use any data type you like.","ref":"web_customizing.html#current-user","title":"Customizing the Dashboard - Current User","type":"extras"},{"doc":"During normal operation users can modify running queues and interact with jobs through the dashboard. In some situations actions such as pausing a queue may be undesired, or even dangerous for operations. Through the resolve_access/1 callback you can tailor precisely which actions the current user can do. The default access level is :all , which permits all users to do any action. To set the dashboard read only and prevent users from performing any actions at all: @impl true def resolve_access ( _user ) , do : :read_only Alternatively, you can use the resolved user to allow admins write access and keep all other users read only: @impl true def resolve_access ( user ) do if user . admin? , do : :all , else : :read_only end You can also specify fine grained access for each of the possible dashboard actions: @impl true def resolve_access ( user ) do if user . admin? do [ cancel_jobs : true , delete_jobs : true , retry_jobs : true ] else :read_only end end This configuration allows admins to cancel jobs, delete jobs and retry jobs. They still can't pause or scale queues because actions which aren't listed are considered disabled . The available fine grained access controls are: :pause_queues :scale_queues :cancel_jobs :delete_jobs :retry_jobs","ref":"web_customizing.html#action-controls","title":"Customizing the Dashboard - Action Controls","type":"extras"},{"doc":"The refresh rate controls how frequently the server pulls statistics from the database, and when data is pushed from the server. The default refresh rate is 1 second, but you can customize it with a resolve_refresh/0 callback. For example, to set the default refresh to 5 seconds: @impl true def resolve_refresh ( _user ) , do : 5 Possible values are: 1 , 2 , 5 , 15 or -1 to disable refreshing. Note that this only sets the default. Users may still choose a different refresh for themselves while viewing the dashboard.","ref":"web_customizing.html#default-refresh","title":"Customizing the Dashboard - Default Refresh","type":"extras"},{"doc":"The search bar supports basic and advanced syntax to whittle down jobs based on multiple fields. Without any qualifiers, your terms are matched against worker , tags , args and meta fields using loose matching or tsquery checks, as appropriate. For more advanced syntax, check the sections below. Note: Multi-field advanced searching is only supported for PostgreSQL 11+. For older versions, only the worker is searched","ref":"searching.html","title":"Searching Jobs","type":"extras"},{"doc":"With the in: qualifier you can restrict your search to the worker , args , meta , tags , or any combination of these. foo in:args — only search within args , using a tsquery based search foo in:meta — only search within meta , using a tsquery based search foo in:tags — only search within tags , using a tsquery based search myapp in:worker — only search within the worker , using a loose ilike style match To search through multiple fields, join them together with a comma. For example: foo in:args,meta foo in:tags,meta,args foo in:worker,tags","ref":"searching.html#qualifier-syntax","title":"Searching Jobs - Qualifier Syntax","type":"extras"},{"doc":"For the jsonb fields, args and meta , you can also use &quot;dot&quot; path syntax to restrict search to nested data. a1b2c3d4 in:meta.worker_id business in:args.user.plan.name Naturally, you can combine path syntax with multi-field syntax: foo in:args.batch_id,meta.worker_id foo in:args.user.plan,tags","ref":"searching.html#nested-syntax","title":"Searching Jobs - Nested Syntax","type":"extras"},{"doc":"The id: qualifier restricts results to one or more jobs by id. Filter down to multiple jobs by separating ids with a comma: id:123 id:123,124,125","ref":"searching.html#id-search","title":"Searching Jobs - Id Search","type":"extras"},{"doc":"The priority field is searchable as well. Use priority: and any combination of values between 0 and 3 to filter jobs by priority: priority:0 priority:0,1 priority:0,1,2,3","ref":"searching.html#priority-search","title":"Searching Jobs - Priority Search","type":"extras"},{"doc":"If your search query contains whitespace, you will need to surround it with quotation marks. For example: &quot;foo bar&quot; alpha not &quot;foo bar&quot;","ref":"searching.html#quoted-terms","title":"Searching Jobs - Quoted Terms","type":"extras"},{"doc":"You can exclude results containing a certain word, using the not syntax. The not operator can only be used for args , meta and tags . It does not work for worker . not alpha foo -bar in:tags","ref":"searching.html#exclude-syntax","title":"Searching Jobs - Exclude Syntax","type":"extras"},{"doc":"You can't use the following wildcard characters as part of your search query: , : ; / \\ ' = * ! ? # $ &amp; + ^ | ~ &lt; &gt; ( ) { } [ ] The search will stripe these symbols and ignore them.","ref":"searching.html#considerations","title":"Searching Jobs - Considerations","type":"extras"},{"doc":"Here are a few examples that combine the available syntax and demonstrate what's possible: alpha in:worker alpha -omega in:tags,meta alpha not omega in:tags priority:0 alpha not &quot;super alpha&quot; in:tags pro in:args.account.plan","ref":"searching.html#advanced-examples","title":"Searching Jobs - Advanced Examples","type":"extras"},{"doc":"Oban Web emits telemetry events for instrumentation, logging, error reporting and activity auditing.","ref":"web_telemetry.html","title":"Telemetry","type":"extras"},{"doc":"An action event is emitted whenever a user performs a write operation with the dashboard, e.g. pausing a queue, cancelling a job, etc. The dashboard emits the following events: [:oban_web, :action, :start] [:oban_web, :action, :stop] [:oban_web, :action, :exception] Action events include the action name, Oban config, the user that performed the action ( if available ), and relevant metadata. In addition, failed actions provide the error type, the error itself, and the stacktrace. The following chart shows the base metadata for each event: event measures metadata :start :system_time :action, :config, :user :stop :duration :action, :config, :user :exception :duration :action, :config, :user, :kind, :error, :stacktrace For :exception events the metadata includes details about what caused the failure. The :kind value is determined by how an error occurred. This chart breaks down the possible actions and their specific metadata: action metadata :pause_queues :queue :scale_queues :queue , :limit :cancel_jobs :job_ids :delete_jobs :job_ids :retry_jobs :job_ids","ref":"web_telemetry.html#action-events","title":"Telemetry - Action Events","type":"extras"},{"doc":"The Oban.Web.Telemetry module ships with a default handler that logs structured JSON for :stop and :exception events. To attach the logger, call attach_default_logger/1 as your application starts: def start ( _type , _args ) do children = [ MyApp.Repo , MyApp.Endpoint , { Oban , oban_opts ( ) } ] Oban.Telemetry . attach_default_logger ( :info ) Oban.Web.Telemetry . attach_default_logger ( :info ) Supervisor . start_link ( children , [ strategy : :one_for_one , name : MyApp.Supervisor ] ) end Here is an example of the JSON output for an action:stop event: { &quot;action&quot;:&quot;cancel_jobs&quot;, &quot;duration&quot;:2544, &quot;event&quot;:&quot;action:stop&quot;, &quot;job_ids&quot;:[290950], &quot;oban_name&quot;:&quot;Oban&quot;, &quot;source&quot;:&quot;oban_web&quot;, &quot;user&quot;:1818 } Event metadata is passed through directly along with these constant fields: duration — Action duration, recorded in native units and logged as microseconds source — Always &quot;oban_web&quot;, which is useful for log filtering user — If the dashboard was mounted with a resolver that implements resolve_user/1 this is the user's id, otherwise null oban_name — The instance that the dashboard is linked to, typically this is &quot;Oban&quot; unless an application is using multiple Oban instances.","ref":"web_telemetry.html#action-logging","title":"Telemetry - Action Logging","type":"extras"}]